<UserControl x:Class="Microsoft.VisualStudio.FSharp.UIResources.AdvancedOptionsControl"
             xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
             xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
             xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" 
             xmlns:d="http://schemas.microsoft.com/expression/blend/2008" 
             xmlns:local="clr-namespace:Microsoft.VisualStudio.FSharp.UIResources"
             mc:Ignorable="d" 
             d:DesignHeight="600" d:DesignWidth="400">
    <UserControl.Resources>
        <ResourceDictionary>
            <ResourceDictionary.MergedDictionaries>
                <ResourceDictionary Source="OptionPageStyle.xaml"/>
            </ResourceDictionary.MergedDictionaries>
        </ResourceDictionary>
    </UserControl.Resources>
    <Grid>
        <ScrollViewer VerticalScrollBarVisibility="Auto">
            <StackPanel>
                <GroupBox Header="{x:Static local:Strings.Block_Structure}">
                    <CheckBox x:Name="toggleBlockStructure" IsChecked="{Binding IsBlockStructureEnabled}"
                              Content="{x:Static local:Strings.Show_guides}"/>
                </GroupBox>
                <GroupBox Header="{x:Static local:Strings.Outlining}">
                    <CheckBox x:Name="toggleOutlining" IsChecked="{Binding IsOutliningEnabled}"
                              Content="{x:Static local:Strings.Show_Outlining}"/>
                </GroupBox>
                <GroupBox Header="{x:Static local:Strings.Inline_Hints}">
                    <StackPanel>
                        <CheckBox x:Name="toggleInlineTypeHints" IsChecked="{Binding IsInlineTypeHintsEnabled}"
                                  Content="{x:Static local:Strings.Show_Inline_Type_Hints}"/>
                        <CheckBox x:Name="toggleReturnTypeHints" IsChecked="{Binding IsInlineReturnTypeHintsEnabled}"
                                  Content="{x:Static local:Strings.Show_Return_Type_Hints}"/>
                        <CheckBox x:Name="toggleParameterNameHints" IsChecked="{Binding IsInlineParameterNameHintsEnabled}"
                                  Content="{x:Static local:Strings.Show_Inline_Parameter_Name_Hints}"/>
                    </StackPanel>
                </GroupBox>
                <GroupBox Header="{x:Static local:Strings.LiveBuffers}">
                    <CheckBox x:Name="toggleLiveBuffers" IsChecked="{Binding IsUseLiveBuffersEnabled}"
                              Content="{x:Static local:Strings.Enable_Live_Buffers}"/>
                </GroupBox>
                <GroupBox Header="{x:Static local:Strings.TransparentCompiler}">
                    <StackPanel>
                        <Label Content="{x:Static local:Strings.TransparentCompiler_Disclaimer1}" />
                        <Label Content="{x:Static local:Strings.TransparentCompiler_Disclaimer2}" />
                        <Label Content="{x:Static local:Strings.TransparentCompiler_Disclaimer3}" />
                    <CheckBox x:Name="toggleTransparentCompiler" IsChecked="{Binding UseTransparentCompiler}"
                              Content="{x:Static local:Strings.Use_Transparent_Compiler}"/>
                    <CheckBox x:Name="snapshotReuse" IsChecked="{Binding TransparentCompilerSnapshotReuse}"
                              Content="{x:Static local:Strings.Transparent_Compiler_Snapshot_Reuse}"/>
                    </StackPanel>
                </GroupBox>
                <GroupBox Header="{x:Static local:Strings.AdditionalTelemetry}">
                    <CheckBox x:Name="toggleSendAdditionalTelemetry" IsChecked="{Binding SendAdditionalTelemetry}"
                              Content="{x:Static local:Strings.Send_Additional_Telemetry}"/>
                </GroupBox>
                <GroupBox Header="{x:Static local:Strings.Background_analysis}">
                    <CheckBox x:Name="toggleSolutionBackgroundAnalysis" IsChecked="{Binding SolutionBackgroundAnalysis}"
                              Content="{x:Static local:Strings.Analyze_full_solution_on_background}"/>
                </GroupBox>
            </StackPanel>
        </ScrollViewer>
    </Grid>
</UserControl>
