using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace Microsoft.VisualStudio.FSharp.UIResources
{
    /// <summary>
    /// Interaction logic for AdvancedOptionsControl.xaml
    /// </summary>
    public partial class AdvancedOptionsControl : UserControl
    {
        public AdvancedOptionsControl()
        {
            InitializeComponent();
        }
    }
}
