// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.EntityFrameworkCore.TestModels.ComplexNavigationsModel;

#nullable disable

public class ComplexNavigationGlobalization
{
    public string Text { get; set; }
    public ComplexNavigationLanguage Language { get; set; }
}
