// #Regression #NoMT #CompilerOptions 
//<Expects id="FS1064" status="error">Unrecognized platform 'PPC', valid values are 'x86', 'x64', 'Itanium', 'anycpu32bitpreferred', and 'anycpu'</Expects>

exit 1

