// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.EntityFrameworkCore.TestModels.AspNetIdentity;

public class IdentityDbContext<TUser> : IdentityDbContext<TUser, IdentityRole, string>
    where TUser : IdentityUser
{
    public IdentityDbContext(DbContextOptions options)
        : base(options)
    {
    }

    protected IdentityDbContext()
    {
    }
}
