// <auto-generated />
using Microsoft.EntityFrameworkCore;
using Microsoft.EntityFrameworkCore.Infrastructure;
using TestNamespace;

#pragma warning disable 219, 612, 618
#nullable disable

[assembly: DbContextModel(typeof(DbContext), typeof(DbContextModel))]
