// #Regression #Conformance #LexicalAnalysis 
//<Expects id="FS0000" span="(3,1)" status="error">#! may only appear as the first line at the start of a file</Expects>
#!/bin/usr/env fsharpi --exec
exit 0
