// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.EntityFrameworkCore.TestModels.InheritanceModel;

public enum DrinkType
{
    Drink,
    Coke,
    Lilt,
    Tea
}
