// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#nullable disable

namespace Microsoft.Build.Framework.XamlTypes
{
    /// <summary>
    /// The RuleSchema provides a strongly typed identity handle to the underlying schema data model.
    /// </summary>
    public abstract class RuleSchema
    {
    }
}
