// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.EntityFrameworkCore;

#nullable disable

public class SerializationSqlServerTest(F1SqlServerFixture fixture) : SerializationTestBase<F1SqlServerFixture>(fixture);
