// #Regression #Conformance #TypesAndModules #Namespaces 
// Regression test for FSHARP1.0:5354
// Namespace fragment warning should not underlines the entire file in blue

// Second file
let note = "error does not show in the last file, so adding this file"
