// #Regression #Conformance #SignatureFiles #Namespaces 


// Verify an empty file with just hash directives
//<Expects id="FS0988" status="warning">Main module of program is empty: nothing will happen when it is run</Expects>
