// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#nullable disable

namespace Microsoft.Build.BackEnd
{
    /// <summary>
    /// Represents a cache for build properties
    /// </summary>
    internal interface IPropertyCache
    {
    }
}
