<!--Copyright, Microsoft Corporation, All rights reserved.-->
<Rule
  Name="ConfigurationGeneralFile"
  DisplayName="General"
  PageTemplate="generic"
  Description="Project item general properties"
  xmlns="http://schemas.microsoft.com/build/2009/properties">

  <Rule.DataSource>
    <DataSource Persistence="ProjectFile" HasConfigurationCondition="False" ItemType="{}{AnyType}" />
  </Rule.DataSource>

  <Rule.Categories>
    <Category Name="Advanced" DisplayName="Advanced" />
    <Category Name="Misc" DisplayName="Misc" />
  </Rule.Categories>

  <DynamicEnumProperty
    Name="{}{ItemType}"
    DisplayName="Build Action"
    Category="Advanced"
    Description="How the file relates to the build and deployment processes."
    EnumProvider="ItemTypes" />

  <EnumProperty
    Name="CopyToOutputDirectory"
    DisplayName="Copy to Output Directory"
    Category="Advanced"
    Description="Specifies the source file will be copied to the output directory.">
    <EnumValue Name="Never" DisplayName="Do not copy" />
    <EnumValue Name="Always" DisplayName="Copy always" />
    <EnumValue Name="PreserveNewest" DisplayName="Copy if newer" />
  </EnumProperty>

  <StringProperty
    Name="Identity"
    Visible="false"
    ReadOnly="true"
    Category="Misc"
    Description="The item specified in the Include attribute.">
    <StringProperty.DataSource>
      <DataSource Persistence="Intrinsic" ItemType="{}{AnyType}" PersistedName="Identity" />
    </StringProperty.DataSource>
  </StringProperty>

  <StringProperty
    Name="FullPath"
    DisplayName="Full Path"
    ReadOnly="true"
    Category="Misc"
    Description="Location of the file.">
    <StringProperty.DataSource>
      <DataSource Persistence="Intrinsic" ItemType="{}{AnyType}" PersistedName="FullPath" />
    </StringProperty.DataSource>
  </StringProperty>

  <StringProperty
    Name="FileNameAndExtension"
    DisplayName="File Name"
    ReadOnly="true"
    Category="Misc"
    Description="Name of the file or folder.">
    <StringProperty.DataSource>
      <DataSource Persistence="Intrinsic" ItemType="{}{AnyType}" PersistedName="FileNameAndExtension" />
    </StringProperty.DataSource>
  </StringProperty>

  <StringProperty Name="URL" ReadOnly="true" Visible="false">
    <StringProperty.DataSource>
      <DataSource Persistence="Intrinsic" ItemType="{}{AnyType}" PersistedName="FullPath" />
    </StringProperty.DataSource>
  </StringProperty>
  <StringProperty Name="DependentUpon" Visible="False" Description="The leaf name of the file that this item appears as a child to in the project tree." />
  <BoolProperty Name="Visible" Visible="false" />
  <StringProperty Name="Link" Visible="false" />
  <StringProperty Name="Extension" Visible="False" ReadOnly="true">
    <StringProperty.DataSource>
      <DataSource Persistence="Intrinsic" ItemType="{}{AnyType}" PersistedName="Extension" />
    </StringProperty.DataSource>
  </StringProperty>
  <StringProperty Name="GeneratorTarget" Visible="false" />
  <EnumProperty Name="SubType" Visible="false">
    <EnumValue Name="Designer" />
    <EnumValue Name="Component" />
    <EnumValue Name="Control" />
    <EnumValue Name="Form" />
    <EnumValue Name="Code" />
  </EnumProperty>

  <StringProperty Name="LastGenOutput" Visible="false" Description="The filename of the last file generated as a result of the SFG." />
  <BoolProperty Name="DesignTime" Visible="false" Description="A value indicating whether this file has a designer." />
  <BoolProperty Name="AutoGen" Visible="false" Description="A value indicating whether this is a generated file." />
</Rule>