// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.DotNet.SourceBuild.Tests;

public enum DotNetTemplate
{
    Console,
    ClassLib,
    XUnit,
    NUnit,
    MSTest,
    Web,
    Mvc,
    Razor,
    BlazorWasm,
    WebApi,
    WebApp,
    Worker,
}
