// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.
// See the LICENSE file in the project root for more information.

namespace FSharp.Editor.IntegrationTests;

internal static class WellKnownProjectTemplates
{
    public const string FSharpNetCoreClassLibrary = "Microsoft.FSharp.NETCore.ClassLibrary";
    public const string FSharpNetCoreConsoleApplication = "Microsoft.FSharp.NETCore.ConsoleApplication";
    public const string FSharpNetCoreXUnitTest = "Microsoft.FSharp.NETCore.XUnitTest";
}
