// #Regression #OCaml #ReqNOMT 
//<Expects status="error" span="(3,1)" id="FS0062">This construct is for ML compatibility\. Consider using a file with extension '\.ml' or '\.mli' instead\. You can disable this warning by using '--mlcompatibility' or '--nowarn:62'\.$</Expects>
#indent "off"
#q;;
