// A simple C# class with a static method.
// In the F# code, referencing this assembly, we will "hide" C
// and try to access it via 'global'
public class C
{
    public static int M()
    {
        return 1;
    }
}
