namespace BaseClass01
{
    public class BaseClass01
    {
        public virtual int Foo { get; set; }
    }

    public interface Interface01 
    {
        int Foo { get; set; }
    }

    public abstract class AbstractClass01
    {
        public virtual int Foo { get; set; }
    }
}
