// #Conformance #TypesAndModules #Modules 
// Productions
//<Expects status="success"></Expects>
#light

module M6 = begin 
                module M6 = Microsoft.FSharp.Collections.Array
            end
