namespace Foo

type A = bool
type B = sbyte
type C = byte
type D = int16
type E = uint16
type F = int
type G = uint
type H = nativeint
type I = unativeint
type J = int64
type K = single 
type L = float32
type M = float
type N = double
type O = bigint
type P = decimal
type Q = System.Char
type R = System.String
type S = System.Byte
type T = unit