//<Expects status="Error" span="(4,1)" id="FS0998">Package manager key 'unk' was not registered</Expects>
//<Expects status="Error" span="(4,1)" id="FS0999">Processing of a script fragment has stopped because an exception has been raised</Expects>

#r "unk: blubblub"

let x = 1