// <auto-generated />
using System;
using System.Collections.Generic;
using System.Net;
using System.Net.NetworkInformation;
using System.Runtime.CompilerServices;
using Microsoft.EntityFrameworkCore.Scaffolding;

#pragma warning disable 219, 612, 618
#nullable disable

namespace TestNamespace
{
    public static class ManyTypesUnsafeAccessors
    {
        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Id>k__BackingField")]
        public static extern ref CompiledModelTestBase.ManyTypesId Id(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Bool>k__BackingField")]
        public static extern ref bool Bool(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<BoolArray>k__BackingField")]
        public static extern ref bool[] BoolArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<BoolNestedCollection>k__BackingField")]
        public static extern ref bool[][] BoolNestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_boolReadOnlyCollection")]
        public static extern ref List<bool> _boolReadOnlyCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<BoolToStringConverterProperty>k__BackingField")]
        public static extern ref bool BoolToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<BoolToTwoValuesConverterProperty>k__BackingField")]
        public static extern ref bool BoolToTwoValuesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<BoolToZeroOneConverterProperty>k__BackingField")]
        public static extern ref bool BoolToZeroOneConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Bytes>k__BackingField")]
        public static extern ref byte[] Bytes(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<BytesToStringConverterProperty>k__BackingField")]
        public static extern ref byte[] BytesToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<CastingConverterProperty>k__BackingField")]
        public static extern ref int CastingConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Char>k__BackingField")]
        public static extern ref char Char(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<CharArray>k__BackingField")]
        public static extern ref char[] CharArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<CharNestedCollection>k__BackingField")]
        public static extern ref char[][] CharNestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<CharToStringConverterProperty>k__BackingField")]
        public static extern ref char CharToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateOnly>k__BackingField")]
        public static extern ref DateOnly DateOnly(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateOnlyToStringConverterProperty>k__BackingField")]
        public static extern ref DateOnly DateOnlyToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTime>k__BackingField")]
        public static extern ref DateTime DateTime(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTimeOffsetToBinaryConverterProperty>k__BackingField")]
        public static extern ref DateTimeOffset DateTimeOffsetToBinaryConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTimeOffsetToBytesConverterProperty>k__BackingField")]
        public static extern ref DateTimeOffset DateTimeOffsetToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTimeOffsetToStringConverterProperty>k__BackingField")]
        public static extern ref DateTimeOffset DateTimeOffsetToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTimeToBinaryConverterProperty>k__BackingField")]
        public static extern ref DateTime DateTimeToBinaryConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTimeToStringConverterProperty>k__BackingField")]
        public static extern ref DateTime DateTimeToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DateTimeToTicksConverterProperty>k__BackingField")]
        public static extern ref DateTime DateTimeToTicksConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Decimal>k__BackingField")]
        public static extern ref decimal Decimal(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DecimalArray>k__BackingField")]
        public static extern ref decimal[] DecimalArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DecimalNumberToBytesConverterProperty>k__BackingField")]
        public static extern ref decimal DecimalNumberToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DecimalNumberToStringConverterProperty>k__BackingField")]
        public static extern ref decimal DecimalNumberToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Double>k__BackingField")]
        public static extern ref double Double(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DoubleArray>k__BackingField")]
        public static extern ref double[] DoubleArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DoubleNumberToBytesConverterProperty>k__BackingField")]
        public static extern ref double DoubleNumberToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<DoubleNumberToStringConverterProperty>k__BackingField")]
        public static extern ref double DoubleNumberToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum16>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum16 Enum16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum16AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum16 Enum16AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum32>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum32 Enum32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum32AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum32 Enum32AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum64>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum64 Enum64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum64AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum64 Enum64AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum8>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum8 Enum8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Enum8AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum8 Enum8AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumToNumberConverterProperty>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum32 EnumToNumberConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumToStringConverterProperty>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum32 EnumToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU16>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU16 EnumU16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU16AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU16 EnumU16AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU32>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU32 EnumU32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU32AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU32 EnumU32AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU64>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU64 EnumU64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU64AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU64 EnumU64AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU8>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU8 EnumU8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<EnumU8AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU8 EnumU8AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Float>k__BackingField")]
        public static extern ref float Float(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<FloatArray>k__BackingField")]
        public static extern ref float[] FloatArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Guid>k__BackingField")]
        public static extern ref Guid Guid(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<GuidToBytesConverterProperty>k__BackingField")]
        public static extern ref Guid GuidToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<GuidToStringConverterProperty>k__BackingField")]
        public static extern ref Guid GuidToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<IPAddress>k__BackingField")]
        public static extern ref IPAddress IPAddress(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<IPAddressToBytesConverterProperty>k__BackingField")]
        public static extern ref IPAddress IPAddressToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<IPAddressToStringConverterProperty>k__BackingField")]
        public static extern ref IPAddress IPAddressToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int16>k__BackingField")]
        public static extern ref short Int16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int16Array>k__BackingField")]
        public static extern ref short[] Int16Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int32>k__BackingField")]
        public static extern ref int Int32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int32Array>k__BackingField")]
        public static extern ref int[] Int32Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int32NestedCollection>k__BackingField")]
        public static extern ref int[][] Int32NestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_int32ReadOnlyCollection")]
        public static extern ref List<int> _int32ReadOnlyCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int64>k__BackingField")]
        public static extern ref long Int64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int64Array>k__BackingField")]
        public static extern ref long[] Int64Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int64NestedCollection>k__BackingField")]
        public static extern ref IList<long[]>[] Int64NestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int8>k__BackingField")]
        public static extern ref sbyte Int8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int8Array>k__BackingField")]
        public static extern ref sbyte[] Int8Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Int8NestedCollection>k__BackingField")]
        public static extern ref sbyte[][][] Int8NestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<IntNumberToBytesConverterProperty>k__BackingField")]
        public static extern ref int IntNumberToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<IntNumberToStringConverterProperty>k__BackingField")]
        public static extern ref int IntNumberToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullIntToNullStringConverterProperty>k__BackingField")]
        public static extern ref int? NullIntToNullStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableBool>k__BackingField")]
        public static extern ref bool? NullableBool(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableBoolArray>k__BackingField")]
        public static extern ref bool?[] NullableBoolArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableBytes>k__BackingField")]
        public static extern ref byte[] NullableBytes(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableChar>k__BackingField")]
        public static extern ref char? NullableChar(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableCharArray>k__BackingField")]
        public static extern ref char?[] NullableCharArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableDateOnly>k__BackingField")]
        public static extern ref DateOnly? NullableDateOnly(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableDateTime>k__BackingField")]
        public static extern ref DateTime? NullableDateTime(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableDecimal>k__BackingField")]
        public static extern ref decimal? NullableDecimal(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableDecimalArray>k__BackingField")]
        public static extern ref decimal?[] NullableDecimalArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableDouble>k__BackingField")]
        public static extern ref double? NullableDouble(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableDoubleArray>k__BackingField")]
        public static extern ref double?[] NullableDoubleArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum16>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum16? NullableEnum16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum16AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum16? NullableEnum16AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum32>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum32? NullableEnum32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum32AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum32? NullableEnum32AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum64>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum64? NullableEnum64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum64AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum64? NullableEnum64AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum8>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum8? NullableEnum8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnum8AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.Enum8? NullableEnum8AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU16>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU16? NullableEnumU16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU16AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU16? NullableEnumU16AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU32>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU32? NullableEnumU32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU32AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU32? NullableEnumU32AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU64>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU64? NullableEnumU64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU64AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU64? NullableEnumU64AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU8>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU8? NullableEnumU8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableEnumU8AsString>k__BackingField")]
        public static extern ref CompiledModelTestBase.EnumU8? NullableEnumU8AsString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableFloat>k__BackingField")]
        public static extern ref float? NullableFloat(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableFloatArray>k__BackingField")]
        public static extern ref float?[] NullableFloatArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableGuid>k__BackingField")]
        public static extern ref Guid? NullableGuid(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableIPAddress>k__BackingField")]
        public static extern ref IPAddress NullableIPAddress(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt16>k__BackingField")]
        public static extern ref short? NullableInt16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt16Array>k__BackingField")]
        public static extern ref short?[] NullableInt16Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt32>k__BackingField")]
        public static extern ref int? NullableInt32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt32Array>k__BackingField")]
        public static extern ref int?[] NullableInt32Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt32NestedCollection>k__BackingField")]
        public static extern ref int?[][] NullableInt32NestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt64>k__BackingField")]
        public static extern ref long? NullableInt64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt64Array>k__BackingField")]
        public static extern ref long?[] NullableInt64Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt64NestedCollection>k__BackingField")]
        public static extern ref List<long?[][]> NullableInt64NestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt8>k__BackingField")]
        public static extern ref sbyte? NullableInt8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableInt8Array>k__BackingField")]
        public static extern ref sbyte?[] NullableInt8Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullablePhysicalAddress>k__BackingField")]
        public static extern ref PhysicalAddress NullablePhysicalAddress(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableString>k__BackingField")]
        public static extern ref string NullableString(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableStringArray>k__BackingField")]
        public static extern ref string[] NullableStringArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableStringNestedCollection>k__BackingField")]
        public static extern ref string[][] NullableStringNestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableTimeOnly>k__BackingField")]
        public static extern ref TimeOnly? NullableTimeOnly(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableTimeSpan>k__BackingField")]
        public static extern ref TimeSpan? NullableTimeSpan(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt16>k__BackingField")]
        public static extern ref ushort? NullableUInt16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt16Array>k__BackingField")]
        public static extern ref ushort?[] NullableUInt16Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt32>k__BackingField")]
        public static extern ref uint? NullableUInt32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt32Array>k__BackingField")]
        public static extern ref uint?[] NullableUInt32Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt64>k__BackingField")]
        public static extern ref ulong? NullableUInt64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt64Array>k__BackingField")]
        public static extern ref ulong?[] NullableUInt64Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt8>k__BackingField")]
        public static extern ref byte? NullableUInt8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUInt8Array>k__BackingField")]
        public static extern ref byte?[] NullableUInt8Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<NullableUri>k__BackingField")]
        public static extern ref Uri NullableUri(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<PhysicalAddress>k__BackingField")]
        public static extern ref PhysicalAddress PhysicalAddress(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<PhysicalAddressToBytesConverterProperty>k__BackingField")]
        public static extern ref PhysicalAddress PhysicalAddressToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<PhysicalAddressToStringConverterProperty>k__BackingField")]
        public static extern ref PhysicalAddress PhysicalAddressToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<String>k__BackingField")]
        public static extern ref string String(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringArray>k__BackingField")]
        public static extern ref string[] StringArray(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringNestedCollection>k__BackingField")]
        public static extern ref string[][] StringNestedCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_stringReadOnlyCollection")]
        public static extern ref List<string> _stringReadOnlyCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToBoolConverterProperty>k__BackingField")]
        public static extern ref string StringToBoolConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToBytesConverterProperty>k__BackingField")]
        public static extern ref string StringToBytesConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToCharConverterProperty>k__BackingField")]
        public static extern ref string StringToCharConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToDateOnlyConverterProperty>k__BackingField")]
        public static extern ref string StringToDateOnlyConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToDateTimeConverterProperty>k__BackingField")]
        public static extern ref string StringToDateTimeConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToDateTimeOffsetConverterProperty>k__BackingField")]
        public static extern ref string StringToDateTimeOffsetConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToDecimalNumberConverterProperty>k__BackingField")]
        public static extern ref string StringToDecimalNumberConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToDoubleNumberConverterProperty>k__BackingField")]
        public static extern ref string StringToDoubleNumberConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToEnumConverterProperty>k__BackingField")]
        public static extern ref string StringToEnumConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToGuidConverterProperty>k__BackingField")]
        public static extern ref string StringToGuidConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToIntNumberConverterProperty>k__BackingField")]
        public static extern ref string StringToIntNumberConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToTimeOnlyConverterProperty>k__BackingField")]
        public static extern ref string StringToTimeOnlyConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToTimeSpanConverterProperty>k__BackingField")]
        public static extern ref string StringToTimeSpanConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<StringToUriConverterProperty>k__BackingField")]
        public static extern ref string StringToUriConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<TimeOnly>k__BackingField")]
        public static extern ref TimeOnly TimeOnly(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<TimeOnlyToStringConverterProperty>k__BackingField")]
        public static extern ref TimeOnly TimeOnlyToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<TimeOnlyToTicksConverterProperty>k__BackingField")]
        public static extern ref TimeOnly TimeOnlyToTicksConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<TimeSpan>k__BackingField")]
        public static extern ref TimeSpan TimeSpan(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<TimeSpanToStringConverterProperty>k__BackingField")]
        public static extern ref TimeSpan TimeSpanToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<TimeSpanToTicksConverterProperty>k__BackingField")]
        public static extern ref TimeSpan TimeSpanToTicksConverterProperty(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt16>k__BackingField")]
        public static extern ref ushort UInt16(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt16Array>k__BackingField")]
        public static extern ref ushort[] UInt16Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt32>k__BackingField")]
        public static extern ref uint UInt32(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt32Array>k__BackingField")]
        public static extern ref uint[] UInt32Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt64>k__BackingField")]
        public static extern ref ulong UInt64(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt64Array>k__BackingField")]
        public static extern ref ulong[] UInt64Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt8>k__BackingField")]
        public static extern ref byte UInt8(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UInt8Array>k__BackingField")]
        public static extern ref byte[] UInt8Array(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_uInt8ReadOnlyCollection")]
        public static extern ref List<byte> _uInt8ReadOnlyCollection(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Uri>k__BackingField")]
        public static extern ref Uri Uri(CompiledModelTestBase.ManyTypes @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<UriToStringConverterProperty>k__BackingField")]
        public static extern ref Uri UriToStringConverterProperty(CompiledModelTestBase.ManyTypes @this);
    }
}
