module TestData

let rnd = new System.Random()

let expectedValues() = 
    let x = 1
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    let x = x + rnd.Next(3) 
    x
printfn "expectedValues() = %A" (expectedValues())
printf "TEST PASSED OK" ;
