// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.EntityFrameworkCore.Storage;

/// <summary>
///     A type representing CLR type of the JsonTypeMapping.
/// </summary>
public sealed class JsonTypePlaceholder
{
    private JsonTypePlaceholder()
    {
    }
}
