// #Conformance #LexicalAnalysis 
                               









                                  






                          let                                      



                                     square  
                                                            x                        

                                                                         =   x * x



                            in
         
                 printf
                                            "%d"                                                   (




          square








                                                  42

)                              
