// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

//using Microsoft.EntityFrameworkCore.NativeAotTests.Models;

//using var context = new NativeAotContext();

//context.Add(new User { Name = "Andriy" });
//await context.SaveChangesAsync();

return 100;
