// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

using Microsoft.Build.BackEnd;

#nullable disable

namespace Microsoft.Build.Shared
{
    internal interface IElementLocation : IMSBuildElementLocation, ITranslatable { }
}
