// #Regression #Conformance #TypesAndModules #Exceptions 
// Exception types
// Exception labels must begin with an uppercase letter



#light

exception ı of int     // err 
exception i of int     // err
