// #Conformance #ObjectOrientedTypes #TypeExtensions 
#light
namespace NS
  module M = 
    type Lib with
    
    // members: property, method, abstract, implementing dispatch slots, overloaded. 
          override x.M i = i
    
  
