// #Regression #Diagnostics 
// Regression test for FSHARP1.0:4245
//<Expects status="notin">NONTERM</Expects>
//<Expects id="FS0010" span="(6,1-6,2)" status="error">Unexpected character '.' in implementation file</Expects>

\;;
