// #Regression #OCaml 
//<Expects status="error" id="FS0062">This construct is for ML compatibility\. The file extensions '\.ml' and '\.mli' are for ML compatibility\. You can disable this warning by using '--mlcompatibility' or '--nowarn:62'\.$</Expects>
exit 0
