// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.EntityFrameworkCore.TestUtilities;

public class TestModificationCommandBatch(
    ModificationCommandBatchFactoryDependencies dependencies,
    int? maxBatchSize) : SingularModificationCommandBatch(dependencies)
{
    protected override int MaxBatchSize { get; } = maxBatchSize ?? 42;
}
