// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace Microsoft.Build.BackEnd
{
    internal readonly record struct InternPathIds(int DirectoryId, int FileNameId);
}
