// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

// All tests in OwnedNavigationsProjectionSqliteTest currently fail because of #26708
// (Stop generating composite keys for owned collections on SQLite)


