// #Regression #Conformance #SignatureFiles #Namespaces 
// Normally this would produce a warning, but
// since this compiles an EXE one shouldn't be emitted.

//<Expects id="FS0191" status="warning">Files in libraries or multiple-file applications should begin with a namespace or module declaration</Expects>

exit 0
