// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

using Microsoft.EntityFrameworkCore.Benchmarks.Models.AdventureWorks;

namespace Microsoft.EntityFrameworkCore.Benchmarks.Query;

public class NavigationsQuerySqlServerTests : NavigationsQueryTests
{
    protected override AdventureWorksContextBase CreateContext()
        => AdventureWorksSqlServerFixture.CreateContext();
}
