// <auto-generated />
using System;
using System.Runtime.CompilerServices;
using Microsoft.EntityFrameworkCore.Scaffolding;

#pragma warning disable 219, 612, 618
#nullable disable

namespace TestNamespace
{
    public static class DependentDerivedUnsafeAccessors<TKey>
    {
        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Data>k__BackingField")]
        public static extern ref string Data(CompiledModelTestBase.DependentDerived<TKey> @this);
    }
}
