/// <summary>
/// A C# class with virtual and abstract methods
/// Consumed by E_CallToUnimplementedMethod02.fs
/// </summary>
public abstract class B
{
    public virtual double M(int i)
    {
        return 1.0;
    }

    public abstract void M(string i);
}
