// #Regression #NoMT #FSI #RequiresENU 
// Regression test for FSharp1.0:4164 - FSI emit "please report to fsbugs" error on malformed usage of .. operator
//<Expects status="success"> Timing now on</Expects>
//<Expects status="success"> Timing now off</Expects>

#time "on"
#time "off"
exit 0
