// <auto-generated />
using System;
using System.Collections.Generic;
using System.Net;
using System.Runtime.CompilerServices;
using Microsoft.EntityFrameworkCore.Scaffolding;

#pragma warning disable 219, 612, 618
#nullable disable

namespace TestNamespace
{
    public static class OwnedTypeUnsafeAccessors
    {
        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_details")]
        public static extern ref string _details(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Number>k__BackingField")]
        public static extern ref int Number(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_refTypeArray")]
        public static extern ref IPAddress[] _refTypeArray(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_refTypeEnumerable")]
        public static extern ref IEnumerable<string> _refTypeEnumerable(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_refTypeIList")]
        public static extern ref IList<string> _refTypeIList(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_refTypeList")]
        public static extern ref List<IPAddress> _refTypeList(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_valueTypeArray")]
        public static extern ref DateTime[] _valueTypeArray(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_valueTypeEnumerable")]
        public static extern ref IEnumerable<byte> _valueTypeEnumerable(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<ValueTypeIList>k__BackingField")]
        public static extern ref IList<byte> ValueTypeIList(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "_valueTypeList")]
        public static extern ref List<short> _valueTypeList(CompiledModelTestBase.OwnedType @this);

        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "<Principal>k__BackingField")]
        public static extern ref CompiledModelTestBase.PrincipalBase Principal(CompiledModelTestBase.OwnedType @this);
    }
}
