// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

using Microsoft.EntityFrameworkCore.Cosmos.Metadata.Internal;

namespace Microsoft.EntityFrameworkCore.Cosmos.Storage.Internal;

/// <summary>
///     This is an internal API that supports the Entity Framework Core infrastructure and not subject to
///     the same compatibility standards as public APIs. It may be changed or removed without notice in
///     any release. You should only use it directly in your code with extreme caution and knowing that
///     doing so can result in application failures when updating to a new Entity Framework Core release.
/// </summary>
public readonly record struct ContainerProperties(
    string Id,
    IReadOnlyList<string> PartitionKeyStoreNames,
    int? AnalyticalStoreTimeToLiveInSeconds,
    int? DefaultTimeToLive,
    ThroughputProperties? Throughput,
    IReadOnlyList<IIndex> Indexes,
    IReadOnlyList<(IProperty Property, CosmosVectorType VectorType)> Vectors,
    string DefaultFullTextLanguage,
    IReadOnlyList<(IProperty Property, string? Language)> FullTextProperties);
