// <auto-generated />
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using Microsoft.EntityFrameworkCore.Scaffolding;

#pragma warning disable 219, 612, 618
#nullable disable

namespace TestNamespace
{
    public static class PrincipalDerivedUnsafeAccessors<TDependent>
        where TDependent : class
    {
        [UnsafeAccessor(UnsafeAccessorKind.Field, Name = "ManyOwned")]
        public static extern ref IList<CompiledModelTestBase.OwnedType> ManyOwned(CompiledModelTestBase.PrincipalDerived<TDependent> @this);
    }
}
