//------------------------------------------------------------------------------
// <auto-generated>
//     此代码由工具生成。
//     运行时版本:4.0.30319.42000
//
//     对此文件的更改可能会导致不正确的行为，并且如果
//     重新生成代码，这些更改将会丢失。
// </auto-generated>
//------------------------------------------------------------------------------

namespace Microsoft.VisualStudio.FSharp.UIResources {
    using System;
    
    
    /// <summary>
    ///   一个强类型的资源类，用于查找本地化的字符串等。
    /// </summary>
    // 此类是由 StronglyTypedResourceBuilder
    // 类通过类似于 ResGen 或 Visual Studio 的工具自动生成的。
    // 若要添加或移除成员，请编辑 .ResX 文件，然后重新运行 ResGen
    // (以 /str 作为命令选项)，或重新生成 VS 项目。
    [global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "17.0.0.0")]
    [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
    [global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]
    public class Strings {
        
        private static global::System.Resources.ResourceManager resourceMan;
        
        private static global::System.Globalization.CultureInfo resourceCulture;
        
        [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        internal Strings() {
        }
        
        /// <summary>
        ///   返回此类使用的缓存的 ResourceManager 实例。
        /// </summary>
        [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
        public static global::System.Resources.ResourceManager ResourceManager {
            get {
                if (object.ReferenceEquals(resourceMan, null)) {
                    global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.VisualStudio.FSharp.UIResources.Strings", typeof(Strings).Assembly);
                    resourceMan = temp;
                }
                return resourceMan;
            }
        }
        
        /// <summary>
        ///   重写当前线程的 CurrentUICulture 属性，对
        ///   使用此强类型资源类的所有资源查找执行重写。
        /// </summary>
        [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
        public static global::System.Globalization.CultureInfo Culture {
            get {
                return resourceCulture;
            }
            set {
                resourceCulture = value;
            }
        }
        
        /// <summary>
        ///   查找类似 Additional performance telemetry (experimental) 的本地化字符串。
        /// </summary>
        public static string AdditionalTelemetry {
            get {
                return ResourceManager.GetString("AdditionalTelemetry", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Always place open statements at the top level 的本地化字符串。
        /// </summary>
        public static string Always_place_opens_at_top_level {
            get {
                return ResourceManager.GetString("Always_place_opens_at_top_level", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Keep analyzing the entire solution for diagnostics as a low priority background task (requires restart) 的本地化字符串。
        /// </summary>
        public static string Analyze_full_solution_on_background {
            get {
                return ResourceManager.GetString("Analyze_full_solution_on_background", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Background analysis 的本地化字符串。
        /// </summary>
        public static string Background_analysis {
            get {
                return ResourceManager.GetString("Background_analysis", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Block Structure Guides 的本地化字符串。
        /// </summary>
        public static string Block_Structure {
            get {
                return ResourceManager.GetString("Block_Structure", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Code Fixes 的本地化字符串。
        /// </summary>
        public static string Code_Fixes {
            get {
                return ResourceManager.GetString("Code_Fixes", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Completion Lists 的本地化字符串。
        /// </summary>
        public static string Completion_Lists {
            get {
                return ResourceManager.GetString("Completion_Lists", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 D_ash underline 的本地化字符串。
        /// </summary>
        public static string Dash_underline {
            get {
                return ResourceManager.GetString("Dash_underline", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Diagnostics 的本地化字符串。
        /// </summary>
        public static string Diagnostics {
            get {
                return ResourceManager.GetString("Diagnostics", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 D_ot underline 的本地化字符串。
        /// </summary>
        public static string Dot_underline {
            get {
                return ResourceManager.GetString("Dot_underline", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Keep background symbol keys 的本地化字符串。
        /// </summary>
        public static string Enable_Background_ItemKeyStore_And_Semantic_Classification {
            get {
                return ResourceManager.GetString("Enable_Background_ItemKeyStore_And_Semantic_Classification", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Enable fast find references &amp; rename (experimental) 的本地化字符串。
        /// </summary>
        public static string Enable_Fast_Find_References {
            get {
                return ResourceManager.GetString("Enable_Fast_Find_References", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 _Enable in-memory cross project references 的本地化字符串。
        /// </summary>
        public static string Enable_in_memory_cross_project_references {
            get {
                return ResourceManager.GetString("Enable_in_memory_cross_project_references", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Use live (unsaved) buffers for analysis (restart required) 的本地化字符串。
        /// </summary>
        public static string Enable_Live_Buffers {
            get {
                return ResourceManager.GetString("Enable_Live_Buffers", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Enable parallel reference resolution 的本地化字符串。
        /// </summary>
        public static string Enable_Parallel_Reference_Resolution {
            get {
                return ResourceManager.GetString("Enable_Parallel_Reference_Resolution", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Enable partial type checking 的本地化字符串。
        /// </summary>
        public static string Enable_partial_type_checking {
            get {
                return ResourceManager.GetString("Enable_partial_type_checking", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Enable stale data for IntelliSense features 的本地化字符串。
        /// </summary>
        public static string Enable_Stale_IntelliSense_Results {
            get {
                return ResourceManager.GetString("Enable_Stale_IntelliSense_Results", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Always add new line on enter 的本地化字符串。
        /// </summary>
        public static string Enter_key_always {
            get {
                return ResourceManager.GetString("Enter_key_always", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Never add new line on enter 的本地化字符串。
        /// </summary>
        public static string Enter_key_never {
            get {
                return ResourceManager.GetString("Enter_key_never", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Only add new line on enter after end of fully typed word 的本地化字符串。
        /// </summary>
        public static string Enter_key_only {
            get {
                return ResourceManager.GetString("Enter_key_only", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Enter key behavior 的本地化字符串。
        /// </summary>
        public static string Enter_Key_Rule {
            get {
                return ResourceManager.GetString("Enter_Key_Rule", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Find References Performance Options 的本地化字符串。
        /// </summary>
        public static string Find_References_Performance {
            get {
                return ResourceManager.GetString("Find_References_Performance", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Re-format indentation on paste (Experimental) 的本地化字符串。
        /// </summary>
        public static string Format_on_paste {
            get {
                return ResourceManager.GetString("Format_on_paste", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Formatting 的本地化字符串。
        /// </summary>
        public static string Formatting {
            get {
                return ResourceManager.GetString("Formatting", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Generate default implementation body for overrided method 的本地化字符串。
        /// </summary>
        public static string Generate_Body_For_Overridden_Method {
            get {
                return ResourceManager.GetString("Generate_Body_For_Overridden_Method", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Inline Hints 的本地化字符串。
        /// </summary>
        public static string Inline_Hints {
            get {
                return ResourceManager.GetString("Inline_Hints", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 IntelliSense Performance Options 的本地化字符串。
        /// </summary>
        public static string IntelliSense_Performance {
            get {
                return ResourceManager.GetString("IntelliSense_Performance", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Keep all background intermediate resolutions (increases memory usage) 的本地化字符串。
        /// </summary>
        public static string Keep_All_Background_Resolutions {
            get {
                return ResourceManager.GetString("Keep_All_Background_Resolutions", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Keep all background symbol uses (increases memory usage) 的本地化字符串。
        /// </summary>
        public static string Keep_All_Background_Symbol_Uses {
            get {
                return ResourceManager.GetString("Keep_All_Background_Symbol_Uses", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Performance 的本地化字符串。
        /// </summary>
        public static string Language_Service_Performance {
            get {
                return ResourceManager.GetString("Language_Service_Performance", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Language service settings (advanced) 的本地化字符串。
        /// </summary>
        public static string Language_Service_Settings {
            get {
                return ResourceManager.GetString("Language_Service_Settings", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Live Buffers 的本地化字符串。
        /// </summary>
        public static string LiveBuffers {
            get {
                return ResourceManager.GetString("LiveBuffers", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Navigation links 的本地化字符串。
        /// </summary>
        public static string Navigation_links {
            get {
                return ResourceManager.GetString("Navigation_links", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Outlining 的本地化字符串。
        /// </summary>
        public static string Outlining {
            get {
                return ResourceManager.GetString("Outlining", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Parallelization (requires restart) 的本地化字符串。
        /// </summary>
        public static string Parallelization {
            get {
                return ResourceManager.GetString("Parallelization", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Preferred description width in characters 的本地化字符串。
        /// </summary>
        public static string Preferred_description_width_in_characters {
            get {
                return ResourceManager.GetString("Preferred_description_width_in_characters", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 F# Project and Caching Performance Options 的本地化字符串。
        /// </summary>
        public static string Project_Performance {
            get {
                return ResourceManager.GetString("Project_Performance", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Remove unnecessary parentheses (experimental, might affect typing performance) 的本地化字符串。
        /// </summary>
        public static string Remove_parens_code_fix {
            get {
                return ResourceManager.GetString("Remove_parens_code_fix", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Send additional performance telemetry 的本地化字符串。
        /// </summary>
        public static string Send_Additional_Telemetry {
            get {
                return ResourceManager.GetString("Send_Additional_Telemetry", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Show s_ymbols in unopened namespaces 的本地化字符串。
        /// </summary>
        public static string Show_all_symbols {
            get {
                return ResourceManager.GetString("Show_all_symbols", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Show completion list after a character is _deleted 的本地化字符串。
        /// </summary>
        public static string Show_completion_list_after_a_character_is_deleted {
            get {
                return ResourceManager.GetString("Show_completion_list_after_a_character_is_deleted", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 _Show completion list after a character is typed 的本地化字符串。
        /// </summary>
        public static string Show_completion_list_after_a_character_is_typed {
            get {
                return ResourceManager.GetString("Show_completion_list_after_a_character_is_typed", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Show structure guidelines for F# code 的本地化字符串。
        /// </summary>
        public static string Show_guides {
            get {
                return ResourceManager.GetString("Show_guides", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Display inline parameter name hints (preview) 的本地化字符串。
        /// </summary>
        public static string Show_Inline_Parameter_Name_Hints {
            get {
                return ResourceManager.GetString("Show_Inline_Parameter_Name_Hints", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Display inline type hints (preview) 的本地化字符串。
        /// </summary>
        public static string Show_Inline_Type_Hints {
            get {
                return ResourceManager.GetString("Show_Inline_Type_Hints", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 S_how navigation links as 的本地化字符串。
        /// </summary>
        public static string Show_navigation_links_as {
            get {
                return ResourceManager.GetString("Show_navigation_links_as", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Show outlining and collapsible nodes for F# code 的本地化字符串。
        /// </summary>
        public static string Show_Outlining {
            get {
                return ResourceManager.GetString("Show_Outlining", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Show remarks in Quick Info 的本地化字符串。
        /// </summary>
        public static string Show_remarks_in_Quick_Info {
            get {
                return ResourceManager.GetString("Show_remarks_in_Quick_Info", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Display return type hints (preview) 的本地化字符串。
        /// </summary>
        public static string Show_Return_Type_Hints {
            get {
                return ResourceManager.GetString("Show_Return_Type_Hints", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Simplify names (remove unnecessary qualifiers) 的本地化字符串。
        /// </summary>
        public static string Simplify_name_code_fix {
            get {
                return ResourceManager.GetString("Simplify_name_code_fix", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 _Solid underline 的本地化字符串。
        /// </summary>
        public static string Solid_underline {
            get {
                return ResourceManager.GetString("Solid_underline", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Suggest names for unresolved identifiers 的本地化字符串。
        /// </summary>
        public static string Suggest_names_for_errors_code_fix {
            get {
                return ResourceManager.GetString("Suggest_names_for_errors_code_fix", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Text hover 的本地化字符串。
        /// </summary>
        public static string Text_hover {
            get {
                return ResourceManager.GetString("Text_hover", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Time until stale results are used (in milliseconds) 的本地化字符串。
        /// </summary>
        public static string Time_until_stale_completion {
            get {
                return ResourceManager.GetString("Time_until_stale_completion", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 In-memory cross-project references store project-level data in memory to allow IDE features to work across projects. 的本地化字符串。
        /// </summary>
        public static string Tooltip_in_memory_cross_project_references {
            get {
                return ResourceManager.GetString("Tooltip_in_memory_cross_project_references", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Format signature to the given width by adding line breaks conforming with F# syntax rules.  的本地化字符串。
        /// </summary>
        public static string Tooltip_preferred_description_width_in_characters {
            get {
                return ResourceManager.GetString("Tooltip_preferred_description_width_in_characters", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Transparent Compiler Cache Factor 的本地化字符串。
        /// </summary>
        public static string Transparent_Compiler_Cache_Factor {
            get {
                return ResourceManager.GetString("Transparent_Compiler_Cache_Factor", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Higher number means more memory will be used for caching. Changing the value wipes cache. 的本地化字符串。
        /// </summary>
        public static string Transparent_Compiler_Cache_Factor_Tooltip {
            get {
                return ResourceManager.GetString("Transparent_Compiler_Cache_Factor_Tooltip", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Create new project snapshots from existing ones 的本地化字符串。
        /// </summary>
        public static string Transparent_Compiler_Snapshot_Reuse {
            get {
                return ResourceManager.GetString("Transparent_Compiler_Snapshot_Reuse", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Transparent Compiler (experimental) 的本地化字符串。
        /// </summary>
        public static string TransparentCompiler {
            get {
                return ResourceManager.GetString("TransparentCompiler", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 WARNING! Transparent Compiler does not yet support all features and can cause crashes or give incorrect results. 的本地化字符串。
        /// </summary>
        public static string TransparentCompiler_Disclaimer1 {
            get {
                return ResourceManager.GetString("TransparentCompiler_Disclaimer1", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Use at your own risk! 的本地化字符串。
        /// </summary>
        public static string TransparentCompiler_Disclaimer2 {
            get {
                return ResourceManager.GetString("TransparentCompiler_Disclaimer2", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 By checking this you also opt-in for additional performance telemetry 的本地化字符串。
        /// </summary>
        public static string TransparentCompiler_Disclaimer3 {
            get {
                return ResourceManager.GetString("TransparentCompiler_Disclaimer3", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Analyze and suggest fixes for unused values 的本地化字符串。
        /// </summary>
        public static string Unused_declaration_code_fix {
            get {
                return ResourceManager.GetString("Unused_declaration_code_fix", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Remove unused open statements 的本地化字符串。
        /// </summary>
        public static string Unused_opens_code_fix {
            get {
                return ResourceManager.GetString("Unused_opens_code_fix", resourceCulture);
            }
        }
        
        /// <summary>
        ///   查找类似 Use Transparent Compiler (restart required) 的本地化字符串。
        /// </summary>
        public static string Use_Transparent_Compiler {
            get {
                return ResourceManager.GetString("Use_Transparent_Compiler", resourceCulture);
            }
        }
    }
}
