// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

namespace InvalidCustomCheck
{
    public sealed class InvalidCheck
    {
        public string FriendlyName => "InvalidCheck";
    }
}
