// Copyright (c) Microsoft Corporation.  All Rights Reserved.  See License.txt in the project root for license information.

namespace Microsoft.FSharp

open System.Reflection
open System.Runtime.CompilerServices
open System.Runtime.InteropServices
open Microsoft.VisualStudio.Shell

[<assembly: AssemblyCulture("")>]

do()
