// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

// <auto-generated />

namespace Microsoft.EntityFrameworkCore.ModelBuilding;

public static class GiantModel
{
    public class RelatedEntity1
    {
        public int Id { get; set; }

        public RelatedEntity2000 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity2> ChildEntities { get; set; }
    }

    public class RelatedEntity2
    {
        public int Id { get; set; }

        public RelatedEntity1 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity3> ChildEntities { get; set; }
    }

    public class RelatedEntity3
    {
        public int Id { get; set; }

        public RelatedEntity2 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity4> ChildEntities { get; set; }
    }

    public class RelatedEntity4
    {
        public int Id { get; set; }

        public RelatedEntity3 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity5> ChildEntities { get; set; }
    }

    public class RelatedEntity5
    {
        public int Id { get; set; }

        public RelatedEntity4 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity6> ChildEntities { get; set; }
    }

    public class RelatedEntity6
    {
        public int Id { get; set; }

        public RelatedEntity5 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity7> ChildEntities { get; set; }
    }

    public class RelatedEntity7
    {
        public int Id { get; set; }

        public RelatedEntity6 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity8> ChildEntities { get; set; }
    }

    public class RelatedEntity8
    {
        public int Id { get; set; }

        public RelatedEntity7 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity9> ChildEntities { get; set; }
    }

    public class RelatedEntity9
    {
        public int Id { get; set; }

        public RelatedEntity8 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity10> ChildEntities { get; set; }
    }

    public class RelatedEntity10
    {
        public int Id { get; set; }

        public RelatedEntity9 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity11> ChildEntities { get; set; }
    }

    public class RelatedEntity11
    {
        public int Id { get; set; }

        public RelatedEntity10 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity12> ChildEntities { get; set; }
    }

    public class RelatedEntity12
    {
        public int Id { get; set; }

        public RelatedEntity11 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity13> ChildEntities { get; set; }
    }

    public class RelatedEntity13
    {
        public int Id { get; set; }

        public RelatedEntity12 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity14> ChildEntities { get; set; }
    }

    public class RelatedEntity14
    {
        public int Id { get; set; }

        public RelatedEntity13 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity15> ChildEntities { get; set; }
    }

    public class RelatedEntity15
    {
        public int Id { get; set; }

        public RelatedEntity14 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity16> ChildEntities { get; set; }
    }

    public class RelatedEntity16
    {
        public int Id { get; set; }

        public RelatedEntity15 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity17> ChildEntities { get; set; }
    }

    public class RelatedEntity17
    {
        public int Id { get; set; }

        public RelatedEntity16 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity18> ChildEntities { get; set; }
    }

    public class RelatedEntity18
    {
        public int Id { get; set; }

        public RelatedEntity17 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity19> ChildEntities { get; set; }
    }

    public class RelatedEntity19
    {
        public int Id { get; set; }

        public RelatedEntity18 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity20> ChildEntities { get; set; }
    }

    public class RelatedEntity20
    {
        public int Id { get; set; }

        public RelatedEntity19 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity21> ChildEntities { get; set; }
    }

    public class RelatedEntity21
    {
        public int Id { get; set; }

        public RelatedEntity20 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity22> ChildEntities { get; set; }
    }

    public class RelatedEntity22
    {
        public int Id { get; set; }

        public RelatedEntity21 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity23> ChildEntities { get; set; }
    }

    public class RelatedEntity23
    {
        public int Id { get; set; }

        public RelatedEntity22 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity24> ChildEntities { get; set; }
    }

    public class RelatedEntity24
    {
        public int Id { get; set; }

        public RelatedEntity23 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity25> ChildEntities { get; set; }
    }

    public class RelatedEntity25
    {
        public int Id { get; set; }

        public RelatedEntity24 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity26> ChildEntities { get; set; }
    }

    public class RelatedEntity26
    {
        public int Id { get; set; }

        public RelatedEntity25 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity27> ChildEntities { get; set; }
    }

    public class RelatedEntity27
    {
        public int Id { get; set; }

        public RelatedEntity26 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity28> ChildEntities { get; set; }
    }

    public class RelatedEntity28
    {
        public int Id { get; set; }

        public RelatedEntity27 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity29> ChildEntities { get; set; }
    }

    public class RelatedEntity29
    {
        public int Id { get; set; }

        public RelatedEntity28 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity30> ChildEntities { get; set; }
    }

    public class RelatedEntity30
    {
        public int Id { get; set; }

        public RelatedEntity29 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity31> ChildEntities { get; set; }
    }

    public class RelatedEntity31
    {
        public int Id { get; set; }

        public RelatedEntity30 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity32> ChildEntities { get; set; }
    }

    public class RelatedEntity32
    {
        public int Id { get; set; }

        public RelatedEntity31 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity33> ChildEntities { get; set; }
    }

    public class RelatedEntity33
    {
        public int Id { get; set; }

        public RelatedEntity32 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity34> ChildEntities { get; set; }
    }

    public class RelatedEntity34
    {
        public int Id { get; set; }

        public RelatedEntity33 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity35> ChildEntities { get; set; }
    }

    public class RelatedEntity35
    {
        public int Id { get; set; }

        public RelatedEntity34 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity36> ChildEntities { get; set; }
    }

    public class RelatedEntity36
    {
        public int Id { get; set; }

        public RelatedEntity35 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity37> ChildEntities { get; set; }
    }

    public class RelatedEntity37
    {
        public int Id { get; set; }

        public RelatedEntity36 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity38> ChildEntities { get; set; }
    }

    public class RelatedEntity38
    {
        public int Id { get; set; }

        public RelatedEntity37 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity39> ChildEntities { get; set; }
    }

    public class RelatedEntity39
    {
        public int Id { get; set; }

        public RelatedEntity38 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity40> ChildEntities { get; set; }
    }

    public class RelatedEntity40
    {
        public int Id { get; set; }

        public RelatedEntity39 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity41> ChildEntities { get; set; }
    }

    public class RelatedEntity41
    {
        public int Id { get; set; }

        public RelatedEntity40 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity42> ChildEntities { get; set; }
    }

    public class RelatedEntity42
    {
        public int Id { get; set; }

        public RelatedEntity41 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity43> ChildEntities { get; set; }
    }

    public class RelatedEntity43
    {
        public int Id { get; set; }

        public RelatedEntity42 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity44> ChildEntities { get; set; }
    }

    public class RelatedEntity44
    {
        public int Id { get; set; }

        public RelatedEntity43 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity45> ChildEntities { get; set; }
    }

    public class RelatedEntity45
    {
        public int Id { get; set; }

        public RelatedEntity44 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity46> ChildEntities { get; set; }
    }

    public class RelatedEntity46
    {
        public int Id { get; set; }

        public RelatedEntity45 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity47> ChildEntities { get; set; }
    }

    public class RelatedEntity47
    {
        public int Id { get; set; }

        public RelatedEntity46 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity48> ChildEntities { get; set; }
    }

    public class RelatedEntity48
    {
        public int Id { get; set; }

        public RelatedEntity47 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity49> ChildEntities { get; set; }
    }

    public class RelatedEntity49
    {
        public int Id { get; set; }

        public RelatedEntity48 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity50> ChildEntities { get; set; }
    }

    public class RelatedEntity50
    {
        public int Id { get; set; }

        public RelatedEntity49 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity51> ChildEntities { get; set; }
    }

    public class RelatedEntity51
    {
        public int Id { get; set; }

        public RelatedEntity50 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity52> ChildEntities { get; set; }
    }

    public class RelatedEntity52
    {
        public int Id { get; set; }

        public RelatedEntity51 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity53> ChildEntities { get; set; }
    }

    public class RelatedEntity53
    {
        public int Id { get; set; }

        public RelatedEntity52 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity54> ChildEntities { get; set; }
    }

    public class RelatedEntity54
    {
        public int Id { get; set; }

        public RelatedEntity53 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity55> ChildEntities { get; set; }
    }

    public class RelatedEntity55
    {
        public int Id { get; set; }

        public RelatedEntity54 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity56> ChildEntities { get; set; }
    }

    public class RelatedEntity56
    {
        public int Id { get; set; }

        public RelatedEntity55 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity57> ChildEntities { get; set; }
    }

    public class RelatedEntity57
    {
        public int Id { get; set; }

        public RelatedEntity56 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity58> ChildEntities { get; set; }
    }

    public class RelatedEntity58
    {
        public int Id { get; set; }

        public RelatedEntity57 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity59> ChildEntities { get; set; }
    }

    public class RelatedEntity59
    {
        public int Id { get; set; }

        public RelatedEntity58 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity60> ChildEntities { get; set; }
    }

    public class RelatedEntity60
    {
        public int Id { get; set; }

        public RelatedEntity59 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity61> ChildEntities { get; set; }
    }

    public class RelatedEntity61
    {
        public int Id { get; set; }

        public RelatedEntity60 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity62> ChildEntities { get; set; }
    }

    public class RelatedEntity62
    {
        public int Id { get; set; }

        public RelatedEntity61 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity63> ChildEntities { get; set; }
    }

    public class RelatedEntity63
    {
        public int Id { get; set; }

        public RelatedEntity62 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity64> ChildEntities { get; set; }
    }

    public class RelatedEntity64
    {
        public int Id { get; set; }

        public RelatedEntity63 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity65> ChildEntities { get; set; }
    }

    public class RelatedEntity65
    {
        public int Id { get; set; }

        public RelatedEntity64 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity66> ChildEntities { get; set; }
    }

    public class RelatedEntity66
    {
        public int Id { get; set; }

        public RelatedEntity65 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity67> ChildEntities { get; set; }
    }

    public class RelatedEntity67
    {
        public int Id { get; set; }

        public RelatedEntity66 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity68> ChildEntities { get; set; }
    }

    public class RelatedEntity68
    {
        public int Id { get; set; }

        public RelatedEntity67 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity69> ChildEntities { get; set; }
    }

    public class RelatedEntity69
    {
        public int Id { get; set; }

        public RelatedEntity68 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity70> ChildEntities { get; set; }
    }

    public class RelatedEntity70
    {
        public int Id { get; set; }

        public RelatedEntity69 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity71> ChildEntities { get; set; }
    }

    public class RelatedEntity71
    {
        public int Id { get; set; }

        public RelatedEntity70 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity72> ChildEntities { get; set; }
    }

    public class RelatedEntity72
    {
        public int Id { get; set; }

        public RelatedEntity71 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity73> ChildEntities { get; set; }
    }

    public class RelatedEntity73
    {
        public int Id { get; set; }

        public RelatedEntity72 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity74> ChildEntities { get; set; }
    }

    public class RelatedEntity74
    {
        public int Id { get; set; }

        public RelatedEntity73 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity75> ChildEntities { get; set; }
    }

    public class RelatedEntity75
    {
        public int Id { get; set; }

        public RelatedEntity74 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity76> ChildEntities { get; set; }
    }

    public class RelatedEntity76
    {
        public int Id { get; set; }

        public RelatedEntity75 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity77> ChildEntities { get; set; }
    }

    public class RelatedEntity77
    {
        public int Id { get; set; }

        public RelatedEntity76 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity78> ChildEntities { get; set; }
    }

    public class RelatedEntity78
    {
        public int Id { get; set; }

        public RelatedEntity77 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity79> ChildEntities { get; set; }
    }

    public class RelatedEntity79
    {
        public int Id { get; set; }

        public RelatedEntity78 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity80> ChildEntities { get; set; }
    }

    public class RelatedEntity80
    {
        public int Id { get; set; }

        public RelatedEntity79 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity81> ChildEntities { get; set; }
    }

    public class RelatedEntity81
    {
        public int Id { get; set; }

        public RelatedEntity80 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity82> ChildEntities { get; set; }
    }

    public class RelatedEntity82
    {
        public int Id { get; set; }

        public RelatedEntity81 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity83> ChildEntities { get; set; }
    }

    public class RelatedEntity83
    {
        public int Id { get; set; }

        public RelatedEntity82 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity84> ChildEntities { get; set; }
    }

    public class RelatedEntity84
    {
        public int Id { get; set; }

        public RelatedEntity83 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity85> ChildEntities { get; set; }
    }

    public class RelatedEntity85
    {
        public int Id { get; set; }

        public RelatedEntity84 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity86> ChildEntities { get; set; }
    }

    public class RelatedEntity86
    {
        public int Id { get; set; }

        public RelatedEntity85 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity87> ChildEntities { get; set; }
    }

    public class RelatedEntity87
    {
        public int Id { get; set; }

        public RelatedEntity86 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity88> ChildEntities { get; set; }
    }

    public class RelatedEntity88
    {
        public int Id { get; set; }

        public RelatedEntity87 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity89> ChildEntities { get; set; }
    }

    public class RelatedEntity89
    {
        public int Id { get; set; }

        public RelatedEntity88 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity90> ChildEntities { get; set; }
    }

    public class RelatedEntity90
    {
        public int Id { get; set; }

        public RelatedEntity89 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity91> ChildEntities { get; set; }
    }

    public class RelatedEntity91
    {
        public int Id { get; set; }

        public RelatedEntity90 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity92> ChildEntities { get; set; }
    }

    public class RelatedEntity92
    {
        public int Id { get; set; }

        public RelatedEntity91 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity93> ChildEntities { get; set; }
    }

    public class RelatedEntity93
    {
        public int Id { get; set; }

        public RelatedEntity92 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity94> ChildEntities { get; set; }
    }

    public class RelatedEntity94
    {
        public int Id { get; set; }

        public RelatedEntity93 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity95> ChildEntities { get; set; }
    }

    public class RelatedEntity95
    {
        public int Id { get; set; }

        public RelatedEntity94 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity96> ChildEntities { get; set; }
    }

    public class RelatedEntity96
    {
        public int Id { get; set; }

        public RelatedEntity95 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity97> ChildEntities { get; set; }
    }

    public class RelatedEntity97
    {
        public int Id { get; set; }

        public RelatedEntity96 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity98> ChildEntities { get; set; }
    }

    public class RelatedEntity98
    {
        public int Id { get; set; }

        public RelatedEntity97 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity99> ChildEntities { get; set; }
    }

    public class RelatedEntity99
    {
        public int Id { get; set; }

        public RelatedEntity98 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity100> ChildEntities { get; set; }
    }

    public class RelatedEntity100
    {
        public int Id { get; set; }

        public RelatedEntity99 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity101> ChildEntities { get; set; }
    }

    public class RelatedEntity101
    {
        public int Id { get; set; }

        public RelatedEntity100 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity102> ChildEntities { get; set; }
    }

    public class RelatedEntity102
    {
        public int Id { get; set; }

        public RelatedEntity101 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity103> ChildEntities { get; set; }
    }

    public class RelatedEntity103
    {
        public int Id { get; set; }

        public RelatedEntity102 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity104> ChildEntities { get; set; }
    }

    public class RelatedEntity104
    {
        public int Id { get; set; }

        public RelatedEntity103 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity105> ChildEntities { get; set; }
    }

    public class RelatedEntity105
    {
        public int Id { get; set; }

        public RelatedEntity104 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity106> ChildEntities { get; set; }
    }

    public class RelatedEntity106
    {
        public int Id { get; set; }

        public RelatedEntity105 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity107> ChildEntities { get; set; }
    }

    public class RelatedEntity107
    {
        public int Id { get; set; }

        public RelatedEntity106 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity108> ChildEntities { get; set; }
    }

    public class RelatedEntity108
    {
        public int Id { get; set; }

        public RelatedEntity107 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity109> ChildEntities { get; set; }
    }

    public class RelatedEntity109
    {
        public int Id { get; set; }

        public RelatedEntity108 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity110> ChildEntities { get; set; }
    }

    public class RelatedEntity110
    {
        public int Id { get; set; }

        public RelatedEntity109 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity111> ChildEntities { get; set; }
    }

    public class RelatedEntity111
    {
        public int Id { get; set; }

        public RelatedEntity110 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity112> ChildEntities { get; set; }
    }

    public class RelatedEntity112
    {
        public int Id { get; set; }

        public RelatedEntity111 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity113> ChildEntities { get; set; }
    }

    public class RelatedEntity113
    {
        public int Id { get; set; }

        public RelatedEntity112 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity114> ChildEntities { get; set; }
    }

    public class RelatedEntity114
    {
        public int Id { get; set; }

        public RelatedEntity113 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity115> ChildEntities { get; set; }
    }

    public class RelatedEntity115
    {
        public int Id { get; set; }

        public RelatedEntity114 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity116> ChildEntities { get; set; }
    }

    public class RelatedEntity116
    {
        public int Id { get; set; }

        public RelatedEntity115 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity117> ChildEntities { get; set; }
    }

    public class RelatedEntity117
    {
        public int Id { get; set; }

        public RelatedEntity116 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity118> ChildEntities { get; set; }
    }

    public class RelatedEntity118
    {
        public int Id { get; set; }

        public RelatedEntity117 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity119> ChildEntities { get; set; }
    }

    public class RelatedEntity119
    {
        public int Id { get; set; }

        public RelatedEntity118 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity120> ChildEntities { get; set; }
    }

    public class RelatedEntity120
    {
        public int Id { get; set; }

        public RelatedEntity119 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity121> ChildEntities { get; set; }
    }

    public class RelatedEntity121
    {
        public int Id { get; set; }

        public RelatedEntity120 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity122> ChildEntities { get; set; }
    }

    public class RelatedEntity122
    {
        public int Id { get; set; }

        public RelatedEntity121 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity123> ChildEntities { get; set; }
    }

    public class RelatedEntity123
    {
        public int Id { get; set; }

        public RelatedEntity122 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity124> ChildEntities { get; set; }
    }

    public class RelatedEntity124
    {
        public int Id { get; set; }

        public RelatedEntity123 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity125> ChildEntities { get; set; }
    }

    public class RelatedEntity125
    {
        public int Id { get; set; }

        public RelatedEntity124 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity126> ChildEntities { get; set; }
    }

    public class RelatedEntity126
    {
        public int Id { get; set; }

        public RelatedEntity125 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity127> ChildEntities { get; set; }
    }

    public class RelatedEntity127
    {
        public int Id { get; set; }

        public RelatedEntity126 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity128> ChildEntities { get; set; }
    }

    public class RelatedEntity128
    {
        public int Id { get; set; }

        public RelatedEntity127 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity129> ChildEntities { get; set; }
    }

    public class RelatedEntity129
    {
        public int Id { get; set; }

        public RelatedEntity128 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity130> ChildEntities { get; set; }
    }

    public class RelatedEntity130
    {
        public int Id { get; set; }

        public RelatedEntity129 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity131> ChildEntities { get; set; }
    }

    public class RelatedEntity131
    {
        public int Id { get; set; }

        public RelatedEntity130 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity132> ChildEntities { get; set; }
    }

    public class RelatedEntity132
    {
        public int Id { get; set; }

        public RelatedEntity131 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity133> ChildEntities { get; set; }
    }

    public class RelatedEntity133
    {
        public int Id { get; set; }

        public RelatedEntity132 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity134> ChildEntities { get; set; }
    }

    public class RelatedEntity134
    {
        public int Id { get; set; }

        public RelatedEntity133 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity135> ChildEntities { get; set; }
    }

    public class RelatedEntity135
    {
        public int Id { get; set; }

        public RelatedEntity134 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity136> ChildEntities { get; set; }
    }

    public class RelatedEntity136
    {
        public int Id { get; set; }

        public RelatedEntity135 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity137> ChildEntities { get; set; }
    }

    public class RelatedEntity137
    {
        public int Id { get; set; }

        public RelatedEntity136 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity138> ChildEntities { get; set; }
    }

    public class RelatedEntity138
    {
        public int Id { get; set; }

        public RelatedEntity137 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity139> ChildEntities { get; set; }
    }

    public class RelatedEntity139
    {
        public int Id { get; set; }

        public RelatedEntity138 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity140> ChildEntities { get; set; }
    }

    public class RelatedEntity140
    {
        public int Id { get; set; }

        public RelatedEntity139 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity141> ChildEntities { get; set; }
    }

    public class RelatedEntity141
    {
        public int Id { get; set; }

        public RelatedEntity140 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity142> ChildEntities { get; set; }
    }

    public class RelatedEntity142
    {
        public int Id { get; set; }

        public RelatedEntity141 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity143> ChildEntities { get; set; }
    }

    public class RelatedEntity143
    {
        public int Id { get; set; }

        public RelatedEntity142 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity144> ChildEntities { get; set; }
    }

    public class RelatedEntity144
    {
        public int Id { get; set; }

        public RelatedEntity143 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity145> ChildEntities { get; set; }
    }

    public class RelatedEntity145
    {
        public int Id { get; set; }

        public RelatedEntity144 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity146> ChildEntities { get; set; }
    }

    public class RelatedEntity146
    {
        public int Id { get; set; }

        public RelatedEntity145 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity147> ChildEntities { get; set; }
    }

    public class RelatedEntity147
    {
        public int Id { get; set; }

        public RelatedEntity146 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity148> ChildEntities { get; set; }
    }

    public class RelatedEntity148
    {
        public int Id { get; set; }

        public RelatedEntity147 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity149> ChildEntities { get; set; }
    }

    public class RelatedEntity149
    {
        public int Id { get; set; }

        public RelatedEntity148 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity150> ChildEntities { get; set; }
    }

    public class RelatedEntity150
    {
        public int Id { get; set; }

        public RelatedEntity149 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity151> ChildEntities { get; set; }
    }

    public class RelatedEntity151
    {
        public int Id { get; set; }

        public RelatedEntity150 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity152> ChildEntities { get; set; }
    }

    public class RelatedEntity152
    {
        public int Id { get; set; }

        public RelatedEntity151 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity153> ChildEntities { get; set; }
    }

    public class RelatedEntity153
    {
        public int Id { get; set; }

        public RelatedEntity152 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity154> ChildEntities { get; set; }
    }

    public class RelatedEntity154
    {
        public int Id { get; set; }

        public RelatedEntity153 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity155> ChildEntities { get; set; }
    }

    public class RelatedEntity155
    {
        public int Id { get; set; }

        public RelatedEntity154 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity156> ChildEntities { get; set; }
    }

    public class RelatedEntity156
    {
        public int Id { get; set; }

        public RelatedEntity155 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity157> ChildEntities { get; set; }
    }

    public class RelatedEntity157
    {
        public int Id { get; set; }

        public RelatedEntity156 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity158> ChildEntities { get; set; }
    }

    public class RelatedEntity158
    {
        public int Id { get; set; }

        public RelatedEntity157 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity159> ChildEntities { get; set; }
    }

    public class RelatedEntity159
    {
        public int Id { get; set; }

        public RelatedEntity158 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity160> ChildEntities { get; set; }
    }

    public class RelatedEntity160
    {
        public int Id { get; set; }

        public RelatedEntity159 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity161> ChildEntities { get; set; }
    }

    public class RelatedEntity161
    {
        public int Id { get; set; }

        public RelatedEntity160 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity162> ChildEntities { get; set; }
    }

    public class RelatedEntity162
    {
        public int Id { get; set; }

        public RelatedEntity161 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity163> ChildEntities { get; set; }
    }

    public class RelatedEntity163
    {
        public int Id { get; set; }

        public RelatedEntity162 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity164> ChildEntities { get; set; }
    }

    public class RelatedEntity164
    {
        public int Id { get; set; }

        public RelatedEntity163 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity165> ChildEntities { get; set; }
    }

    public class RelatedEntity165
    {
        public int Id { get; set; }

        public RelatedEntity164 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity166> ChildEntities { get; set; }
    }

    public class RelatedEntity166
    {
        public int Id { get; set; }

        public RelatedEntity165 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity167> ChildEntities { get; set; }
    }

    public class RelatedEntity167
    {
        public int Id { get; set; }

        public RelatedEntity166 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity168> ChildEntities { get; set; }
    }

    public class RelatedEntity168
    {
        public int Id { get; set; }

        public RelatedEntity167 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity169> ChildEntities { get; set; }
    }

    public class RelatedEntity169
    {
        public int Id { get; set; }

        public RelatedEntity168 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity170> ChildEntities { get; set; }
    }

    public class RelatedEntity170
    {
        public int Id { get; set; }

        public RelatedEntity169 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity171> ChildEntities { get; set; }
    }

    public class RelatedEntity171
    {
        public int Id { get; set; }

        public RelatedEntity170 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity172> ChildEntities { get; set; }
    }

    public class RelatedEntity172
    {
        public int Id { get; set; }

        public RelatedEntity171 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity173> ChildEntities { get; set; }
    }

    public class RelatedEntity173
    {
        public int Id { get; set; }

        public RelatedEntity172 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity174> ChildEntities { get; set; }
    }

    public class RelatedEntity174
    {
        public int Id { get; set; }

        public RelatedEntity173 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity175> ChildEntities { get; set; }
    }

    public class RelatedEntity175
    {
        public int Id { get; set; }

        public RelatedEntity174 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity176> ChildEntities { get; set; }
    }

    public class RelatedEntity176
    {
        public int Id { get; set; }

        public RelatedEntity175 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity177> ChildEntities { get; set; }
    }

    public class RelatedEntity177
    {
        public int Id { get; set; }

        public RelatedEntity176 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity178> ChildEntities { get; set; }
    }

    public class RelatedEntity178
    {
        public int Id { get; set; }

        public RelatedEntity177 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity179> ChildEntities { get; set; }
    }

    public class RelatedEntity179
    {
        public int Id { get; set; }

        public RelatedEntity178 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity180> ChildEntities { get; set; }
    }

    public class RelatedEntity180
    {
        public int Id { get; set; }

        public RelatedEntity179 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity181> ChildEntities { get; set; }
    }

    public class RelatedEntity181
    {
        public int Id { get; set; }

        public RelatedEntity180 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity182> ChildEntities { get; set; }
    }

    public class RelatedEntity182
    {
        public int Id { get; set; }

        public RelatedEntity181 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity183> ChildEntities { get; set; }
    }

    public class RelatedEntity183
    {
        public int Id { get; set; }

        public RelatedEntity182 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity184> ChildEntities { get; set; }
    }

    public class RelatedEntity184
    {
        public int Id { get; set; }

        public RelatedEntity183 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity185> ChildEntities { get; set; }
    }

    public class RelatedEntity185
    {
        public int Id { get; set; }

        public RelatedEntity184 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity186> ChildEntities { get; set; }
    }

    public class RelatedEntity186
    {
        public int Id { get; set; }

        public RelatedEntity185 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity187> ChildEntities { get; set; }
    }

    public class RelatedEntity187
    {
        public int Id { get; set; }

        public RelatedEntity186 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity188> ChildEntities { get; set; }
    }

    public class RelatedEntity188
    {
        public int Id { get; set; }

        public RelatedEntity187 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity189> ChildEntities { get; set; }
    }

    public class RelatedEntity189
    {
        public int Id { get; set; }

        public RelatedEntity188 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity190> ChildEntities { get; set; }
    }

    public class RelatedEntity190
    {
        public int Id { get; set; }

        public RelatedEntity189 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity191> ChildEntities { get; set; }
    }

    public class RelatedEntity191
    {
        public int Id { get; set; }

        public RelatedEntity190 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity192> ChildEntities { get; set; }
    }

    public class RelatedEntity192
    {
        public int Id { get; set; }

        public RelatedEntity191 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity193> ChildEntities { get; set; }
    }

    public class RelatedEntity193
    {
        public int Id { get; set; }

        public RelatedEntity192 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity194> ChildEntities { get; set; }
    }

    public class RelatedEntity194
    {
        public int Id { get; set; }

        public RelatedEntity193 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity195> ChildEntities { get; set; }
    }

    public class RelatedEntity195
    {
        public int Id { get; set; }

        public RelatedEntity194 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity196> ChildEntities { get; set; }
    }

    public class RelatedEntity196
    {
        public int Id { get; set; }

        public RelatedEntity195 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity197> ChildEntities { get; set; }
    }

    public class RelatedEntity197
    {
        public int Id { get; set; }

        public RelatedEntity196 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity198> ChildEntities { get; set; }
    }

    public class RelatedEntity198
    {
        public int Id { get; set; }

        public RelatedEntity197 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity199> ChildEntities { get; set; }
    }

    public class RelatedEntity199
    {
        public int Id { get; set; }

        public RelatedEntity198 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity200> ChildEntities { get; set; }
    }

    public class RelatedEntity200
    {
        public int Id { get; set; }

        public RelatedEntity199 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity201> ChildEntities { get; set; }
    }

    public class RelatedEntity201
    {
        public int Id { get; set; }

        public RelatedEntity200 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity202> ChildEntities { get; set; }
    }

    public class RelatedEntity202
    {
        public int Id { get; set; }

        public RelatedEntity201 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity203> ChildEntities { get; set; }
    }

    public class RelatedEntity203
    {
        public int Id { get; set; }

        public RelatedEntity202 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity204> ChildEntities { get; set; }
    }

    public class RelatedEntity204
    {
        public int Id { get; set; }

        public RelatedEntity203 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity205> ChildEntities { get; set; }
    }

    public class RelatedEntity205
    {
        public int Id { get; set; }

        public RelatedEntity204 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity206> ChildEntities { get; set; }
    }

    public class RelatedEntity206
    {
        public int Id { get; set; }

        public RelatedEntity205 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity207> ChildEntities { get; set; }
    }

    public class RelatedEntity207
    {
        public int Id { get; set; }

        public RelatedEntity206 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity208> ChildEntities { get; set; }
    }

    public class RelatedEntity208
    {
        public int Id { get; set; }

        public RelatedEntity207 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity209> ChildEntities { get; set; }
    }

    public class RelatedEntity209
    {
        public int Id { get; set; }

        public RelatedEntity208 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity210> ChildEntities { get; set; }
    }

    public class RelatedEntity210
    {
        public int Id { get; set; }

        public RelatedEntity209 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity211> ChildEntities { get; set; }
    }

    public class RelatedEntity211
    {
        public int Id { get; set; }

        public RelatedEntity210 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity212> ChildEntities { get; set; }
    }

    public class RelatedEntity212
    {
        public int Id { get; set; }

        public RelatedEntity211 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity213> ChildEntities { get; set; }
    }

    public class RelatedEntity213
    {
        public int Id { get; set; }

        public RelatedEntity212 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity214> ChildEntities { get; set; }
    }

    public class RelatedEntity214
    {
        public int Id { get; set; }

        public RelatedEntity213 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity215> ChildEntities { get; set; }
    }

    public class RelatedEntity215
    {
        public int Id { get; set; }

        public RelatedEntity214 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity216> ChildEntities { get; set; }
    }

    public class RelatedEntity216
    {
        public int Id { get; set; }

        public RelatedEntity215 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity217> ChildEntities { get; set; }
    }

    public class RelatedEntity217
    {
        public int Id { get; set; }

        public RelatedEntity216 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity218> ChildEntities { get; set; }
    }

    public class RelatedEntity218
    {
        public int Id { get; set; }

        public RelatedEntity217 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity219> ChildEntities { get; set; }
    }

    public class RelatedEntity219
    {
        public int Id { get; set; }

        public RelatedEntity218 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity220> ChildEntities { get; set; }
    }

    public class RelatedEntity220
    {
        public int Id { get; set; }

        public RelatedEntity219 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity221> ChildEntities { get; set; }
    }

    public class RelatedEntity221
    {
        public int Id { get; set; }

        public RelatedEntity220 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity222> ChildEntities { get; set; }
    }

    public class RelatedEntity222
    {
        public int Id { get; set; }

        public RelatedEntity221 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity223> ChildEntities { get; set; }
    }

    public class RelatedEntity223
    {
        public int Id { get; set; }

        public RelatedEntity222 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity224> ChildEntities { get; set; }
    }

    public class RelatedEntity224
    {
        public int Id { get; set; }

        public RelatedEntity223 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity225> ChildEntities { get; set; }
    }

    public class RelatedEntity225
    {
        public int Id { get; set; }

        public RelatedEntity224 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity226> ChildEntities { get; set; }
    }

    public class RelatedEntity226
    {
        public int Id { get; set; }

        public RelatedEntity225 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity227> ChildEntities { get; set; }
    }

    public class RelatedEntity227
    {
        public int Id { get; set; }

        public RelatedEntity226 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity228> ChildEntities { get; set; }
    }

    public class RelatedEntity228
    {
        public int Id { get; set; }

        public RelatedEntity227 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity229> ChildEntities { get; set; }
    }

    public class RelatedEntity229
    {
        public int Id { get; set; }

        public RelatedEntity228 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity230> ChildEntities { get; set; }
    }

    public class RelatedEntity230
    {
        public int Id { get; set; }

        public RelatedEntity229 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity231> ChildEntities { get; set; }
    }

    public class RelatedEntity231
    {
        public int Id { get; set; }

        public RelatedEntity230 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity232> ChildEntities { get; set; }
    }

    public class RelatedEntity232
    {
        public int Id { get; set; }

        public RelatedEntity231 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity233> ChildEntities { get; set; }
    }

    public class RelatedEntity233
    {
        public int Id { get; set; }

        public RelatedEntity232 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity234> ChildEntities { get; set; }
    }

    public class RelatedEntity234
    {
        public int Id { get; set; }

        public RelatedEntity233 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity235> ChildEntities { get; set; }
    }

    public class RelatedEntity235
    {
        public int Id { get; set; }

        public RelatedEntity234 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity236> ChildEntities { get; set; }
    }

    public class RelatedEntity236
    {
        public int Id { get; set; }

        public RelatedEntity235 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity237> ChildEntities { get; set; }
    }

    public class RelatedEntity237
    {
        public int Id { get; set; }

        public RelatedEntity236 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity238> ChildEntities { get; set; }
    }

    public class RelatedEntity238
    {
        public int Id { get; set; }

        public RelatedEntity237 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity239> ChildEntities { get; set; }
    }

    public class RelatedEntity239
    {
        public int Id { get; set; }

        public RelatedEntity238 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity240> ChildEntities { get; set; }
    }

    public class RelatedEntity240
    {
        public int Id { get; set; }

        public RelatedEntity239 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity241> ChildEntities { get; set; }
    }

    public class RelatedEntity241
    {
        public int Id { get; set; }

        public RelatedEntity240 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity242> ChildEntities { get; set; }
    }

    public class RelatedEntity242
    {
        public int Id { get; set; }

        public RelatedEntity241 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity243> ChildEntities { get; set; }
    }

    public class RelatedEntity243
    {
        public int Id { get; set; }

        public RelatedEntity242 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity244> ChildEntities { get; set; }
    }

    public class RelatedEntity244
    {
        public int Id { get; set; }

        public RelatedEntity243 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity245> ChildEntities { get; set; }
    }

    public class RelatedEntity245
    {
        public int Id { get; set; }

        public RelatedEntity244 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity246> ChildEntities { get; set; }
    }

    public class RelatedEntity246
    {
        public int Id { get; set; }

        public RelatedEntity245 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity247> ChildEntities { get; set; }
    }

    public class RelatedEntity247
    {
        public int Id { get; set; }

        public RelatedEntity246 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity248> ChildEntities { get; set; }
    }

    public class RelatedEntity248
    {
        public int Id { get; set; }

        public RelatedEntity247 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity249> ChildEntities { get; set; }
    }

    public class RelatedEntity249
    {
        public int Id { get; set; }

        public RelatedEntity248 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity250> ChildEntities { get; set; }
    }

    public class RelatedEntity250
    {
        public int Id { get; set; }

        public RelatedEntity249 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity251> ChildEntities { get; set; }
    }

    public class RelatedEntity251
    {
        public int Id { get; set; }

        public RelatedEntity250 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity252> ChildEntities { get; set; }
    }

    public class RelatedEntity252
    {
        public int Id { get; set; }

        public RelatedEntity251 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity253> ChildEntities { get; set; }
    }

    public class RelatedEntity253
    {
        public int Id { get; set; }

        public RelatedEntity252 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity254> ChildEntities { get; set; }
    }

    public class RelatedEntity254
    {
        public int Id { get; set; }

        public RelatedEntity253 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity255> ChildEntities { get; set; }
    }

    public class RelatedEntity255
    {
        public int Id { get; set; }

        public RelatedEntity254 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity256> ChildEntities { get; set; }
    }

    public class RelatedEntity256
    {
        public int Id { get; set; }

        public RelatedEntity255 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity257> ChildEntities { get; set; }
    }

    public class RelatedEntity257
    {
        public int Id { get; set; }

        public RelatedEntity256 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity258> ChildEntities { get; set; }
    }

    public class RelatedEntity258
    {
        public int Id { get; set; }

        public RelatedEntity257 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity259> ChildEntities { get; set; }
    }

    public class RelatedEntity259
    {
        public int Id { get; set; }

        public RelatedEntity258 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity260> ChildEntities { get; set; }
    }

    public class RelatedEntity260
    {
        public int Id { get; set; }

        public RelatedEntity259 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity261> ChildEntities { get; set; }
    }

    public class RelatedEntity261
    {
        public int Id { get; set; }

        public RelatedEntity260 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity262> ChildEntities { get; set; }
    }

    public class RelatedEntity262
    {
        public int Id { get; set; }

        public RelatedEntity261 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity263> ChildEntities { get; set; }
    }

    public class RelatedEntity263
    {
        public int Id { get; set; }

        public RelatedEntity262 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity264> ChildEntities { get; set; }
    }

    public class RelatedEntity264
    {
        public int Id { get; set; }

        public RelatedEntity263 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity265> ChildEntities { get; set; }
    }

    public class RelatedEntity265
    {
        public int Id { get; set; }

        public RelatedEntity264 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity266> ChildEntities { get; set; }
    }

    public class RelatedEntity266
    {
        public int Id { get; set; }

        public RelatedEntity265 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity267> ChildEntities { get; set; }
    }

    public class RelatedEntity267
    {
        public int Id { get; set; }

        public RelatedEntity266 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity268> ChildEntities { get; set; }
    }

    public class RelatedEntity268
    {
        public int Id { get; set; }

        public RelatedEntity267 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity269> ChildEntities { get; set; }
    }

    public class RelatedEntity269
    {
        public int Id { get; set; }

        public RelatedEntity268 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity270> ChildEntities { get; set; }
    }

    public class RelatedEntity270
    {
        public int Id { get; set; }

        public RelatedEntity269 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity271> ChildEntities { get; set; }
    }

    public class RelatedEntity271
    {
        public int Id { get; set; }

        public RelatedEntity270 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity272> ChildEntities { get; set; }
    }

    public class RelatedEntity272
    {
        public int Id { get; set; }

        public RelatedEntity271 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity273> ChildEntities { get; set; }
    }

    public class RelatedEntity273
    {
        public int Id { get; set; }

        public RelatedEntity272 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity274> ChildEntities { get; set; }
    }

    public class RelatedEntity274
    {
        public int Id { get; set; }

        public RelatedEntity273 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity275> ChildEntities { get; set; }
    }

    public class RelatedEntity275
    {
        public int Id { get; set; }

        public RelatedEntity274 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity276> ChildEntities { get; set; }
    }

    public class RelatedEntity276
    {
        public int Id { get; set; }

        public RelatedEntity275 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity277> ChildEntities { get; set; }
    }

    public class RelatedEntity277
    {
        public int Id { get; set; }

        public RelatedEntity276 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity278> ChildEntities { get; set; }
    }

    public class RelatedEntity278
    {
        public int Id { get; set; }

        public RelatedEntity277 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity279> ChildEntities { get; set; }
    }

    public class RelatedEntity279
    {
        public int Id { get; set; }

        public RelatedEntity278 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity280> ChildEntities { get; set; }
    }

    public class RelatedEntity280
    {
        public int Id { get; set; }

        public RelatedEntity279 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity281> ChildEntities { get; set; }
    }

    public class RelatedEntity281
    {
        public int Id { get; set; }

        public RelatedEntity280 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity282> ChildEntities { get; set; }
    }

    public class RelatedEntity282
    {
        public int Id { get; set; }

        public RelatedEntity281 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity283> ChildEntities { get; set; }
    }

    public class RelatedEntity283
    {
        public int Id { get; set; }

        public RelatedEntity282 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity284> ChildEntities { get; set; }
    }

    public class RelatedEntity284
    {
        public int Id { get; set; }

        public RelatedEntity283 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity285> ChildEntities { get; set; }
    }

    public class RelatedEntity285
    {
        public int Id { get; set; }

        public RelatedEntity284 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity286> ChildEntities { get; set; }
    }

    public class RelatedEntity286
    {
        public int Id { get; set; }

        public RelatedEntity285 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity287> ChildEntities { get; set; }
    }

    public class RelatedEntity287
    {
        public int Id { get; set; }

        public RelatedEntity286 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity288> ChildEntities { get; set; }
    }

    public class RelatedEntity288
    {
        public int Id { get; set; }

        public RelatedEntity287 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity289> ChildEntities { get; set; }
    }

    public class RelatedEntity289
    {
        public int Id { get; set; }

        public RelatedEntity288 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity290> ChildEntities { get; set; }
    }

    public class RelatedEntity290
    {
        public int Id { get; set; }

        public RelatedEntity289 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity291> ChildEntities { get; set; }
    }

    public class RelatedEntity291
    {
        public int Id { get; set; }

        public RelatedEntity290 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity292> ChildEntities { get; set; }
    }

    public class RelatedEntity292
    {
        public int Id { get; set; }

        public RelatedEntity291 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity293> ChildEntities { get; set; }
    }

    public class RelatedEntity293
    {
        public int Id { get; set; }

        public RelatedEntity292 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity294> ChildEntities { get; set; }
    }

    public class RelatedEntity294
    {
        public int Id { get; set; }

        public RelatedEntity293 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity295> ChildEntities { get; set; }
    }

    public class RelatedEntity295
    {
        public int Id { get; set; }

        public RelatedEntity294 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity296> ChildEntities { get; set; }
    }

    public class RelatedEntity296
    {
        public int Id { get; set; }

        public RelatedEntity295 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity297> ChildEntities { get; set; }
    }

    public class RelatedEntity297
    {
        public int Id { get; set; }

        public RelatedEntity296 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity298> ChildEntities { get; set; }
    }

    public class RelatedEntity298
    {
        public int Id { get; set; }

        public RelatedEntity297 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity299> ChildEntities { get; set; }
    }

    public class RelatedEntity299
    {
        public int Id { get; set; }

        public RelatedEntity298 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity300> ChildEntities { get; set; }
    }

    public class RelatedEntity300
    {
        public int Id { get; set; }

        public RelatedEntity299 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity301> ChildEntities { get; set; }
    }

    public class RelatedEntity301
    {
        public int Id { get; set; }

        public RelatedEntity300 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity302> ChildEntities { get; set; }
    }

    public class RelatedEntity302
    {
        public int Id { get; set; }

        public RelatedEntity301 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity303> ChildEntities { get; set; }
    }

    public class RelatedEntity303
    {
        public int Id { get; set; }

        public RelatedEntity302 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity304> ChildEntities { get; set; }
    }

    public class RelatedEntity304
    {
        public int Id { get; set; }

        public RelatedEntity303 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity305> ChildEntities { get; set; }
    }

    public class RelatedEntity305
    {
        public int Id { get; set; }

        public RelatedEntity304 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity306> ChildEntities { get; set; }
    }

    public class RelatedEntity306
    {
        public int Id { get; set; }

        public RelatedEntity305 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity307> ChildEntities { get; set; }
    }

    public class RelatedEntity307
    {
        public int Id { get; set; }

        public RelatedEntity306 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity308> ChildEntities { get; set; }
    }

    public class RelatedEntity308
    {
        public int Id { get; set; }

        public RelatedEntity307 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity309> ChildEntities { get; set; }
    }

    public class RelatedEntity309
    {
        public int Id { get; set; }

        public RelatedEntity308 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity310> ChildEntities { get; set; }
    }

    public class RelatedEntity310
    {
        public int Id { get; set; }

        public RelatedEntity309 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity311> ChildEntities { get; set; }
    }

    public class RelatedEntity311
    {
        public int Id { get; set; }

        public RelatedEntity310 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity312> ChildEntities { get; set; }
    }

    public class RelatedEntity312
    {
        public int Id { get; set; }

        public RelatedEntity311 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity313> ChildEntities { get; set; }
    }

    public class RelatedEntity313
    {
        public int Id { get; set; }

        public RelatedEntity312 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity314> ChildEntities { get; set; }
    }

    public class RelatedEntity314
    {
        public int Id { get; set; }

        public RelatedEntity313 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity315> ChildEntities { get; set; }
    }

    public class RelatedEntity315
    {
        public int Id { get; set; }

        public RelatedEntity314 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity316> ChildEntities { get; set; }
    }

    public class RelatedEntity316
    {
        public int Id { get; set; }

        public RelatedEntity315 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity317> ChildEntities { get; set; }
    }

    public class RelatedEntity317
    {
        public int Id { get; set; }

        public RelatedEntity316 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity318> ChildEntities { get; set; }
    }

    public class RelatedEntity318
    {
        public int Id { get; set; }

        public RelatedEntity317 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity319> ChildEntities { get; set; }
    }

    public class RelatedEntity319
    {
        public int Id { get; set; }

        public RelatedEntity318 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity320> ChildEntities { get; set; }
    }

    public class RelatedEntity320
    {
        public int Id { get; set; }

        public RelatedEntity319 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity321> ChildEntities { get; set; }
    }

    public class RelatedEntity321
    {
        public int Id { get; set; }

        public RelatedEntity320 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity322> ChildEntities { get; set; }
    }

    public class RelatedEntity322
    {
        public int Id { get; set; }

        public RelatedEntity321 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity323> ChildEntities { get; set; }
    }

    public class RelatedEntity323
    {
        public int Id { get; set; }

        public RelatedEntity322 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity324> ChildEntities { get; set; }
    }

    public class RelatedEntity324
    {
        public int Id { get; set; }

        public RelatedEntity323 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity325> ChildEntities { get; set; }
    }

    public class RelatedEntity325
    {
        public int Id { get; set; }

        public RelatedEntity324 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity326> ChildEntities { get; set; }
    }

    public class RelatedEntity326
    {
        public int Id { get; set; }

        public RelatedEntity325 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity327> ChildEntities { get; set; }
    }

    public class RelatedEntity327
    {
        public int Id { get; set; }

        public RelatedEntity326 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity328> ChildEntities { get; set; }
    }

    public class RelatedEntity328
    {
        public int Id { get; set; }

        public RelatedEntity327 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity329> ChildEntities { get; set; }
    }

    public class RelatedEntity329
    {
        public int Id { get; set; }

        public RelatedEntity328 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity330> ChildEntities { get; set; }
    }

    public class RelatedEntity330
    {
        public int Id { get; set; }

        public RelatedEntity329 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity331> ChildEntities { get; set; }
    }

    public class RelatedEntity331
    {
        public int Id { get; set; }

        public RelatedEntity330 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity332> ChildEntities { get; set; }
    }

    public class RelatedEntity332
    {
        public int Id { get; set; }

        public RelatedEntity331 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity333> ChildEntities { get; set; }
    }

    public class RelatedEntity333
    {
        public int Id { get; set; }

        public RelatedEntity332 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity334> ChildEntities { get; set; }
    }

    public class RelatedEntity334
    {
        public int Id { get; set; }

        public RelatedEntity333 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity335> ChildEntities { get; set; }
    }

    public class RelatedEntity335
    {
        public int Id { get; set; }

        public RelatedEntity334 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity336> ChildEntities { get; set; }
    }

    public class RelatedEntity336
    {
        public int Id { get; set; }

        public RelatedEntity335 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity337> ChildEntities { get; set; }
    }

    public class RelatedEntity337
    {
        public int Id { get; set; }

        public RelatedEntity336 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity338> ChildEntities { get; set; }
    }

    public class RelatedEntity338
    {
        public int Id { get; set; }

        public RelatedEntity337 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity339> ChildEntities { get; set; }
    }

    public class RelatedEntity339
    {
        public int Id { get; set; }

        public RelatedEntity338 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity340> ChildEntities { get; set; }
    }

    public class RelatedEntity340
    {
        public int Id { get; set; }

        public RelatedEntity339 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity341> ChildEntities { get; set; }
    }

    public class RelatedEntity341
    {
        public int Id { get; set; }

        public RelatedEntity340 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity342> ChildEntities { get; set; }
    }

    public class RelatedEntity342
    {
        public int Id { get; set; }

        public RelatedEntity341 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity343> ChildEntities { get; set; }
    }

    public class RelatedEntity343
    {
        public int Id { get; set; }

        public RelatedEntity342 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity344> ChildEntities { get; set; }
    }

    public class RelatedEntity344
    {
        public int Id { get; set; }

        public RelatedEntity343 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity345> ChildEntities { get; set; }
    }

    public class RelatedEntity345
    {
        public int Id { get; set; }

        public RelatedEntity344 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity346> ChildEntities { get; set; }
    }

    public class RelatedEntity346
    {
        public int Id { get; set; }

        public RelatedEntity345 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity347> ChildEntities { get; set; }
    }

    public class RelatedEntity347
    {
        public int Id { get; set; }

        public RelatedEntity346 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity348> ChildEntities { get; set; }
    }

    public class RelatedEntity348
    {
        public int Id { get; set; }

        public RelatedEntity347 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity349> ChildEntities { get; set; }
    }

    public class RelatedEntity349
    {
        public int Id { get; set; }

        public RelatedEntity348 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity350> ChildEntities { get; set; }
    }

    public class RelatedEntity350
    {
        public int Id { get; set; }

        public RelatedEntity349 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity351> ChildEntities { get; set; }
    }

    public class RelatedEntity351
    {
        public int Id { get; set; }

        public RelatedEntity350 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity352> ChildEntities { get; set; }
    }

    public class RelatedEntity352
    {
        public int Id { get; set; }

        public RelatedEntity351 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity353> ChildEntities { get; set; }
    }

    public class RelatedEntity353
    {
        public int Id { get; set; }

        public RelatedEntity352 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity354> ChildEntities { get; set; }
    }

    public class RelatedEntity354
    {
        public int Id { get; set; }

        public RelatedEntity353 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity355> ChildEntities { get; set; }
    }

    public class RelatedEntity355
    {
        public int Id { get; set; }

        public RelatedEntity354 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity356> ChildEntities { get; set; }
    }

    public class RelatedEntity356
    {
        public int Id { get; set; }

        public RelatedEntity355 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity357> ChildEntities { get; set; }
    }

    public class RelatedEntity357
    {
        public int Id { get; set; }

        public RelatedEntity356 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity358> ChildEntities { get; set; }
    }

    public class RelatedEntity358
    {
        public int Id { get; set; }

        public RelatedEntity357 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity359> ChildEntities { get; set; }
    }

    public class RelatedEntity359
    {
        public int Id { get; set; }

        public RelatedEntity358 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity360> ChildEntities { get; set; }
    }

    public class RelatedEntity360
    {
        public int Id { get; set; }

        public RelatedEntity359 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity361> ChildEntities { get; set; }
    }

    public class RelatedEntity361
    {
        public int Id { get; set; }

        public RelatedEntity360 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity362> ChildEntities { get; set; }
    }

    public class RelatedEntity362
    {
        public int Id { get; set; }

        public RelatedEntity361 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity363> ChildEntities { get; set; }
    }

    public class RelatedEntity363
    {
        public int Id { get; set; }

        public RelatedEntity362 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity364> ChildEntities { get; set; }
    }

    public class RelatedEntity364
    {
        public int Id { get; set; }

        public RelatedEntity363 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity365> ChildEntities { get; set; }
    }

    public class RelatedEntity365
    {
        public int Id { get; set; }

        public RelatedEntity364 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity366> ChildEntities { get; set; }
    }

    public class RelatedEntity366
    {
        public int Id { get; set; }

        public RelatedEntity365 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity367> ChildEntities { get; set; }
    }

    public class RelatedEntity367
    {
        public int Id { get; set; }

        public RelatedEntity366 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity368> ChildEntities { get; set; }
    }

    public class RelatedEntity368
    {
        public int Id { get; set; }

        public RelatedEntity367 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity369> ChildEntities { get; set; }
    }

    public class RelatedEntity369
    {
        public int Id { get; set; }

        public RelatedEntity368 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity370> ChildEntities { get; set; }
    }

    public class RelatedEntity370
    {
        public int Id { get; set; }

        public RelatedEntity369 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity371> ChildEntities { get; set; }
    }

    public class RelatedEntity371
    {
        public int Id { get; set; }

        public RelatedEntity370 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity372> ChildEntities { get; set; }
    }

    public class RelatedEntity372
    {
        public int Id { get; set; }

        public RelatedEntity371 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity373> ChildEntities { get; set; }
    }

    public class RelatedEntity373
    {
        public int Id { get; set; }

        public RelatedEntity372 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity374> ChildEntities { get; set; }
    }

    public class RelatedEntity374
    {
        public int Id { get; set; }

        public RelatedEntity373 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity375> ChildEntities { get; set; }
    }

    public class RelatedEntity375
    {
        public int Id { get; set; }

        public RelatedEntity374 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity376> ChildEntities { get; set; }
    }

    public class RelatedEntity376
    {
        public int Id { get; set; }

        public RelatedEntity375 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity377> ChildEntities { get; set; }
    }

    public class RelatedEntity377
    {
        public int Id { get; set; }

        public RelatedEntity376 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity378> ChildEntities { get; set; }
    }

    public class RelatedEntity378
    {
        public int Id { get; set; }

        public RelatedEntity377 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity379> ChildEntities { get; set; }
    }

    public class RelatedEntity379
    {
        public int Id { get; set; }

        public RelatedEntity378 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity380> ChildEntities { get; set; }
    }

    public class RelatedEntity380
    {
        public int Id { get; set; }

        public RelatedEntity379 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity381> ChildEntities { get; set; }
    }

    public class RelatedEntity381
    {
        public int Id { get; set; }

        public RelatedEntity380 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity382> ChildEntities { get; set; }
    }

    public class RelatedEntity382
    {
        public int Id { get; set; }

        public RelatedEntity381 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity383> ChildEntities { get; set; }
    }

    public class RelatedEntity383
    {
        public int Id { get; set; }

        public RelatedEntity382 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity384> ChildEntities { get; set; }
    }

    public class RelatedEntity384
    {
        public int Id { get; set; }

        public RelatedEntity383 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity385> ChildEntities { get; set; }
    }

    public class RelatedEntity385
    {
        public int Id { get; set; }

        public RelatedEntity384 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity386> ChildEntities { get; set; }
    }

    public class RelatedEntity386
    {
        public int Id { get; set; }

        public RelatedEntity385 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity387> ChildEntities { get; set; }
    }

    public class RelatedEntity387
    {
        public int Id { get; set; }

        public RelatedEntity386 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity388> ChildEntities { get; set; }
    }

    public class RelatedEntity388
    {
        public int Id { get; set; }

        public RelatedEntity387 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity389> ChildEntities { get; set; }
    }

    public class RelatedEntity389
    {
        public int Id { get; set; }

        public RelatedEntity388 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity390> ChildEntities { get; set; }
    }

    public class RelatedEntity390
    {
        public int Id { get; set; }

        public RelatedEntity389 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity391> ChildEntities { get; set; }
    }

    public class RelatedEntity391
    {
        public int Id { get; set; }

        public RelatedEntity390 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity392> ChildEntities { get; set; }
    }

    public class RelatedEntity392
    {
        public int Id { get; set; }

        public RelatedEntity391 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity393> ChildEntities { get; set; }
    }

    public class RelatedEntity393
    {
        public int Id { get; set; }

        public RelatedEntity392 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity394> ChildEntities { get; set; }
    }

    public class RelatedEntity394
    {
        public int Id { get; set; }

        public RelatedEntity393 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity395> ChildEntities { get; set; }
    }

    public class RelatedEntity395
    {
        public int Id { get; set; }

        public RelatedEntity394 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity396> ChildEntities { get; set; }
    }

    public class RelatedEntity396
    {
        public int Id { get; set; }

        public RelatedEntity395 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity397> ChildEntities { get; set; }
    }

    public class RelatedEntity397
    {
        public int Id { get; set; }

        public RelatedEntity396 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity398> ChildEntities { get; set; }
    }

    public class RelatedEntity398
    {
        public int Id { get; set; }

        public RelatedEntity397 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity399> ChildEntities { get; set; }
    }

    public class RelatedEntity399
    {
        public int Id { get; set; }

        public RelatedEntity398 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity400> ChildEntities { get; set; }
    }

    public class RelatedEntity400
    {
        public int Id { get; set; }

        public RelatedEntity399 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity401> ChildEntities { get; set; }
    }

    public class RelatedEntity401
    {
        public int Id { get; set; }

        public RelatedEntity400 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity402> ChildEntities { get; set; }
    }

    public class RelatedEntity402
    {
        public int Id { get; set; }

        public RelatedEntity401 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity403> ChildEntities { get; set; }
    }

    public class RelatedEntity403
    {
        public int Id { get; set; }

        public RelatedEntity402 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity404> ChildEntities { get; set; }
    }

    public class RelatedEntity404
    {
        public int Id { get; set; }

        public RelatedEntity403 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity405> ChildEntities { get; set; }
    }

    public class RelatedEntity405
    {
        public int Id { get; set; }

        public RelatedEntity404 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity406> ChildEntities { get; set; }
    }

    public class RelatedEntity406
    {
        public int Id { get; set; }

        public RelatedEntity405 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity407> ChildEntities { get; set; }
    }

    public class RelatedEntity407
    {
        public int Id { get; set; }

        public RelatedEntity406 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity408> ChildEntities { get; set; }
    }

    public class RelatedEntity408
    {
        public int Id { get; set; }

        public RelatedEntity407 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity409> ChildEntities { get; set; }
    }

    public class RelatedEntity409
    {
        public int Id { get; set; }

        public RelatedEntity408 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity410> ChildEntities { get; set; }
    }

    public class RelatedEntity410
    {
        public int Id { get; set; }

        public RelatedEntity409 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity411> ChildEntities { get; set; }
    }

    public class RelatedEntity411
    {
        public int Id { get; set; }

        public RelatedEntity410 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity412> ChildEntities { get; set; }
    }

    public class RelatedEntity412
    {
        public int Id { get; set; }

        public RelatedEntity411 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity413> ChildEntities { get; set; }
    }

    public class RelatedEntity413
    {
        public int Id { get; set; }

        public RelatedEntity412 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity414> ChildEntities { get; set; }
    }

    public class RelatedEntity414
    {
        public int Id { get; set; }

        public RelatedEntity413 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity415> ChildEntities { get; set; }
    }

    public class RelatedEntity415
    {
        public int Id { get; set; }

        public RelatedEntity414 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity416> ChildEntities { get; set; }
    }

    public class RelatedEntity416
    {
        public int Id { get; set; }

        public RelatedEntity415 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity417> ChildEntities { get; set; }
    }

    public class RelatedEntity417
    {
        public int Id { get; set; }

        public RelatedEntity416 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity418> ChildEntities { get; set; }
    }

    public class RelatedEntity418
    {
        public int Id { get; set; }

        public RelatedEntity417 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity419> ChildEntities { get; set; }
    }

    public class RelatedEntity419
    {
        public int Id { get; set; }

        public RelatedEntity418 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity420> ChildEntities { get; set; }
    }

    public class RelatedEntity420
    {
        public int Id { get; set; }

        public RelatedEntity419 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity421> ChildEntities { get; set; }
    }

    public class RelatedEntity421
    {
        public int Id { get; set; }

        public RelatedEntity420 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity422> ChildEntities { get; set; }
    }

    public class RelatedEntity422
    {
        public int Id { get; set; }

        public RelatedEntity421 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity423> ChildEntities { get; set; }
    }

    public class RelatedEntity423
    {
        public int Id { get; set; }

        public RelatedEntity422 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity424> ChildEntities { get; set; }
    }

    public class RelatedEntity424
    {
        public int Id { get; set; }

        public RelatedEntity423 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity425> ChildEntities { get; set; }
    }

    public class RelatedEntity425
    {
        public int Id { get; set; }

        public RelatedEntity424 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity426> ChildEntities { get; set; }
    }

    public class RelatedEntity426
    {
        public int Id { get; set; }

        public RelatedEntity425 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity427> ChildEntities { get; set; }
    }

    public class RelatedEntity427
    {
        public int Id { get; set; }

        public RelatedEntity426 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity428> ChildEntities { get; set; }
    }

    public class RelatedEntity428
    {
        public int Id { get; set; }

        public RelatedEntity427 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity429> ChildEntities { get; set; }
    }

    public class RelatedEntity429
    {
        public int Id { get; set; }

        public RelatedEntity428 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity430> ChildEntities { get; set; }
    }

    public class RelatedEntity430
    {
        public int Id { get; set; }

        public RelatedEntity429 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity431> ChildEntities { get; set; }
    }

    public class RelatedEntity431
    {
        public int Id { get; set; }

        public RelatedEntity430 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity432> ChildEntities { get; set; }
    }

    public class RelatedEntity432
    {
        public int Id { get; set; }

        public RelatedEntity431 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity433> ChildEntities { get; set; }
    }

    public class RelatedEntity433
    {
        public int Id { get; set; }

        public RelatedEntity432 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity434> ChildEntities { get; set; }
    }

    public class RelatedEntity434
    {
        public int Id { get; set; }

        public RelatedEntity433 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity435> ChildEntities { get; set; }
    }

    public class RelatedEntity435
    {
        public int Id { get; set; }

        public RelatedEntity434 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity436> ChildEntities { get; set; }
    }

    public class RelatedEntity436
    {
        public int Id { get; set; }

        public RelatedEntity435 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity437> ChildEntities { get; set; }
    }

    public class RelatedEntity437
    {
        public int Id { get; set; }

        public RelatedEntity436 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity438> ChildEntities { get; set; }
    }

    public class RelatedEntity438
    {
        public int Id { get; set; }

        public RelatedEntity437 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity439> ChildEntities { get; set; }
    }

    public class RelatedEntity439
    {
        public int Id { get; set; }

        public RelatedEntity438 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity440> ChildEntities { get; set; }
    }

    public class RelatedEntity440
    {
        public int Id { get; set; }

        public RelatedEntity439 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity441> ChildEntities { get; set; }
    }

    public class RelatedEntity441
    {
        public int Id { get; set; }

        public RelatedEntity440 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity442> ChildEntities { get; set; }
    }

    public class RelatedEntity442
    {
        public int Id { get; set; }

        public RelatedEntity441 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity443> ChildEntities { get; set; }
    }

    public class RelatedEntity443
    {
        public int Id { get; set; }

        public RelatedEntity442 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity444> ChildEntities { get; set; }
    }

    public class RelatedEntity444
    {
        public int Id { get; set; }

        public RelatedEntity443 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity445> ChildEntities { get; set; }
    }

    public class RelatedEntity445
    {
        public int Id { get; set; }

        public RelatedEntity444 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity446> ChildEntities { get; set; }
    }

    public class RelatedEntity446
    {
        public int Id { get; set; }

        public RelatedEntity445 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity447> ChildEntities { get; set; }
    }

    public class RelatedEntity447
    {
        public int Id { get; set; }

        public RelatedEntity446 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity448> ChildEntities { get; set; }
    }

    public class RelatedEntity448
    {
        public int Id { get; set; }

        public RelatedEntity447 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity449> ChildEntities { get; set; }
    }

    public class RelatedEntity449
    {
        public int Id { get; set; }

        public RelatedEntity448 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity450> ChildEntities { get; set; }
    }

    public class RelatedEntity450
    {
        public int Id { get; set; }

        public RelatedEntity449 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity451> ChildEntities { get; set; }
    }

    public class RelatedEntity451
    {
        public int Id { get; set; }

        public RelatedEntity450 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity452> ChildEntities { get; set; }
    }

    public class RelatedEntity452
    {
        public int Id { get; set; }

        public RelatedEntity451 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity453> ChildEntities { get; set; }
    }

    public class RelatedEntity453
    {
        public int Id { get; set; }

        public RelatedEntity452 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity454> ChildEntities { get; set; }
    }

    public class RelatedEntity454
    {
        public int Id { get; set; }

        public RelatedEntity453 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity455> ChildEntities { get; set; }
    }

    public class RelatedEntity455
    {
        public int Id { get; set; }

        public RelatedEntity454 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity456> ChildEntities { get; set; }
    }

    public class RelatedEntity456
    {
        public int Id { get; set; }

        public RelatedEntity455 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity457> ChildEntities { get; set; }
    }

    public class RelatedEntity457
    {
        public int Id { get; set; }

        public RelatedEntity456 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity458> ChildEntities { get; set; }
    }

    public class RelatedEntity458
    {
        public int Id { get; set; }

        public RelatedEntity457 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity459> ChildEntities { get; set; }
    }

    public class RelatedEntity459
    {
        public int Id { get; set; }

        public RelatedEntity458 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity460> ChildEntities { get; set; }
    }

    public class RelatedEntity460
    {
        public int Id { get; set; }

        public RelatedEntity459 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity461> ChildEntities { get; set; }
    }

    public class RelatedEntity461
    {
        public int Id { get; set; }

        public RelatedEntity460 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity462> ChildEntities { get; set; }
    }

    public class RelatedEntity462
    {
        public int Id { get; set; }

        public RelatedEntity461 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity463> ChildEntities { get; set; }
    }

    public class RelatedEntity463
    {
        public int Id { get; set; }

        public RelatedEntity462 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity464> ChildEntities { get; set; }
    }

    public class RelatedEntity464
    {
        public int Id { get; set; }

        public RelatedEntity463 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity465> ChildEntities { get; set; }
    }

    public class RelatedEntity465
    {
        public int Id { get; set; }

        public RelatedEntity464 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity466> ChildEntities { get; set; }
    }

    public class RelatedEntity466
    {
        public int Id { get; set; }

        public RelatedEntity465 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity467> ChildEntities { get; set; }
    }

    public class RelatedEntity467
    {
        public int Id { get; set; }

        public RelatedEntity466 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity468> ChildEntities { get; set; }
    }

    public class RelatedEntity468
    {
        public int Id { get; set; }

        public RelatedEntity467 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity469> ChildEntities { get; set; }
    }

    public class RelatedEntity469
    {
        public int Id { get; set; }

        public RelatedEntity468 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity470> ChildEntities { get; set; }
    }

    public class RelatedEntity470
    {
        public int Id { get; set; }

        public RelatedEntity469 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity471> ChildEntities { get; set; }
    }

    public class RelatedEntity471
    {
        public int Id { get; set; }

        public RelatedEntity470 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity472> ChildEntities { get; set; }
    }

    public class RelatedEntity472
    {
        public int Id { get; set; }

        public RelatedEntity471 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity473> ChildEntities { get; set; }
    }

    public class RelatedEntity473
    {
        public int Id { get; set; }

        public RelatedEntity472 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity474> ChildEntities { get; set; }
    }

    public class RelatedEntity474
    {
        public int Id { get; set; }

        public RelatedEntity473 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity475> ChildEntities { get; set; }
    }

    public class RelatedEntity475
    {
        public int Id { get; set; }

        public RelatedEntity474 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity476> ChildEntities { get; set; }
    }

    public class RelatedEntity476
    {
        public int Id { get; set; }

        public RelatedEntity475 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity477> ChildEntities { get; set; }
    }

    public class RelatedEntity477
    {
        public int Id { get; set; }

        public RelatedEntity476 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity478> ChildEntities { get; set; }
    }

    public class RelatedEntity478
    {
        public int Id { get; set; }

        public RelatedEntity477 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity479> ChildEntities { get; set; }
    }

    public class RelatedEntity479
    {
        public int Id { get; set; }

        public RelatedEntity478 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity480> ChildEntities { get; set; }
    }

    public class RelatedEntity480
    {
        public int Id { get; set; }

        public RelatedEntity479 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity481> ChildEntities { get; set; }
    }

    public class RelatedEntity481
    {
        public int Id { get; set; }

        public RelatedEntity480 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity482> ChildEntities { get; set; }
    }

    public class RelatedEntity482
    {
        public int Id { get; set; }

        public RelatedEntity481 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity483> ChildEntities { get; set; }
    }

    public class RelatedEntity483
    {
        public int Id { get; set; }

        public RelatedEntity482 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity484> ChildEntities { get; set; }
    }

    public class RelatedEntity484
    {
        public int Id { get; set; }

        public RelatedEntity483 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity485> ChildEntities { get; set; }
    }

    public class RelatedEntity485
    {
        public int Id { get; set; }

        public RelatedEntity484 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity486> ChildEntities { get; set; }
    }

    public class RelatedEntity486
    {
        public int Id { get; set; }

        public RelatedEntity485 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity487> ChildEntities { get; set; }
    }

    public class RelatedEntity487
    {
        public int Id { get; set; }

        public RelatedEntity486 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity488> ChildEntities { get; set; }
    }

    public class RelatedEntity488
    {
        public int Id { get; set; }

        public RelatedEntity487 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity489> ChildEntities { get; set; }
    }

    public class RelatedEntity489
    {
        public int Id { get; set; }

        public RelatedEntity488 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity490> ChildEntities { get; set; }
    }

    public class RelatedEntity490
    {
        public int Id { get; set; }

        public RelatedEntity489 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity491> ChildEntities { get; set; }
    }

    public class RelatedEntity491
    {
        public int Id { get; set; }

        public RelatedEntity490 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity492> ChildEntities { get; set; }
    }

    public class RelatedEntity492
    {
        public int Id { get; set; }

        public RelatedEntity491 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity493> ChildEntities { get; set; }
    }

    public class RelatedEntity493
    {
        public int Id { get; set; }

        public RelatedEntity492 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity494> ChildEntities { get; set; }
    }

    public class RelatedEntity494
    {
        public int Id { get; set; }

        public RelatedEntity493 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity495> ChildEntities { get; set; }
    }

    public class RelatedEntity495
    {
        public int Id { get; set; }

        public RelatedEntity494 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity496> ChildEntities { get; set; }
    }

    public class RelatedEntity496
    {
        public int Id { get; set; }

        public RelatedEntity495 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity497> ChildEntities { get; set; }
    }

    public class RelatedEntity497
    {
        public int Id { get; set; }

        public RelatedEntity496 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity498> ChildEntities { get; set; }
    }

    public class RelatedEntity498
    {
        public int Id { get; set; }

        public RelatedEntity497 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity499> ChildEntities { get; set; }
    }

    public class RelatedEntity499
    {
        public int Id { get; set; }

        public RelatedEntity498 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity500> ChildEntities { get; set; }
    }

    public class RelatedEntity500
    {
        public int Id { get; set; }

        public RelatedEntity499 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity501> ChildEntities { get; set; }
    }

    public class RelatedEntity501
    {
        public int Id { get; set; }

        public RelatedEntity500 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity502> ChildEntities { get; set; }
    }

    public class RelatedEntity502
    {
        public int Id { get; set; }

        public RelatedEntity501 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity503> ChildEntities { get; set; }
    }

    public class RelatedEntity503
    {
        public int Id { get; set; }

        public RelatedEntity502 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity504> ChildEntities { get; set; }
    }

    public class RelatedEntity504
    {
        public int Id { get; set; }

        public RelatedEntity503 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity505> ChildEntities { get; set; }
    }

    public class RelatedEntity505
    {
        public int Id { get; set; }

        public RelatedEntity504 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity506> ChildEntities { get; set; }
    }

    public class RelatedEntity506
    {
        public int Id { get; set; }

        public RelatedEntity505 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity507> ChildEntities { get; set; }
    }

    public class RelatedEntity507
    {
        public int Id { get; set; }

        public RelatedEntity506 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity508> ChildEntities { get; set; }
    }

    public class RelatedEntity508
    {
        public int Id { get; set; }

        public RelatedEntity507 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity509> ChildEntities { get; set; }
    }

    public class RelatedEntity509
    {
        public int Id { get; set; }

        public RelatedEntity508 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity510> ChildEntities { get; set; }
    }

    public class RelatedEntity510
    {
        public int Id { get; set; }

        public RelatedEntity509 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity511> ChildEntities { get; set; }
    }

    public class RelatedEntity511
    {
        public int Id { get; set; }

        public RelatedEntity510 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity512> ChildEntities { get; set; }
    }

    public class RelatedEntity512
    {
        public int Id { get; set; }

        public RelatedEntity511 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity513> ChildEntities { get; set; }
    }

    public class RelatedEntity513
    {
        public int Id { get; set; }

        public RelatedEntity512 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity514> ChildEntities { get; set; }
    }

    public class RelatedEntity514
    {
        public int Id { get; set; }

        public RelatedEntity513 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity515> ChildEntities { get; set; }
    }

    public class RelatedEntity515
    {
        public int Id { get; set; }

        public RelatedEntity514 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity516> ChildEntities { get; set; }
    }

    public class RelatedEntity516
    {
        public int Id { get; set; }

        public RelatedEntity515 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity517> ChildEntities { get; set; }
    }

    public class RelatedEntity517
    {
        public int Id { get; set; }

        public RelatedEntity516 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity518> ChildEntities { get; set; }
    }

    public class RelatedEntity518
    {
        public int Id { get; set; }

        public RelatedEntity517 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity519> ChildEntities { get; set; }
    }

    public class RelatedEntity519
    {
        public int Id { get; set; }

        public RelatedEntity518 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity520> ChildEntities { get; set; }
    }

    public class RelatedEntity520
    {
        public int Id { get; set; }

        public RelatedEntity519 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity521> ChildEntities { get; set; }
    }

    public class RelatedEntity521
    {
        public int Id { get; set; }

        public RelatedEntity520 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity522> ChildEntities { get; set; }
    }

    public class RelatedEntity522
    {
        public int Id { get; set; }

        public RelatedEntity521 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity523> ChildEntities { get; set; }
    }

    public class RelatedEntity523
    {
        public int Id { get; set; }

        public RelatedEntity522 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity524> ChildEntities { get; set; }
    }

    public class RelatedEntity524
    {
        public int Id { get; set; }

        public RelatedEntity523 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity525> ChildEntities { get; set; }
    }

    public class RelatedEntity525
    {
        public int Id { get; set; }

        public RelatedEntity524 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity526> ChildEntities { get; set; }
    }

    public class RelatedEntity526
    {
        public int Id { get; set; }

        public RelatedEntity525 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity527> ChildEntities { get; set; }
    }

    public class RelatedEntity527
    {
        public int Id { get; set; }

        public RelatedEntity526 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity528> ChildEntities { get; set; }
    }

    public class RelatedEntity528
    {
        public int Id { get; set; }

        public RelatedEntity527 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity529> ChildEntities { get; set; }
    }

    public class RelatedEntity529
    {
        public int Id { get; set; }

        public RelatedEntity528 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity530> ChildEntities { get; set; }
    }

    public class RelatedEntity530
    {
        public int Id { get; set; }

        public RelatedEntity529 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity531> ChildEntities { get; set; }
    }

    public class RelatedEntity531
    {
        public int Id { get; set; }

        public RelatedEntity530 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity532> ChildEntities { get; set; }
    }

    public class RelatedEntity532
    {
        public int Id { get; set; }

        public RelatedEntity531 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity533> ChildEntities { get; set; }
    }

    public class RelatedEntity533
    {
        public int Id { get; set; }

        public RelatedEntity532 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity534> ChildEntities { get; set; }
    }

    public class RelatedEntity534
    {
        public int Id { get; set; }

        public RelatedEntity533 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity535> ChildEntities { get; set; }
    }

    public class RelatedEntity535
    {
        public int Id { get; set; }

        public RelatedEntity534 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity536> ChildEntities { get; set; }
    }

    public class RelatedEntity536
    {
        public int Id { get; set; }

        public RelatedEntity535 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity537> ChildEntities { get; set; }
    }

    public class RelatedEntity537
    {
        public int Id { get; set; }

        public RelatedEntity536 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity538> ChildEntities { get; set; }
    }

    public class RelatedEntity538
    {
        public int Id { get; set; }

        public RelatedEntity537 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity539> ChildEntities { get; set; }
    }

    public class RelatedEntity539
    {
        public int Id { get; set; }

        public RelatedEntity538 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity540> ChildEntities { get; set; }
    }

    public class RelatedEntity540
    {
        public int Id { get; set; }

        public RelatedEntity539 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity541> ChildEntities { get; set; }
    }

    public class RelatedEntity541
    {
        public int Id { get; set; }

        public RelatedEntity540 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity542> ChildEntities { get; set; }
    }

    public class RelatedEntity542
    {
        public int Id { get; set; }

        public RelatedEntity541 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity543> ChildEntities { get; set; }
    }

    public class RelatedEntity543
    {
        public int Id { get; set; }

        public RelatedEntity542 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity544> ChildEntities { get; set; }
    }

    public class RelatedEntity544
    {
        public int Id { get; set; }

        public RelatedEntity543 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity545> ChildEntities { get; set; }
    }

    public class RelatedEntity545
    {
        public int Id { get; set; }

        public RelatedEntity544 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity546> ChildEntities { get; set; }
    }

    public class RelatedEntity546
    {
        public int Id { get; set; }

        public RelatedEntity545 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity547> ChildEntities { get; set; }
    }

    public class RelatedEntity547
    {
        public int Id { get; set; }

        public RelatedEntity546 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity548> ChildEntities { get; set; }
    }

    public class RelatedEntity548
    {
        public int Id { get; set; }

        public RelatedEntity547 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity549> ChildEntities { get; set; }
    }

    public class RelatedEntity549
    {
        public int Id { get; set; }

        public RelatedEntity548 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity550> ChildEntities { get; set; }
    }

    public class RelatedEntity550
    {
        public int Id { get; set; }

        public RelatedEntity549 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity551> ChildEntities { get; set; }
    }

    public class RelatedEntity551
    {
        public int Id { get; set; }

        public RelatedEntity550 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity552> ChildEntities { get; set; }
    }

    public class RelatedEntity552
    {
        public int Id { get; set; }

        public RelatedEntity551 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity553> ChildEntities { get; set; }
    }

    public class RelatedEntity553
    {
        public int Id { get; set; }

        public RelatedEntity552 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity554> ChildEntities { get; set; }
    }

    public class RelatedEntity554
    {
        public int Id { get; set; }

        public RelatedEntity553 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity555> ChildEntities { get; set; }
    }

    public class RelatedEntity555
    {
        public int Id { get; set; }

        public RelatedEntity554 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity556> ChildEntities { get; set; }
    }

    public class RelatedEntity556
    {
        public int Id { get; set; }

        public RelatedEntity555 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity557> ChildEntities { get; set; }
    }

    public class RelatedEntity557
    {
        public int Id { get; set; }

        public RelatedEntity556 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity558> ChildEntities { get; set; }
    }

    public class RelatedEntity558
    {
        public int Id { get; set; }

        public RelatedEntity557 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity559> ChildEntities { get; set; }
    }

    public class RelatedEntity559
    {
        public int Id { get; set; }

        public RelatedEntity558 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity560> ChildEntities { get; set; }
    }

    public class RelatedEntity560
    {
        public int Id { get; set; }

        public RelatedEntity559 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity561> ChildEntities { get; set; }
    }

    public class RelatedEntity561
    {
        public int Id { get; set; }

        public RelatedEntity560 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity562> ChildEntities { get; set; }
    }

    public class RelatedEntity562
    {
        public int Id { get; set; }

        public RelatedEntity561 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity563> ChildEntities { get; set; }
    }

    public class RelatedEntity563
    {
        public int Id { get; set; }

        public RelatedEntity562 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity564> ChildEntities { get; set; }
    }

    public class RelatedEntity564
    {
        public int Id { get; set; }

        public RelatedEntity563 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity565> ChildEntities { get; set; }
    }

    public class RelatedEntity565
    {
        public int Id { get; set; }

        public RelatedEntity564 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity566> ChildEntities { get; set; }
    }

    public class RelatedEntity566
    {
        public int Id { get; set; }

        public RelatedEntity565 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity567> ChildEntities { get; set; }
    }

    public class RelatedEntity567
    {
        public int Id { get; set; }

        public RelatedEntity566 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity568> ChildEntities { get; set; }
    }

    public class RelatedEntity568
    {
        public int Id { get; set; }

        public RelatedEntity567 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity569> ChildEntities { get; set; }
    }

    public class RelatedEntity569
    {
        public int Id { get; set; }

        public RelatedEntity568 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity570> ChildEntities { get; set; }
    }

    public class RelatedEntity570
    {
        public int Id { get; set; }

        public RelatedEntity569 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity571> ChildEntities { get; set; }
    }

    public class RelatedEntity571
    {
        public int Id { get; set; }

        public RelatedEntity570 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity572> ChildEntities { get; set; }
    }

    public class RelatedEntity572
    {
        public int Id { get; set; }

        public RelatedEntity571 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity573> ChildEntities { get; set; }
    }

    public class RelatedEntity573
    {
        public int Id { get; set; }

        public RelatedEntity572 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity574> ChildEntities { get; set; }
    }

    public class RelatedEntity574
    {
        public int Id { get; set; }

        public RelatedEntity573 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity575> ChildEntities { get; set; }
    }

    public class RelatedEntity575
    {
        public int Id { get; set; }

        public RelatedEntity574 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity576> ChildEntities { get; set; }
    }

    public class RelatedEntity576
    {
        public int Id { get; set; }

        public RelatedEntity575 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity577> ChildEntities { get; set; }
    }

    public class RelatedEntity577
    {
        public int Id { get; set; }

        public RelatedEntity576 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity578> ChildEntities { get; set; }
    }

    public class RelatedEntity578
    {
        public int Id { get; set; }

        public RelatedEntity577 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity579> ChildEntities { get; set; }
    }

    public class RelatedEntity579
    {
        public int Id { get; set; }

        public RelatedEntity578 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity580> ChildEntities { get; set; }
    }

    public class RelatedEntity580
    {
        public int Id { get; set; }

        public RelatedEntity579 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity581> ChildEntities { get; set; }
    }

    public class RelatedEntity581
    {
        public int Id { get; set; }

        public RelatedEntity580 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity582> ChildEntities { get; set; }
    }

    public class RelatedEntity582
    {
        public int Id { get; set; }

        public RelatedEntity581 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity583> ChildEntities { get; set; }
    }

    public class RelatedEntity583
    {
        public int Id { get; set; }

        public RelatedEntity582 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity584> ChildEntities { get; set; }
    }

    public class RelatedEntity584
    {
        public int Id { get; set; }

        public RelatedEntity583 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity585> ChildEntities { get; set; }
    }

    public class RelatedEntity585
    {
        public int Id { get; set; }

        public RelatedEntity584 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity586> ChildEntities { get; set; }
    }

    public class RelatedEntity586
    {
        public int Id { get; set; }

        public RelatedEntity585 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity587> ChildEntities { get; set; }
    }

    public class RelatedEntity587
    {
        public int Id { get; set; }

        public RelatedEntity586 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity588> ChildEntities { get; set; }
    }

    public class RelatedEntity588
    {
        public int Id { get; set; }

        public RelatedEntity587 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity589> ChildEntities { get; set; }
    }

    public class RelatedEntity589
    {
        public int Id { get; set; }

        public RelatedEntity588 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity590> ChildEntities { get; set; }
    }

    public class RelatedEntity590
    {
        public int Id { get; set; }

        public RelatedEntity589 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity591> ChildEntities { get; set; }
    }

    public class RelatedEntity591
    {
        public int Id { get; set; }

        public RelatedEntity590 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity592> ChildEntities { get; set; }
    }

    public class RelatedEntity592
    {
        public int Id { get; set; }

        public RelatedEntity591 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity593> ChildEntities { get; set; }
    }

    public class RelatedEntity593
    {
        public int Id { get; set; }

        public RelatedEntity592 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity594> ChildEntities { get; set; }
    }

    public class RelatedEntity594
    {
        public int Id { get; set; }

        public RelatedEntity593 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity595> ChildEntities { get; set; }
    }

    public class RelatedEntity595
    {
        public int Id { get; set; }

        public RelatedEntity594 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity596> ChildEntities { get; set; }
    }

    public class RelatedEntity596
    {
        public int Id { get; set; }

        public RelatedEntity595 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity597> ChildEntities { get; set; }
    }

    public class RelatedEntity597
    {
        public int Id { get; set; }

        public RelatedEntity596 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity598> ChildEntities { get; set; }
    }

    public class RelatedEntity598
    {
        public int Id { get; set; }

        public RelatedEntity597 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity599> ChildEntities { get; set; }
    }

    public class RelatedEntity599
    {
        public int Id { get; set; }

        public RelatedEntity598 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity600> ChildEntities { get; set; }
    }

    public class RelatedEntity600
    {
        public int Id { get; set; }

        public RelatedEntity599 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity601> ChildEntities { get; set; }
    }

    public class RelatedEntity601
    {
        public int Id { get; set; }

        public RelatedEntity600 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity602> ChildEntities { get; set; }
    }

    public class RelatedEntity602
    {
        public int Id { get; set; }

        public RelatedEntity601 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity603> ChildEntities { get; set; }
    }

    public class RelatedEntity603
    {
        public int Id { get; set; }

        public RelatedEntity602 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity604> ChildEntities { get; set; }
    }

    public class RelatedEntity604
    {
        public int Id { get; set; }

        public RelatedEntity603 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity605> ChildEntities { get; set; }
    }

    public class RelatedEntity605
    {
        public int Id { get; set; }

        public RelatedEntity604 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity606> ChildEntities { get; set; }
    }

    public class RelatedEntity606
    {
        public int Id { get; set; }

        public RelatedEntity605 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity607> ChildEntities { get; set; }
    }

    public class RelatedEntity607
    {
        public int Id { get; set; }

        public RelatedEntity606 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity608> ChildEntities { get; set; }
    }

    public class RelatedEntity608
    {
        public int Id { get; set; }

        public RelatedEntity607 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity609> ChildEntities { get; set; }
    }

    public class RelatedEntity609
    {
        public int Id { get; set; }

        public RelatedEntity608 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity610> ChildEntities { get; set; }
    }

    public class RelatedEntity610
    {
        public int Id { get; set; }

        public RelatedEntity609 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity611> ChildEntities { get; set; }
    }

    public class RelatedEntity611
    {
        public int Id { get; set; }

        public RelatedEntity610 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity612> ChildEntities { get; set; }
    }

    public class RelatedEntity612
    {
        public int Id { get; set; }

        public RelatedEntity611 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity613> ChildEntities { get; set; }
    }

    public class RelatedEntity613
    {
        public int Id { get; set; }

        public RelatedEntity612 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity614> ChildEntities { get; set; }
    }

    public class RelatedEntity614
    {
        public int Id { get; set; }

        public RelatedEntity613 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity615> ChildEntities { get; set; }
    }

    public class RelatedEntity615
    {
        public int Id { get; set; }

        public RelatedEntity614 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity616> ChildEntities { get; set; }
    }

    public class RelatedEntity616
    {
        public int Id { get; set; }

        public RelatedEntity615 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity617> ChildEntities { get; set; }
    }

    public class RelatedEntity617
    {
        public int Id { get; set; }

        public RelatedEntity616 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity618> ChildEntities { get; set; }
    }

    public class RelatedEntity618
    {
        public int Id { get; set; }

        public RelatedEntity617 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity619> ChildEntities { get; set; }
    }

    public class RelatedEntity619
    {
        public int Id { get; set; }

        public RelatedEntity618 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity620> ChildEntities { get; set; }
    }

    public class RelatedEntity620
    {
        public int Id { get; set; }

        public RelatedEntity619 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity621> ChildEntities { get; set; }
    }

    public class RelatedEntity621
    {
        public int Id { get; set; }

        public RelatedEntity620 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity622> ChildEntities { get; set; }
    }

    public class RelatedEntity622
    {
        public int Id { get; set; }

        public RelatedEntity621 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity623> ChildEntities { get; set; }
    }

    public class RelatedEntity623
    {
        public int Id { get; set; }

        public RelatedEntity622 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity624> ChildEntities { get; set; }
    }

    public class RelatedEntity624
    {
        public int Id { get; set; }

        public RelatedEntity623 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity625> ChildEntities { get; set; }
    }

    public class RelatedEntity625
    {
        public int Id { get; set; }

        public RelatedEntity624 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity626> ChildEntities { get; set; }
    }

    public class RelatedEntity626
    {
        public int Id { get; set; }

        public RelatedEntity625 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity627> ChildEntities { get; set; }
    }

    public class RelatedEntity627
    {
        public int Id { get; set; }

        public RelatedEntity626 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity628> ChildEntities { get; set; }
    }

    public class RelatedEntity628
    {
        public int Id { get; set; }

        public RelatedEntity627 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity629> ChildEntities { get; set; }
    }

    public class RelatedEntity629
    {
        public int Id { get; set; }

        public RelatedEntity628 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity630> ChildEntities { get; set; }
    }

    public class RelatedEntity630
    {
        public int Id { get; set; }

        public RelatedEntity629 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity631> ChildEntities { get; set; }
    }

    public class RelatedEntity631
    {
        public int Id { get; set; }

        public RelatedEntity630 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity632> ChildEntities { get; set; }
    }

    public class RelatedEntity632
    {
        public int Id { get; set; }

        public RelatedEntity631 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity633> ChildEntities { get; set; }
    }

    public class RelatedEntity633
    {
        public int Id { get; set; }

        public RelatedEntity632 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity634> ChildEntities { get; set; }
    }

    public class RelatedEntity634
    {
        public int Id { get; set; }

        public RelatedEntity633 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity635> ChildEntities { get; set; }
    }

    public class RelatedEntity635
    {
        public int Id { get; set; }

        public RelatedEntity634 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity636> ChildEntities { get; set; }
    }

    public class RelatedEntity636
    {
        public int Id { get; set; }

        public RelatedEntity635 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity637> ChildEntities { get; set; }
    }

    public class RelatedEntity637
    {
        public int Id { get; set; }

        public RelatedEntity636 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity638> ChildEntities { get; set; }
    }

    public class RelatedEntity638
    {
        public int Id { get; set; }

        public RelatedEntity637 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity639> ChildEntities { get; set; }
    }

    public class RelatedEntity639
    {
        public int Id { get; set; }

        public RelatedEntity638 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity640> ChildEntities { get; set; }
    }

    public class RelatedEntity640
    {
        public int Id { get; set; }

        public RelatedEntity639 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity641> ChildEntities { get; set; }
    }

    public class RelatedEntity641
    {
        public int Id { get; set; }

        public RelatedEntity640 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity642> ChildEntities { get; set; }
    }

    public class RelatedEntity642
    {
        public int Id { get; set; }

        public RelatedEntity641 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity643> ChildEntities { get; set; }
    }

    public class RelatedEntity643
    {
        public int Id { get; set; }

        public RelatedEntity642 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity644> ChildEntities { get; set; }
    }

    public class RelatedEntity644
    {
        public int Id { get; set; }

        public RelatedEntity643 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity645> ChildEntities { get; set; }
    }

    public class RelatedEntity645
    {
        public int Id { get; set; }

        public RelatedEntity644 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity646> ChildEntities { get; set; }
    }

    public class RelatedEntity646
    {
        public int Id { get; set; }

        public RelatedEntity645 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity647> ChildEntities { get; set; }
    }

    public class RelatedEntity647
    {
        public int Id { get; set; }

        public RelatedEntity646 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity648> ChildEntities { get; set; }
    }

    public class RelatedEntity648
    {
        public int Id { get; set; }

        public RelatedEntity647 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity649> ChildEntities { get; set; }
    }

    public class RelatedEntity649
    {
        public int Id { get; set; }

        public RelatedEntity648 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity650> ChildEntities { get; set; }
    }

    public class RelatedEntity650
    {
        public int Id { get; set; }

        public RelatedEntity649 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity651> ChildEntities { get; set; }
    }

    public class RelatedEntity651
    {
        public int Id { get; set; }

        public RelatedEntity650 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity652> ChildEntities { get; set; }
    }

    public class RelatedEntity652
    {
        public int Id { get; set; }

        public RelatedEntity651 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity653> ChildEntities { get; set; }
    }

    public class RelatedEntity653
    {
        public int Id { get; set; }

        public RelatedEntity652 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity654> ChildEntities { get; set; }
    }

    public class RelatedEntity654
    {
        public int Id { get; set; }

        public RelatedEntity653 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity655> ChildEntities { get; set; }
    }

    public class RelatedEntity655
    {
        public int Id { get; set; }

        public RelatedEntity654 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity656> ChildEntities { get; set; }
    }

    public class RelatedEntity656
    {
        public int Id { get; set; }

        public RelatedEntity655 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity657> ChildEntities { get; set; }
    }

    public class RelatedEntity657
    {
        public int Id { get; set; }

        public RelatedEntity656 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity658> ChildEntities { get; set; }
    }

    public class RelatedEntity658
    {
        public int Id { get; set; }

        public RelatedEntity657 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity659> ChildEntities { get; set; }
    }

    public class RelatedEntity659
    {
        public int Id { get; set; }

        public RelatedEntity658 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity660> ChildEntities { get; set; }
    }

    public class RelatedEntity660
    {
        public int Id { get; set; }

        public RelatedEntity659 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity661> ChildEntities { get; set; }
    }

    public class RelatedEntity661
    {
        public int Id { get; set; }

        public RelatedEntity660 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity662> ChildEntities { get; set; }
    }

    public class RelatedEntity662
    {
        public int Id { get; set; }

        public RelatedEntity661 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity663> ChildEntities { get; set; }
    }

    public class RelatedEntity663
    {
        public int Id { get; set; }

        public RelatedEntity662 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity664> ChildEntities { get; set; }
    }

    public class RelatedEntity664
    {
        public int Id { get; set; }

        public RelatedEntity663 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity665> ChildEntities { get; set; }
    }

    public class RelatedEntity665
    {
        public int Id { get; set; }

        public RelatedEntity664 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity666> ChildEntities { get; set; }
    }

    public class RelatedEntity666
    {
        public int Id { get; set; }

        public RelatedEntity665 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity667> ChildEntities { get; set; }
    }

    public class RelatedEntity667
    {
        public int Id { get; set; }

        public RelatedEntity666 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity668> ChildEntities { get; set; }
    }

    public class RelatedEntity668
    {
        public int Id { get; set; }

        public RelatedEntity667 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity669> ChildEntities { get; set; }
    }

    public class RelatedEntity669
    {
        public int Id { get; set; }

        public RelatedEntity668 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity670> ChildEntities { get; set; }
    }

    public class RelatedEntity670
    {
        public int Id { get; set; }

        public RelatedEntity669 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity671> ChildEntities { get; set; }
    }

    public class RelatedEntity671
    {
        public int Id { get; set; }

        public RelatedEntity670 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity672> ChildEntities { get; set; }
    }

    public class RelatedEntity672
    {
        public int Id { get; set; }

        public RelatedEntity671 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity673> ChildEntities { get; set; }
    }

    public class RelatedEntity673
    {
        public int Id { get; set; }

        public RelatedEntity672 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity674> ChildEntities { get; set; }
    }

    public class RelatedEntity674
    {
        public int Id { get; set; }

        public RelatedEntity673 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity675> ChildEntities { get; set; }
    }

    public class RelatedEntity675
    {
        public int Id { get; set; }

        public RelatedEntity674 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity676> ChildEntities { get; set; }
    }

    public class RelatedEntity676
    {
        public int Id { get; set; }

        public RelatedEntity675 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity677> ChildEntities { get; set; }
    }

    public class RelatedEntity677
    {
        public int Id { get; set; }

        public RelatedEntity676 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity678> ChildEntities { get; set; }
    }

    public class RelatedEntity678
    {
        public int Id { get; set; }

        public RelatedEntity677 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity679> ChildEntities { get; set; }
    }

    public class RelatedEntity679
    {
        public int Id { get; set; }

        public RelatedEntity678 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity680> ChildEntities { get; set; }
    }

    public class RelatedEntity680
    {
        public int Id { get; set; }

        public RelatedEntity679 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity681> ChildEntities { get; set; }
    }

    public class RelatedEntity681
    {
        public int Id { get; set; }

        public RelatedEntity680 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity682> ChildEntities { get; set; }
    }

    public class RelatedEntity682
    {
        public int Id { get; set; }

        public RelatedEntity681 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity683> ChildEntities { get; set; }
    }

    public class RelatedEntity683
    {
        public int Id { get; set; }

        public RelatedEntity682 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity684> ChildEntities { get; set; }
    }

    public class RelatedEntity684
    {
        public int Id { get; set; }

        public RelatedEntity683 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity685> ChildEntities { get; set; }
    }

    public class RelatedEntity685
    {
        public int Id { get; set; }

        public RelatedEntity684 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity686> ChildEntities { get; set; }
    }

    public class RelatedEntity686
    {
        public int Id { get; set; }

        public RelatedEntity685 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity687> ChildEntities { get; set; }
    }

    public class RelatedEntity687
    {
        public int Id { get; set; }

        public RelatedEntity686 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity688> ChildEntities { get; set; }
    }

    public class RelatedEntity688
    {
        public int Id { get; set; }

        public RelatedEntity687 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity689> ChildEntities { get; set; }
    }

    public class RelatedEntity689
    {
        public int Id { get; set; }

        public RelatedEntity688 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity690> ChildEntities { get; set; }
    }

    public class RelatedEntity690
    {
        public int Id { get; set; }

        public RelatedEntity689 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity691> ChildEntities { get; set; }
    }

    public class RelatedEntity691
    {
        public int Id { get; set; }

        public RelatedEntity690 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity692> ChildEntities { get; set; }
    }

    public class RelatedEntity692
    {
        public int Id { get; set; }

        public RelatedEntity691 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity693> ChildEntities { get; set; }
    }

    public class RelatedEntity693
    {
        public int Id { get; set; }

        public RelatedEntity692 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity694> ChildEntities { get; set; }
    }

    public class RelatedEntity694
    {
        public int Id { get; set; }

        public RelatedEntity693 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity695> ChildEntities { get; set; }
    }

    public class RelatedEntity695
    {
        public int Id { get; set; }

        public RelatedEntity694 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity696> ChildEntities { get; set; }
    }

    public class RelatedEntity696
    {
        public int Id { get; set; }

        public RelatedEntity695 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity697> ChildEntities { get; set; }
    }

    public class RelatedEntity697
    {
        public int Id { get; set; }

        public RelatedEntity696 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity698> ChildEntities { get; set; }
    }

    public class RelatedEntity698
    {
        public int Id { get; set; }

        public RelatedEntity697 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity699> ChildEntities { get; set; }
    }

    public class RelatedEntity699
    {
        public int Id { get; set; }

        public RelatedEntity698 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity700> ChildEntities { get; set; }
    }

    public class RelatedEntity700
    {
        public int Id { get; set; }

        public RelatedEntity699 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity701> ChildEntities { get; set; }
    }

    public class RelatedEntity701
    {
        public int Id { get; set; }

        public RelatedEntity700 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity702> ChildEntities { get; set; }
    }

    public class RelatedEntity702
    {
        public int Id { get; set; }

        public RelatedEntity701 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity703> ChildEntities { get; set; }
    }

    public class RelatedEntity703
    {
        public int Id { get; set; }

        public RelatedEntity702 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity704> ChildEntities { get; set; }
    }

    public class RelatedEntity704
    {
        public int Id { get; set; }

        public RelatedEntity703 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity705> ChildEntities { get; set; }
    }

    public class RelatedEntity705
    {
        public int Id { get; set; }

        public RelatedEntity704 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity706> ChildEntities { get; set; }
    }

    public class RelatedEntity706
    {
        public int Id { get; set; }

        public RelatedEntity705 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity707> ChildEntities { get; set; }
    }

    public class RelatedEntity707
    {
        public int Id { get; set; }

        public RelatedEntity706 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity708> ChildEntities { get; set; }
    }

    public class RelatedEntity708
    {
        public int Id { get; set; }

        public RelatedEntity707 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity709> ChildEntities { get; set; }
    }

    public class RelatedEntity709
    {
        public int Id { get; set; }

        public RelatedEntity708 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity710> ChildEntities { get; set; }
    }

    public class RelatedEntity710
    {
        public int Id { get; set; }

        public RelatedEntity709 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity711> ChildEntities { get; set; }
    }

    public class RelatedEntity711
    {
        public int Id { get; set; }

        public RelatedEntity710 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity712> ChildEntities { get; set; }
    }

    public class RelatedEntity712
    {
        public int Id { get; set; }

        public RelatedEntity711 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity713> ChildEntities { get; set; }
    }

    public class RelatedEntity713
    {
        public int Id { get; set; }

        public RelatedEntity712 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity714> ChildEntities { get; set; }
    }

    public class RelatedEntity714
    {
        public int Id { get; set; }

        public RelatedEntity713 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity715> ChildEntities { get; set; }
    }

    public class RelatedEntity715
    {
        public int Id { get; set; }

        public RelatedEntity714 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity716> ChildEntities { get; set; }
    }

    public class RelatedEntity716
    {
        public int Id { get; set; }

        public RelatedEntity715 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity717> ChildEntities { get; set; }
    }

    public class RelatedEntity717
    {
        public int Id { get; set; }

        public RelatedEntity716 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity718> ChildEntities { get; set; }
    }

    public class RelatedEntity718
    {
        public int Id { get; set; }

        public RelatedEntity717 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity719> ChildEntities { get; set; }
    }

    public class RelatedEntity719
    {
        public int Id { get; set; }

        public RelatedEntity718 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity720> ChildEntities { get; set; }
    }

    public class RelatedEntity720
    {
        public int Id { get; set; }

        public RelatedEntity719 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity721> ChildEntities { get; set; }
    }

    public class RelatedEntity721
    {
        public int Id { get; set; }

        public RelatedEntity720 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity722> ChildEntities { get; set; }
    }

    public class RelatedEntity722
    {
        public int Id { get; set; }

        public RelatedEntity721 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity723> ChildEntities { get; set; }
    }

    public class RelatedEntity723
    {
        public int Id { get; set; }

        public RelatedEntity722 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity724> ChildEntities { get; set; }
    }

    public class RelatedEntity724
    {
        public int Id { get; set; }

        public RelatedEntity723 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity725> ChildEntities { get; set; }
    }

    public class RelatedEntity725
    {
        public int Id { get; set; }

        public RelatedEntity724 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity726> ChildEntities { get; set; }
    }

    public class RelatedEntity726
    {
        public int Id { get; set; }

        public RelatedEntity725 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity727> ChildEntities { get; set; }
    }

    public class RelatedEntity727
    {
        public int Id { get; set; }

        public RelatedEntity726 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity728> ChildEntities { get; set; }
    }

    public class RelatedEntity728
    {
        public int Id { get; set; }

        public RelatedEntity727 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity729> ChildEntities { get; set; }
    }

    public class RelatedEntity729
    {
        public int Id { get; set; }

        public RelatedEntity728 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity730> ChildEntities { get; set; }
    }

    public class RelatedEntity730
    {
        public int Id { get; set; }

        public RelatedEntity729 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity731> ChildEntities { get; set; }
    }

    public class RelatedEntity731
    {
        public int Id { get; set; }

        public RelatedEntity730 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity732> ChildEntities { get; set; }
    }

    public class RelatedEntity732
    {
        public int Id { get; set; }

        public RelatedEntity731 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity733> ChildEntities { get; set; }
    }

    public class RelatedEntity733
    {
        public int Id { get; set; }

        public RelatedEntity732 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity734> ChildEntities { get; set; }
    }

    public class RelatedEntity734
    {
        public int Id { get; set; }

        public RelatedEntity733 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity735> ChildEntities { get; set; }
    }

    public class RelatedEntity735
    {
        public int Id { get; set; }

        public RelatedEntity734 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity736> ChildEntities { get; set; }
    }

    public class RelatedEntity736
    {
        public int Id { get; set; }

        public RelatedEntity735 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity737> ChildEntities { get; set; }
    }

    public class RelatedEntity737
    {
        public int Id { get; set; }

        public RelatedEntity736 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity738> ChildEntities { get; set; }
    }

    public class RelatedEntity738
    {
        public int Id { get; set; }

        public RelatedEntity737 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity739> ChildEntities { get; set; }
    }

    public class RelatedEntity739
    {
        public int Id { get; set; }

        public RelatedEntity738 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity740> ChildEntities { get; set; }
    }

    public class RelatedEntity740
    {
        public int Id { get; set; }

        public RelatedEntity739 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity741> ChildEntities { get; set; }
    }

    public class RelatedEntity741
    {
        public int Id { get; set; }

        public RelatedEntity740 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity742> ChildEntities { get; set; }
    }

    public class RelatedEntity742
    {
        public int Id { get; set; }

        public RelatedEntity741 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity743> ChildEntities { get; set; }
    }

    public class RelatedEntity743
    {
        public int Id { get; set; }

        public RelatedEntity742 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity744> ChildEntities { get; set; }
    }

    public class RelatedEntity744
    {
        public int Id { get; set; }

        public RelatedEntity743 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity745> ChildEntities { get; set; }
    }

    public class RelatedEntity745
    {
        public int Id { get; set; }

        public RelatedEntity744 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity746> ChildEntities { get; set; }
    }

    public class RelatedEntity746
    {
        public int Id { get; set; }

        public RelatedEntity745 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity747> ChildEntities { get; set; }
    }

    public class RelatedEntity747
    {
        public int Id { get; set; }

        public RelatedEntity746 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity748> ChildEntities { get; set; }
    }

    public class RelatedEntity748
    {
        public int Id { get; set; }

        public RelatedEntity747 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity749> ChildEntities { get; set; }
    }

    public class RelatedEntity749
    {
        public int Id { get; set; }

        public RelatedEntity748 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity750> ChildEntities { get; set; }
    }

    public class RelatedEntity750
    {
        public int Id { get; set; }

        public RelatedEntity749 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity751> ChildEntities { get; set; }
    }

    public class RelatedEntity751
    {
        public int Id { get; set; }

        public RelatedEntity750 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity752> ChildEntities { get; set; }
    }

    public class RelatedEntity752
    {
        public int Id { get; set; }

        public RelatedEntity751 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity753> ChildEntities { get; set; }
    }

    public class RelatedEntity753
    {
        public int Id { get; set; }

        public RelatedEntity752 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity754> ChildEntities { get; set; }
    }

    public class RelatedEntity754
    {
        public int Id { get; set; }

        public RelatedEntity753 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity755> ChildEntities { get; set; }
    }

    public class RelatedEntity755
    {
        public int Id { get; set; }

        public RelatedEntity754 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity756> ChildEntities { get; set; }
    }

    public class RelatedEntity756
    {
        public int Id { get; set; }

        public RelatedEntity755 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity757> ChildEntities { get; set; }
    }

    public class RelatedEntity757
    {
        public int Id { get; set; }

        public RelatedEntity756 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity758> ChildEntities { get; set; }
    }

    public class RelatedEntity758
    {
        public int Id { get; set; }

        public RelatedEntity757 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity759> ChildEntities { get; set; }
    }

    public class RelatedEntity759
    {
        public int Id { get; set; }

        public RelatedEntity758 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity760> ChildEntities { get; set; }
    }

    public class RelatedEntity760
    {
        public int Id { get; set; }

        public RelatedEntity759 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity761> ChildEntities { get; set; }
    }

    public class RelatedEntity761
    {
        public int Id { get; set; }

        public RelatedEntity760 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity762> ChildEntities { get; set; }
    }

    public class RelatedEntity762
    {
        public int Id { get; set; }

        public RelatedEntity761 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity763> ChildEntities { get; set; }
    }

    public class RelatedEntity763
    {
        public int Id { get; set; }

        public RelatedEntity762 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity764> ChildEntities { get; set; }
    }

    public class RelatedEntity764
    {
        public int Id { get; set; }

        public RelatedEntity763 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity765> ChildEntities { get; set; }
    }

    public class RelatedEntity765
    {
        public int Id { get; set; }

        public RelatedEntity764 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity766> ChildEntities { get; set; }
    }

    public class RelatedEntity766
    {
        public int Id { get; set; }

        public RelatedEntity765 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity767> ChildEntities { get; set; }
    }

    public class RelatedEntity767
    {
        public int Id { get; set; }

        public RelatedEntity766 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity768> ChildEntities { get; set; }
    }

    public class RelatedEntity768
    {
        public int Id { get; set; }

        public RelatedEntity767 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity769> ChildEntities { get; set; }
    }

    public class RelatedEntity769
    {
        public int Id { get; set; }

        public RelatedEntity768 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity770> ChildEntities { get; set; }
    }

    public class RelatedEntity770
    {
        public int Id { get; set; }

        public RelatedEntity769 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity771> ChildEntities { get; set; }
    }

    public class RelatedEntity771
    {
        public int Id { get; set; }

        public RelatedEntity770 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity772> ChildEntities { get; set; }
    }

    public class RelatedEntity772
    {
        public int Id { get; set; }

        public RelatedEntity771 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity773> ChildEntities { get; set; }
    }

    public class RelatedEntity773
    {
        public int Id { get; set; }

        public RelatedEntity772 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity774> ChildEntities { get; set; }
    }

    public class RelatedEntity774
    {
        public int Id { get; set; }

        public RelatedEntity773 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity775> ChildEntities { get; set; }
    }

    public class RelatedEntity775
    {
        public int Id { get; set; }

        public RelatedEntity774 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity776> ChildEntities { get; set; }
    }

    public class RelatedEntity776
    {
        public int Id { get; set; }

        public RelatedEntity775 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity777> ChildEntities { get; set; }
    }

    public class RelatedEntity777
    {
        public int Id { get; set; }

        public RelatedEntity776 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity778> ChildEntities { get; set; }
    }

    public class RelatedEntity778
    {
        public int Id { get; set; }

        public RelatedEntity777 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity779> ChildEntities { get; set; }
    }

    public class RelatedEntity779
    {
        public int Id { get; set; }

        public RelatedEntity778 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity780> ChildEntities { get; set; }
    }

    public class RelatedEntity780
    {
        public int Id { get; set; }

        public RelatedEntity779 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity781> ChildEntities { get; set; }
    }

    public class RelatedEntity781
    {
        public int Id { get; set; }

        public RelatedEntity780 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity782> ChildEntities { get; set; }
    }

    public class RelatedEntity782
    {
        public int Id { get; set; }

        public RelatedEntity781 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity783> ChildEntities { get; set; }
    }

    public class RelatedEntity783
    {
        public int Id { get; set; }

        public RelatedEntity782 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity784> ChildEntities { get; set; }
    }

    public class RelatedEntity784
    {
        public int Id { get; set; }

        public RelatedEntity783 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity785> ChildEntities { get; set; }
    }

    public class RelatedEntity785
    {
        public int Id { get; set; }

        public RelatedEntity784 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity786> ChildEntities { get; set; }
    }

    public class RelatedEntity786
    {
        public int Id { get; set; }

        public RelatedEntity785 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity787> ChildEntities { get; set; }
    }

    public class RelatedEntity787
    {
        public int Id { get; set; }

        public RelatedEntity786 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity788> ChildEntities { get; set; }
    }

    public class RelatedEntity788
    {
        public int Id { get; set; }

        public RelatedEntity787 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity789> ChildEntities { get; set; }
    }

    public class RelatedEntity789
    {
        public int Id { get; set; }

        public RelatedEntity788 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity790> ChildEntities { get; set; }
    }

    public class RelatedEntity790
    {
        public int Id { get; set; }

        public RelatedEntity789 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity791> ChildEntities { get; set; }
    }

    public class RelatedEntity791
    {
        public int Id { get; set; }

        public RelatedEntity790 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity792> ChildEntities { get; set; }
    }

    public class RelatedEntity792
    {
        public int Id { get; set; }

        public RelatedEntity791 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity793> ChildEntities { get; set; }
    }

    public class RelatedEntity793
    {
        public int Id { get; set; }

        public RelatedEntity792 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity794> ChildEntities { get; set; }
    }

    public class RelatedEntity794
    {
        public int Id { get; set; }

        public RelatedEntity793 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity795> ChildEntities { get; set; }
    }

    public class RelatedEntity795
    {
        public int Id { get; set; }

        public RelatedEntity794 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity796> ChildEntities { get; set; }
    }

    public class RelatedEntity796
    {
        public int Id { get; set; }

        public RelatedEntity795 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity797> ChildEntities { get; set; }
    }

    public class RelatedEntity797
    {
        public int Id { get; set; }

        public RelatedEntity796 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity798> ChildEntities { get; set; }
    }

    public class RelatedEntity798
    {
        public int Id { get; set; }

        public RelatedEntity797 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity799> ChildEntities { get; set; }
    }

    public class RelatedEntity799
    {
        public int Id { get; set; }

        public RelatedEntity798 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity800> ChildEntities { get; set; }
    }

    public class RelatedEntity800
    {
        public int Id { get; set; }

        public RelatedEntity799 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity801> ChildEntities { get; set; }
    }

    public class RelatedEntity801
    {
        public int Id { get; set; }

        public RelatedEntity800 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity802> ChildEntities { get; set; }
    }

    public class RelatedEntity802
    {
        public int Id { get; set; }

        public RelatedEntity801 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity803> ChildEntities { get; set; }
    }

    public class RelatedEntity803
    {
        public int Id { get; set; }

        public RelatedEntity802 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity804> ChildEntities { get; set; }
    }

    public class RelatedEntity804
    {
        public int Id { get; set; }

        public RelatedEntity803 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity805> ChildEntities { get; set; }
    }

    public class RelatedEntity805
    {
        public int Id { get; set; }

        public RelatedEntity804 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity806> ChildEntities { get; set; }
    }

    public class RelatedEntity806
    {
        public int Id { get; set; }

        public RelatedEntity805 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity807> ChildEntities { get; set; }
    }

    public class RelatedEntity807
    {
        public int Id { get; set; }

        public RelatedEntity806 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity808> ChildEntities { get; set; }
    }

    public class RelatedEntity808
    {
        public int Id { get; set; }

        public RelatedEntity807 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity809> ChildEntities { get; set; }
    }

    public class RelatedEntity809
    {
        public int Id { get; set; }

        public RelatedEntity808 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity810> ChildEntities { get; set; }
    }

    public class RelatedEntity810
    {
        public int Id { get; set; }

        public RelatedEntity809 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity811> ChildEntities { get; set; }
    }

    public class RelatedEntity811
    {
        public int Id { get; set; }

        public RelatedEntity810 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity812> ChildEntities { get; set; }
    }

    public class RelatedEntity812
    {
        public int Id { get; set; }

        public RelatedEntity811 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity813> ChildEntities { get; set; }
    }

    public class RelatedEntity813
    {
        public int Id { get; set; }

        public RelatedEntity812 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity814> ChildEntities { get; set; }
    }

    public class RelatedEntity814
    {
        public int Id { get; set; }

        public RelatedEntity813 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity815> ChildEntities { get; set; }
    }

    public class RelatedEntity815
    {
        public int Id { get; set; }

        public RelatedEntity814 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity816> ChildEntities { get; set; }
    }

    public class RelatedEntity816
    {
        public int Id { get; set; }

        public RelatedEntity815 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity817> ChildEntities { get; set; }
    }

    public class RelatedEntity817
    {
        public int Id { get; set; }

        public RelatedEntity816 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity818> ChildEntities { get; set; }
    }

    public class RelatedEntity818
    {
        public int Id { get; set; }

        public RelatedEntity817 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity819> ChildEntities { get; set; }
    }

    public class RelatedEntity819
    {
        public int Id { get; set; }

        public RelatedEntity818 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity820> ChildEntities { get; set; }
    }

    public class RelatedEntity820
    {
        public int Id { get; set; }

        public RelatedEntity819 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity821> ChildEntities { get; set; }
    }

    public class RelatedEntity821
    {
        public int Id { get; set; }

        public RelatedEntity820 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity822> ChildEntities { get; set; }
    }

    public class RelatedEntity822
    {
        public int Id { get; set; }

        public RelatedEntity821 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity823> ChildEntities { get; set; }
    }

    public class RelatedEntity823
    {
        public int Id { get; set; }

        public RelatedEntity822 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity824> ChildEntities { get; set; }
    }

    public class RelatedEntity824
    {
        public int Id { get; set; }

        public RelatedEntity823 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity825> ChildEntities { get; set; }
    }

    public class RelatedEntity825
    {
        public int Id { get; set; }

        public RelatedEntity824 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity826> ChildEntities { get; set; }
    }

    public class RelatedEntity826
    {
        public int Id { get; set; }

        public RelatedEntity825 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity827> ChildEntities { get; set; }
    }

    public class RelatedEntity827
    {
        public int Id { get; set; }

        public RelatedEntity826 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity828> ChildEntities { get; set; }
    }

    public class RelatedEntity828
    {
        public int Id { get; set; }

        public RelatedEntity827 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity829> ChildEntities { get; set; }
    }

    public class RelatedEntity829
    {
        public int Id { get; set; }

        public RelatedEntity828 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity830> ChildEntities { get; set; }
    }

    public class RelatedEntity830
    {
        public int Id { get; set; }

        public RelatedEntity829 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity831> ChildEntities { get; set; }
    }

    public class RelatedEntity831
    {
        public int Id { get; set; }

        public RelatedEntity830 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity832> ChildEntities { get; set; }
    }

    public class RelatedEntity832
    {
        public int Id { get; set; }

        public RelatedEntity831 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity833> ChildEntities { get; set; }
    }

    public class RelatedEntity833
    {
        public int Id { get; set; }

        public RelatedEntity832 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity834> ChildEntities { get; set; }
    }

    public class RelatedEntity834
    {
        public int Id { get; set; }

        public RelatedEntity833 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity835> ChildEntities { get; set; }
    }

    public class RelatedEntity835
    {
        public int Id { get; set; }

        public RelatedEntity834 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity836> ChildEntities { get; set; }
    }

    public class RelatedEntity836
    {
        public int Id { get; set; }

        public RelatedEntity835 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity837> ChildEntities { get; set; }
    }

    public class RelatedEntity837
    {
        public int Id { get; set; }

        public RelatedEntity836 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity838> ChildEntities { get; set; }
    }

    public class RelatedEntity838
    {
        public int Id { get; set; }

        public RelatedEntity837 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity839> ChildEntities { get; set; }
    }

    public class RelatedEntity839
    {
        public int Id { get; set; }

        public RelatedEntity838 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity840> ChildEntities { get; set; }
    }

    public class RelatedEntity840
    {
        public int Id { get; set; }

        public RelatedEntity839 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity841> ChildEntities { get; set; }
    }

    public class RelatedEntity841
    {
        public int Id { get; set; }

        public RelatedEntity840 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity842> ChildEntities { get; set; }
    }

    public class RelatedEntity842
    {
        public int Id { get; set; }

        public RelatedEntity841 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity843> ChildEntities { get; set; }
    }

    public class RelatedEntity843
    {
        public int Id { get; set; }

        public RelatedEntity842 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity844> ChildEntities { get; set; }
    }

    public class RelatedEntity844
    {
        public int Id { get; set; }

        public RelatedEntity843 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity845> ChildEntities { get; set; }
    }

    public class RelatedEntity845
    {
        public int Id { get; set; }

        public RelatedEntity844 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity846> ChildEntities { get; set; }
    }

    public class RelatedEntity846
    {
        public int Id { get; set; }

        public RelatedEntity845 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity847> ChildEntities { get; set; }
    }

    public class RelatedEntity847
    {
        public int Id { get; set; }

        public RelatedEntity846 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity848> ChildEntities { get; set; }
    }

    public class RelatedEntity848
    {
        public int Id { get; set; }

        public RelatedEntity847 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity849> ChildEntities { get; set; }
    }

    public class RelatedEntity849
    {
        public int Id { get; set; }

        public RelatedEntity848 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity850> ChildEntities { get; set; }
    }

    public class RelatedEntity850
    {
        public int Id { get; set; }

        public RelatedEntity849 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity851> ChildEntities { get; set; }
    }

    public class RelatedEntity851
    {
        public int Id { get; set; }

        public RelatedEntity850 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity852> ChildEntities { get; set; }
    }

    public class RelatedEntity852
    {
        public int Id { get; set; }

        public RelatedEntity851 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity853> ChildEntities { get; set; }
    }

    public class RelatedEntity853
    {
        public int Id { get; set; }

        public RelatedEntity852 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity854> ChildEntities { get; set; }
    }

    public class RelatedEntity854
    {
        public int Id { get; set; }

        public RelatedEntity853 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity855> ChildEntities { get; set; }
    }

    public class RelatedEntity855
    {
        public int Id { get; set; }

        public RelatedEntity854 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity856> ChildEntities { get; set; }
    }

    public class RelatedEntity856
    {
        public int Id { get; set; }

        public RelatedEntity855 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity857> ChildEntities { get; set; }
    }

    public class RelatedEntity857
    {
        public int Id { get; set; }

        public RelatedEntity856 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity858> ChildEntities { get; set; }
    }

    public class RelatedEntity858
    {
        public int Id { get; set; }

        public RelatedEntity857 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity859> ChildEntities { get; set; }
    }

    public class RelatedEntity859
    {
        public int Id { get; set; }

        public RelatedEntity858 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity860> ChildEntities { get; set; }
    }

    public class RelatedEntity860
    {
        public int Id { get; set; }

        public RelatedEntity859 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity861> ChildEntities { get; set; }
    }

    public class RelatedEntity861
    {
        public int Id { get; set; }

        public RelatedEntity860 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity862> ChildEntities { get; set; }
    }

    public class RelatedEntity862
    {
        public int Id { get; set; }

        public RelatedEntity861 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity863> ChildEntities { get; set; }
    }

    public class RelatedEntity863
    {
        public int Id { get; set; }

        public RelatedEntity862 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity864> ChildEntities { get; set; }
    }

    public class RelatedEntity864
    {
        public int Id { get; set; }

        public RelatedEntity863 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity865> ChildEntities { get; set; }
    }

    public class RelatedEntity865
    {
        public int Id { get; set; }

        public RelatedEntity864 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity866> ChildEntities { get; set; }
    }

    public class RelatedEntity866
    {
        public int Id { get; set; }

        public RelatedEntity865 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity867> ChildEntities { get; set; }
    }

    public class RelatedEntity867
    {
        public int Id { get; set; }

        public RelatedEntity866 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity868> ChildEntities { get; set; }
    }

    public class RelatedEntity868
    {
        public int Id { get; set; }

        public RelatedEntity867 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity869> ChildEntities { get; set; }
    }

    public class RelatedEntity869
    {
        public int Id { get; set; }

        public RelatedEntity868 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity870> ChildEntities { get; set; }
    }

    public class RelatedEntity870
    {
        public int Id { get; set; }

        public RelatedEntity869 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity871> ChildEntities { get; set; }
    }

    public class RelatedEntity871
    {
        public int Id { get; set; }

        public RelatedEntity870 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity872> ChildEntities { get; set; }
    }

    public class RelatedEntity872
    {
        public int Id { get; set; }

        public RelatedEntity871 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity873> ChildEntities { get; set; }
    }

    public class RelatedEntity873
    {
        public int Id { get; set; }

        public RelatedEntity872 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity874> ChildEntities { get; set; }
    }

    public class RelatedEntity874
    {
        public int Id { get; set; }

        public RelatedEntity873 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity875> ChildEntities { get; set; }
    }

    public class RelatedEntity875
    {
        public int Id { get; set; }

        public RelatedEntity874 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity876> ChildEntities { get; set; }
    }

    public class RelatedEntity876
    {
        public int Id { get; set; }

        public RelatedEntity875 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity877> ChildEntities { get; set; }
    }

    public class RelatedEntity877
    {
        public int Id { get; set; }

        public RelatedEntity876 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity878> ChildEntities { get; set; }
    }

    public class RelatedEntity878
    {
        public int Id { get; set; }

        public RelatedEntity877 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity879> ChildEntities { get; set; }
    }

    public class RelatedEntity879
    {
        public int Id { get; set; }

        public RelatedEntity878 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity880> ChildEntities { get; set; }
    }

    public class RelatedEntity880
    {
        public int Id { get; set; }

        public RelatedEntity879 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity881> ChildEntities { get; set; }
    }

    public class RelatedEntity881
    {
        public int Id { get; set; }

        public RelatedEntity880 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity882> ChildEntities { get; set; }
    }

    public class RelatedEntity882
    {
        public int Id { get; set; }

        public RelatedEntity881 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity883> ChildEntities { get; set; }
    }

    public class RelatedEntity883
    {
        public int Id { get; set; }

        public RelatedEntity882 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity884> ChildEntities { get; set; }
    }

    public class RelatedEntity884
    {
        public int Id { get; set; }

        public RelatedEntity883 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity885> ChildEntities { get; set; }
    }

    public class RelatedEntity885
    {
        public int Id { get; set; }

        public RelatedEntity884 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity886> ChildEntities { get; set; }
    }

    public class RelatedEntity886
    {
        public int Id { get; set; }

        public RelatedEntity885 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity887> ChildEntities { get; set; }
    }

    public class RelatedEntity887
    {
        public int Id { get; set; }

        public RelatedEntity886 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity888> ChildEntities { get; set; }
    }

    public class RelatedEntity888
    {
        public int Id { get; set; }

        public RelatedEntity887 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity889> ChildEntities { get; set; }
    }

    public class RelatedEntity889
    {
        public int Id { get; set; }

        public RelatedEntity888 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity890> ChildEntities { get; set; }
    }

    public class RelatedEntity890
    {
        public int Id { get; set; }

        public RelatedEntity889 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity891> ChildEntities { get; set; }
    }

    public class RelatedEntity891
    {
        public int Id { get; set; }

        public RelatedEntity890 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity892> ChildEntities { get; set; }
    }

    public class RelatedEntity892
    {
        public int Id { get; set; }

        public RelatedEntity891 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity893> ChildEntities { get; set; }
    }

    public class RelatedEntity893
    {
        public int Id { get; set; }

        public RelatedEntity892 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity894> ChildEntities { get; set; }
    }

    public class RelatedEntity894
    {
        public int Id { get; set; }

        public RelatedEntity893 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity895> ChildEntities { get; set; }
    }

    public class RelatedEntity895
    {
        public int Id { get; set; }

        public RelatedEntity894 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity896> ChildEntities { get; set; }
    }

    public class RelatedEntity896
    {
        public int Id { get; set; }

        public RelatedEntity895 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity897> ChildEntities { get; set; }
    }

    public class RelatedEntity897
    {
        public int Id { get; set; }

        public RelatedEntity896 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity898> ChildEntities { get; set; }
    }

    public class RelatedEntity898
    {
        public int Id { get; set; }

        public RelatedEntity897 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity899> ChildEntities { get; set; }
    }

    public class RelatedEntity899
    {
        public int Id { get; set; }

        public RelatedEntity898 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity900> ChildEntities { get; set; }
    }

    public class RelatedEntity900
    {
        public int Id { get; set; }

        public RelatedEntity899 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity901> ChildEntities { get; set; }
    }

    public class RelatedEntity901
    {
        public int Id { get; set; }

        public RelatedEntity900 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity902> ChildEntities { get; set; }
    }

    public class RelatedEntity902
    {
        public int Id { get; set; }

        public RelatedEntity901 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity903> ChildEntities { get; set; }
    }

    public class RelatedEntity903
    {
        public int Id { get; set; }

        public RelatedEntity902 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity904> ChildEntities { get; set; }
    }

    public class RelatedEntity904
    {
        public int Id { get; set; }

        public RelatedEntity903 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity905> ChildEntities { get; set; }
    }

    public class RelatedEntity905
    {
        public int Id { get; set; }

        public RelatedEntity904 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity906> ChildEntities { get; set; }
    }

    public class RelatedEntity906
    {
        public int Id { get; set; }

        public RelatedEntity905 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity907> ChildEntities { get; set; }
    }

    public class RelatedEntity907
    {
        public int Id { get; set; }

        public RelatedEntity906 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity908> ChildEntities { get; set; }
    }

    public class RelatedEntity908
    {
        public int Id { get; set; }

        public RelatedEntity907 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity909> ChildEntities { get; set; }
    }

    public class RelatedEntity909
    {
        public int Id { get; set; }

        public RelatedEntity908 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity910> ChildEntities { get; set; }
    }

    public class RelatedEntity910
    {
        public int Id { get; set; }

        public RelatedEntity909 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity911> ChildEntities { get; set; }
    }

    public class RelatedEntity911
    {
        public int Id { get; set; }

        public RelatedEntity910 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity912> ChildEntities { get; set; }
    }

    public class RelatedEntity912
    {
        public int Id { get; set; }

        public RelatedEntity911 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity913> ChildEntities { get; set; }
    }

    public class RelatedEntity913
    {
        public int Id { get; set; }

        public RelatedEntity912 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity914> ChildEntities { get; set; }
    }

    public class RelatedEntity914
    {
        public int Id { get; set; }

        public RelatedEntity913 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity915> ChildEntities { get; set; }
    }

    public class RelatedEntity915
    {
        public int Id { get; set; }

        public RelatedEntity914 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity916> ChildEntities { get; set; }
    }

    public class RelatedEntity916
    {
        public int Id { get; set; }

        public RelatedEntity915 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity917> ChildEntities { get; set; }
    }

    public class RelatedEntity917
    {
        public int Id { get; set; }

        public RelatedEntity916 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity918> ChildEntities { get; set; }
    }

    public class RelatedEntity918
    {
        public int Id { get; set; }

        public RelatedEntity917 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity919> ChildEntities { get; set; }
    }

    public class RelatedEntity919
    {
        public int Id { get; set; }

        public RelatedEntity918 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity920> ChildEntities { get; set; }
    }

    public class RelatedEntity920
    {
        public int Id { get; set; }

        public RelatedEntity919 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity921> ChildEntities { get; set; }
    }

    public class RelatedEntity921
    {
        public int Id { get; set; }

        public RelatedEntity920 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity922> ChildEntities { get; set; }
    }

    public class RelatedEntity922
    {
        public int Id { get; set; }

        public RelatedEntity921 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity923> ChildEntities { get; set; }
    }

    public class RelatedEntity923
    {
        public int Id { get; set; }

        public RelatedEntity922 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity924> ChildEntities { get; set; }
    }

    public class RelatedEntity924
    {
        public int Id { get; set; }

        public RelatedEntity923 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity925> ChildEntities { get; set; }
    }

    public class RelatedEntity925
    {
        public int Id { get; set; }

        public RelatedEntity924 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity926> ChildEntities { get; set; }
    }

    public class RelatedEntity926
    {
        public int Id { get; set; }

        public RelatedEntity925 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity927> ChildEntities { get; set; }
    }

    public class RelatedEntity927
    {
        public int Id { get; set; }

        public RelatedEntity926 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity928> ChildEntities { get; set; }
    }

    public class RelatedEntity928
    {
        public int Id { get; set; }

        public RelatedEntity927 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity929> ChildEntities { get; set; }
    }

    public class RelatedEntity929
    {
        public int Id { get; set; }

        public RelatedEntity928 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity930> ChildEntities { get; set; }
    }

    public class RelatedEntity930
    {
        public int Id { get; set; }

        public RelatedEntity929 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity931> ChildEntities { get; set; }
    }

    public class RelatedEntity931
    {
        public int Id { get; set; }

        public RelatedEntity930 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity932> ChildEntities { get; set; }
    }

    public class RelatedEntity932
    {
        public int Id { get; set; }

        public RelatedEntity931 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity933> ChildEntities { get; set; }
    }

    public class RelatedEntity933
    {
        public int Id { get; set; }

        public RelatedEntity932 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity934> ChildEntities { get; set; }
    }

    public class RelatedEntity934
    {
        public int Id { get; set; }

        public RelatedEntity933 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity935> ChildEntities { get; set; }
    }

    public class RelatedEntity935
    {
        public int Id { get; set; }

        public RelatedEntity934 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity936> ChildEntities { get; set; }
    }

    public class RelatedEntity936
    {
        public int Id { get; set; }

        public RelatedEntity935 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity937> ChildEntities { get; set; }
    }

    public class RelatedEntity937
    {
        public int Id { get; set; }

        public RelatedEntity936 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity938> ChildEntities { get; set; }
    }

    public class RelatedEntity938
    {
        public int Id { get; set; }

        public RelatedEntity937 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity939> ChildEntities { get; set; }
    }

    public class RelatedEntity939
    {
        public int Id { get; set; }

        public RelatedEntity938 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity940> ChildEntities { get; set; }
    }

    public class RelatedEntity940
    {
        public int Id { get; set; }

        public RelatedEntity939 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity941> ChildEntities { get; set; }
    }

    public class RelatedEntity941
    {
        public int Id { get; set; }

        public RelatedEntity940 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity942> ChildEntities { get; set; }
    }

    public class RelatedEntity942
    {
        public int Id { get; set; }

        public RelatedEntity941 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity943> ChildEntities { get; set; }
    }

    public class RelatedEntity943
    {
        public int Id { get; set; }

        public RelatedEntity942 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity944> ChildEntities { get; set; }
    }

    public class RelatedEntity944
    {
        public int Id { get; set; }

        public RelatedEntity943 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity945> ChildEntities { get; set; }
    }

    public class RelatedEntity945
    {
        public int Id { get; set; }

        public RelatedEntity944 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity946> ChildEntities { get; set; }
    }

    public class RelatedEntity946
    {
        public int Id { get; set; }

        public RelatedEntity945 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity947> ChildEntities { get; set; }
    }

    public class RelatedEntity947
    {
        public int Id { get; set; }

        public RelatedEntity946 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity948> ChildEntities { get; set; }
    }

    public class RelatedEntity948
    {
        public int Id { get; set; }

        public RelatedEntity947 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity949> ChildEntities { get; set; }
    }

    public class RelatedEntity949
    {
        public int Id { get; set; }

        public RelatedEntity948 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity950> ChildEntities { get; set; }
    }

    public class RelatedEntity950
    {
        public int Id { get; set; }

        public RelatedEntity949 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity951> ChildEntities { get; set; }
    }

    public class RelatedEntity951
    {
        public int Id { get; set; }

        public RelatedEntity950 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity952> ChildEntities { get; set; }
    }

    public class RelatedEntity952
    {
        public int Id { get; set; }

        public RelatedEntity951 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity953> ChildEntities { get; set; }
    }

    public class RelatedEntity953
    {
        public int Id { get; set; }

        public RelatedEntity952 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity954> ChildEntities { get; set; }
    }

    public class RelatedEntity954
    {
        public int Id { get; set; }

        public RelatedEntity953 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity955> ChildEntities { get; set; }
    }

    public class RelatedEntity955
    {
        public int Id { get; set; }

        public RelatedEntity954 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity956> ChildEntities { get; set; }
    }

    public class RelatedEntity956
    {
        public int Id { get; set; }

        public RelatedEntity955 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity957> ChildEntities { get; set; }
    }

    public class RelatedEntity957
    {
        public int Id { get; set; }

        public RelatedEntity956 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity958> ChildEntities { get; set; }
    }

    public class RelatedEntity958
    {
        public int Id { get; set; }

        public RelatedEntity957 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity959> ChildEntities { get; set; }
    }

    public class RelatedEntity959
    {
        public int Id { get; set; }

        public RelatedEntity958 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity960> ChildEntities { get; set; }
    }

    public class RelatedEntity960
    {
        public int Id { get; set; }

        public RelatedEntity959 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity961> ChildEntities { get; set; }
    }

    public class RelatedEntity961
    {
        public int Id { get; set; }

        public RelatedEntity960 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity962> ChildEntities { get; set; }
    }

    public class RelatedEntity962
    {
        public int Id { get; set; }

        public RelatedEntity961 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity963> ChildEntities { get; set; }
    }

    public class RelatedEntity963
    {
        public int Id { get; set; }

        public RelatedEntity962 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity964> ChildEntities { get; set; }
    }

    public class RelatedEntity964
    {
        public int Id { get; set; }

        public RelatedEntity963 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity965> ChildEntities { get; set; }
    }

    public class RelatedEntity965
    {
        public int Id { get; set; }

        public RelatedEntity964 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity966> ChildEntities { get; set; }
    }

    public class RelatedEntity966
    {
        public int Id { get; set; }

        public RelatedEntity965 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity967> ChildEntities { get; set; }
    }

    public class RelatedEntity967
    {
        public int Id { get; set; }

        public RelatedEntity966 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity968> ChildEntities { get; set; }
    }

    public class RelatedEntity968
    {
        public int Id { get; set; }

        public RelatedEntity967 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity969> ChildEntities { get; set; }
    }

    public class RelatedEntity969
    {
        public int Id { get; set; }

        public RelatedEntity968 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity970> ChildEntities { get; set; }
    }

    public class RelatedEntity970
    {
        public int Id { get; set; }

        public RelatedEntity969 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity971> ChildEntities { get; set; }
    }

    public class RelatedEntity971
    {
        public int Id { get; set; }

        public RelatedEntity970 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity972> ChildEntities { get; set; }
    }

    public class RelatedEntity972
    {
        public int Id { get; set; }

        public RelatedEntity971 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity973> ChildEntities { get; set; }
    }

    public class RelatedEntity973
    {
        public int Id { get; set; }

        public RelatedEntity972 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity974> ChildEntities { get; set; }
    }

    public class RelatedEntity974
    {
        public int Id { get; set; }

        public RelatedEntity973 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity975> ChildEntities { get; set; }
    }

    public class RelatedEntity975
    {
        public int Id { get; set; }

        public RelatedEntity974 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity976> ChildEntities { get; set; }
    }

    public class RelatedEntity976
    {
        public int Id { get; set; }

        public RelatedEntity975 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity977> ChildEntities { get; set; }
    }

    public class RelatedEntity977
    {
        public int Id { get; set; }

        public RelatedEntity976 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity978> ChildEntities { get; set; }
    }

    public class RelatedEntity978
    {
        public int Id { get; set; }

        public RelatedEntity977 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity979> ChildEntities { get; set; }
    }

    public class RelatedEntity979
    {
        public int Id { get; set; }

        public RelatedEntity978 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity980> ChildEntities { get; set; }
    }

    public class RelatedEntity980
    {
        public int Id { get; set; }

        public RelatedEntity979 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity981> ChildEntities { get; set; }
    }

    public class RelatedEntity981
    {
        public int Id { get; set; }

        public RelatedEntity980 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity982> ChildEntities { get; set; }
    }

    public class RelatedEntity982
    {
        public int Id { get; set; }

        public RelatedEntity981 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity983> ChildEntities { get; set; }
    }

    public class RelatedEntity983
    {
        public int Id { get; set; }

        public RelatedEntity982 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity984> ChildEntities { get; set; }
    }

    public class RelatedEntity984
    {
        public int Id { get; set; }

        public RelatedEntity983 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity985> ChildEntities { get; set; }
    }

    public class RelatedEntity985
    {
        public int Id { get; set; }

        public RelatedEntity984 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity986> ChildEntities { get; set; }
    }

    public class RelatedEntity986
    {
        public int Id { get; set; }

        public RelatedEntity985 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity987> ChildEntities { get; set; }
    }

    public class RelatedEntity987
    {
        public int Id { get; set; }

        public RelatedEntity986 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity988> ChildEntities { get; set; }
    }

    public class RelatedEntity988
    {
        public int Id { get; set; }

        public RelatedEntity987 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity989> ChildEntities { get; set; }
    }

    public class RelatedEntity989
    {
        public int Id { get; set; }

        public RelatedEntity988 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity990> ChildEntities { get; set; }
    }

    public class RelatedEntity990
    {
        public int Id { get; set; }

        public RelatedEntity989 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity991> ChildEntities { get; set; }
    }

    public class RelatedEntity991
    {
        public int Id { get; set; }

        public RelatedEntity990 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity992> ChildEntities { get; set; }
    }

    public class RelatedEntity992
    {
        public int Id { get; set; }

        public RelatedEntity991 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity993> ChildEntities { get; set; }
    }

    public class RelatedEntity993
    {
        public int Id { get; set; }

        public RelatedEntity992 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity994> ChildEntities { get; set; }
    }

    public class RelatedEntity994
    {
        public int Id { get; set; }

        public RelatedEntity993 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity995> ChildEntities { get; set; }
    }

    public class RelatedEntity995
    {
        public int Id { get; set; }

        public RelatedEntity994 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity996> ChildEntities { get; set; }
    }

    public class RelatedEntity996
    {
        public int Id { get; set; }

        public RelatedEntity995 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity997> ChildEntities { get; set; }
    }

    public class RelatedEntity997
    {
        public int Id { get; set; }

        public RelatedEntity996 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity998> ChildEntities { get; set; }
    }

    public class RelatedEntity998
    {
        public int Id { get; set; }

        public RelatedEntity997 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity999> ChildEntities { get; set; }
    }

    public class RelatedEntity999
    {
        public int Id { get; set; }

        public RelatedEntity998 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1000> ChildEntities { get; set; }
    }

    public class RelatedEntity1000
    {
        public int Id { get; set; }

        public RelatedEntity999 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1001> ChildEntities { get; set; }
    }

    public class RelatedEntity1001
    {
        public int Id { get; set; }

        public RelatedEntity1000 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1002> ChildEntities { get; set; }
    }

    public class RelatedEntity1002
    {
        public int Id { get; set; }

        public RelatedEntity1001 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1003> ChildEntities { get; set; }
    }

    public class RelatedEntity1003
    {
        public int Id { get; set; }

        public RelatedEntity1002 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1004> ChildEntities { get; set; }
    }

    public class RelatedEntity1004
    {
        public int Id { get; set; }

        public RelatedEntity1003 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1005> ChildEntities { get; set; }
    }

    public class RelatedEntity1005
    {
        public int Id { get; set; }

        public RelatedEntity1004 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1006> ChildEntities { get; set; }
    }

    public class RelatedEntity1006
    {
        public int Id { get; set; }

        public RelatedEntity1005 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1007> ChildEntities { get; set; }
    }

    public class RelatedEntity1007
    {
        public int Id { get; set; }

        public RelatedEntity1006 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1008> ChildEntities { get; set; }
    }

    public class RelatedEntity1008
    {
        public int Id { get; set; }

        public RelatedEntity1007 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1009> ChildEntities { get; set; }
    }

    public class RelatedEntity1009
    {
        public int Id { get; set; }

        public RelatedEntity1008 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1010> ChildEntities { get; set; }
    }

    public class RelatedEntity1010
    {
        public int Id { get; set; }

        public RelatedEntity1009 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1011> ChildEntities { get; set; }
    }

    public class RelatedEntity1011
    {
        public int Id { get; set; }

        public RelatedEntity1010 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1012> ChildEntities { get; set; }
    }

    public class RelatedEntity1012
    {
        public int Id { get; set; }

        public RelatedEntity1011 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1013> ChildEntities { get; set; }
    }

    public class RelatedEntity1013
    {
        public int Id { get; set; }

        public RelatedEntity1012 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1014> ChildEntities { get; set; }
    }

    public class RelatedEntity1014
    {
        public int Id { get; set; }

        public RelatedEntity1013 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1015> ChildEntities { get; set; }
    }

    public class RelatedEntity1015
    {
        public int Id { get; set; }

        public RelatedEntity1014 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1016> ChildEntities { get; set; }
    }

    public class RelatedEntity1016
    {
        public int Id { get; set; }

        public RelatedEntity1015 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1017> ChildEntities { get; set; }
    }

    public class RelatedEntity1017
    {
        public int Id { get; set; }

        public RelatedEntity1016 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1018> ChildEntities { get; set; }
    }

    public class RelatedEntity1018
    {
        public int Id { get; set; }

        public RelatedEntity1017 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1019> ChildEntities { get; set; }
    }

    public class RelatedEntity1019
    {
        public int Id { get; set; }

        public RelatedEntity1018 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1020> ChildEntities { get; set; }
    }

    public class RelatedEntity1020
    {
        public int Id { get; set; }

        public RelatedEntity1019 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1021> ChildEntities { get; set; }
    }

    public class RelatedEntity1021
    {
        public int Id { get; set; }

        public RelatedEntity1020 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1022> ChildEntities { get; set; }
    }

    public class RelatedEntity1022
    {
        public int Id { get; set; }

        public RelatedEntity1021 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1023> ChildEntities { get; set; }
    }

    public class RelatedEntity1023
    {
        public int Id { get; set; }

        public RelatedEntity1022 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1024> ChildEntities { get; set; }
    }

    public class RelatedEntity1024
    {
        public int Id { get; set; }

        public RelatedEntity1023 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1025> ChildEntities { get; set; }
    }

    public class RelatedEntity1025
    {
        public int Id { get; set; }

        public RelatedEntity1024 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1026> ChildEntities { get; set; }
    }

    public class RelatedEntity1026
    {
        public int Id { get; set; }

        public RelatedEntity1025 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1027> ChildEntities { get; set; }
    }

    public class RelatedEntity1027
    {
        public int Id { get; set; }

        public RelatedEntity1026 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1028> ChildEntities { get; set; }
    }

    public class RelatedEntity1028
    {
        public int Id { get; set; }

        public RelatedEntity1027 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1029> ChildEntities { get; set; }
    }

    public class RelatedEntity1029
    {
        public int Id { get; set; }

        public RelatedEntity1028 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1030> ChildEntities { get; set; }
    }

    public class RelatedEntity1030
    {
        public int Id { get; set; }

        public RelatedEntity1029 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1031> ChildEntities { get; set; }
    }

    public class RelatedEntity1031
    {
        public int Id { get; set; }

        public RelatedEntity1030 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1032> ChildEntities { get; set; }
    }

    public class RelatedEntity1032
    {
        public int Id { get; set; }

        public RelatedEntity1031 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1033> ChildEntities { get; set; }
    }

    public class RelatedEntity1033
    {
        public int Id { get; set; }

        public RelatedEntity1032 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1034> ChildEntities { get; set; }
    }

    public class RelatedEntity1034
    {
        public int Id { get; set; }

        public RelatedEntity1033 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1035> ChildEntities { get; set; }
    }

    public class RelatedEntity1035
    {
        public int Id { get; set; }

        public RelatedEntity1034 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1036> ChildEntities { get; set; }
    }

    public class RelatedEntity1036
    {
        public int Id { get; set; }

        public RelatedEntity1035 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1037> ChildEntities { get; set; }
    }

    public class RelatedEntity1037
    {
        public int Id { get; set; }

        public RelatedEntity1036 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1038> ChildEntities { get; set; }
    }

    public class RelatedEntity1038
    {
        public int Id { get; set; }

        public RelatedEntity1037 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1039> ChildEntities { get; set; }
    }

    public class RelatedEntity1039
    {
        public int Id { get; set; }

        public RelatedEntity1038 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1040> ChildEntities { get; set; }
    }

    public class RelatedEntity1040
    {
        public int Id { get; set; }

        public RelatedEntity1039 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1041> ChildEntities { get; set; }
    }

    public class RelatedEntity1041
    {
        public int Id { get; set; }

        public RelatedEntity1040 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1042> ChildEntities { get; set; }
    }

    public class RelatedEntity1042
    {
        public int Id { get; set; }

        public RelatedEntity1041 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1043> ChildEntities { get; set; }
    }

    public class RelatedEntity1043
    {
        public int Id { get; set; }

        public RelatedEntity1042 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1044> ChildEntities { get; set; }
    }

    public class RelatedEntity1044
    {
        public int Id { get; set; }

        public RelatedEntity1043 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1045> ChildEntities { get; set; }
    }

    public class RelatedEntity1045
    {
        public int Id { get; set; }

        public RelatedEntity1044 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1046> ChildEntities { get; set; }
    }

    public class RelatedEntity1046
    {
        public int Id { get; set; }

        public RelatedEntity1045 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1047> ChildEntities { get; set; }
    }

    public class RelatedEntity1047
    {
        public int Id { get; set; }

        public RelatedEntity1046 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1048> ChildEntities { get; set; }
    }

    public class RelatedEntity1048
    {
        public int Id { get; set; }

        public RelatedEntity1047 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1049> ChildEntities { get; set; }
    }

    public class RelatedEntity1049
    {
        public int Id { get; set; }

        public RelatedEntity1048 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1050> ChildEntities { get; set; }
    }

    public class RelatedEntity1050
    {
        public int Id { get; set; }

        public RelatedEntity1049 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1051> ChildEntities { get; set; }
    }

    public class RelatedEntity1051
    {
        public int Id { get; set; }

        public RelatedEntity1050 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1052> ChildEntities { get; set; }
    }

    public class RelatedEntity1052
    {
        public int Id { get; set; }

        public RelatedEntity1051 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1053> ChildEntities { get; set; }
    }

    public class RelatedEntity1053
    {
        public int Id { get; set; }

        public RelatedEntity1052 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1054> ChildEntities { get; set; }
    }

    public class RelatedEntity1054
    {
        public int Id { get; set; }

        public RelatedEntity1053 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1055> ChildEntities { get; set; }
    }

    public class RelatedEntity1055
    {
        public int Id { get; set; }

        public RelatedEntity1054 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1056> ChildEntities { get; set; }
    }

    public class RelatedEntity1056
    {
        public int Id { get; set; }

        public RelatedEntity1055 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1057> ChildEntities { get; set; }
    }

    public class RelatedEntity1057
    {
        public int Id { get; set; }

        public RelatedEntity1056 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1058> ChildEntities { get; set; }
    }

    public class RelatedEntity1058
    {
        public int Id { get; set; }

        public RelatedEntity1057 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1059> ChildEntities { get; set; }
    }

    public class RelatedEntity1059
    {
        public int Id { get; set; }

        public RelatedEntity1058 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1060> ChildEntities { get; set; }
    }

    public class RelatedEntity1060
    {
        public int Id { get; set; }

        public RelatedEntity1059 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1061> ChildEntities { get; set; }
    }

    public class RelatedEntity1061
    {
        public int Id { get; set; }

        public RelatedEntity1060 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1062> ChildEntities { get; set; }
    }

    public class RelatedEntity1062
    {
        public int Id { get; set; }

        public RelatedEntity1061 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1063> ChildEntities { get; set; }
    }

    public class RelatedEntity1063
    {
        public int Id { get; set; }

        public RelatedEntity1062 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1064> ChildEntities { get; set; }
    }

    public class RelatedEntity1064
    {
        public int Id { get; set; }

        public RelatedEntity1063 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1065> ChildEntities { get; set; }
    }

    public class RelatedEntity1065
    {
        public int Id { get; set; }

        public RelatedEntity1064 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1066> ChildEntities { get; set; }
    }

    public class RelatedEntity1066
    {
        public int Id { get; set; }

        public RelatedEntity1065 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1067> ChildEntities { get; set; }
    }

    public class RelatedEntity1067
    {
        public int Id { get; set; }

        public RelatedEntity1066 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1068> ChildEntities { get; set; }
    }

    public class RelatedEntity1068
    {
        public int Id { get; set; }

        public RelatedEntity1067 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1069> ChildEntities { get; set; }
    }

    public class RelatedEntity1069
    {
        public int Id { get; set; }

        public RelatedEntity1068 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1070> ChildEntities { get; set; }
    }

    public class RelatedEntity1070
    {
        public int Id { get; set; }

        public RelatedEntity1069 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1071> ChildEntities { get; set; }
    }

    public class RelatedEntity1071
    {
        public int Id { get; set; }

        public RelatedEntity1070 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1072> ChildEntities { get; set; }
    }

    public class RelatedEntity1072
    {
        public int Id { get; set; }

        public RelatedEntity1071 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1073> ChildEntities { get; set; }
    }

    public class RelatedEntity1073
    {
        public int Id { get; set; }

        public RelatedEntity1072 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1074> ChildEntities { get; set; }
    }

    public class RelatedEntity1074
    {
        public int Id { get; set; }

        public RelatedEntity1073 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1075> ChildEntities { get; set; }
    }

    public class RelatedEntity1075
    {
        public int Id { get; set; }

        public RelatedEntity1074 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1076> ChildEntities { get; set; }
    }

    public class RelatedEntity1076
    {
        public int Id { get; set; }

        public RelatedEntity1075 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1077> ChildEntities { get; set; }
    }

    public class RelatedEntity1077
    {
        public int Id { get; set; }

        public RelatedEntity1076 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1078> ChildEntities { get; set; }
    }

    public class RelatedEntity1078
    {
        public int Id { get; set; }

        public RelatedEntity1077 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1079> ChildEntities { get; set; }
    }

    public class RelatedEntity1079
    {
        public int Id { get; set; }

        public RelatedEntity1078 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1080> ChildEntities { get; set; }
    }

    public class RelatedEntity1080
    {
        public int Id { get; set; }

        public RelatedEntity1079 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1081> ChildEntities { get; set; }
    }

    public class RelatedEntity1081
    {
        public int Id { get; set; }

        public RelatedEntity1080 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1082> ChildEntities { get; set; }
    }

    public class RelatedEntity1082
    {
        public int Id { get; set; }

        public RelatedEntity1081 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1083> ChildEntities { get; set; }
    }

    public class RelatedEntity1083
    {
        public int Id { get; set; }

        public RelatedEntity1082 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1084> ChildEntities { get; set; }
    }

    public class RelatedEntity1084
    {
        public int Id { get; set; }

        public RelatedEntity1083 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1085> ChildEntities { get; set; }
    }

    public class RelatedEntity1085
    {
        public int Id { get; set; }

        public RelatedEntity1084 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1086> ChildEntities { get; set; }
    }

    public class RelatedEntity1086
    {
        public int Id { get; set; }

        public RelatedEntity1085 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1087> ChildEntities { get; set; }
    }

    public class RelatedEntity1087
    {
        public int Id { get; set; }

        public RelatedEntity1086 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1088> ChildEntities { get; set; }
    }

    public class RelatedEntity1088
    {
        public int Id { get; set; }

        public RelatedEntity1087 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1089> ChildEntities { get; set; }
    }

    public class RelatedEntity1089
    {
        public int Id { get; set; }

        public RelatedEntity1088 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1090> ChildEntities { get; set; }
    }

    public class RelatedEntity1090
    {
        public int Id { get; set; }

        public RelatedEntity1089 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1091> ChildEntities { get; set; }
    }

    public class RelatedEntity1091
    {
        public int Id { get; set; }

        public RelatedEntity1090 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1092> ChildEntities { get; set; }
    }

    public class RelatedEntity1092
    {
        public int Id { get; set; }

        public RelatedEntity1091 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1093> ChildEntities { get; set; }
    }

    public class RelatedEntity1093
    {
        public int Id { get; set; }

        public RelatedEntity1092 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1094> ChildEntities { get; set; }
    }

    public class RelatedEntity1094
    {
        public int Id { get; set; }

        public RelatedEntity1093 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1095> ChildEntities { get; set; }
    }

    public class RelatedEntity1095
    {
        public int Id { get; set; }

        public RelatedEntity1094 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1096> ChildEntities { get; set; }
    }

    public class RelatedEntity1096
    {
        public int Id { get; set; }

        public RelatedEntity1095 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1097> ChildEntities { get; set; }
    }

    public class RelatedEntity1097
    {
        public int Id { get; set; }

        public RelatedEntity1096 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1098> ChildEntities { get; set; }
    }

    public class RelatedEntity1098
    {
        public int Id { get; set; }

        public RelatedEntity1097 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1099> ChildEntities { get; set; }
    }

    public class RelatedEntity1099
    {
        public int Id { get; set; }

        public RelatedEntity1098 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1100> ChildEntities { get; set; }
    }

    public class RelatedEntity1100
    {
        public int Id { get; set; }

        public RelatedEntity1099 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1101> ChildEntities { get; set; }
    }

    public class RelatedEntity1101
    {
        public int Id { get; set; }

        public RelatedEntity1100 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1102> ChildEntities { get; set; }
    }

    public class RelatedEntity1102
    {
        public int Id { get; set; }

        public RelatedEntity1101 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1103> ChildEntities { get; set; }
    }

    public class RelatedEntity1103
    {
        public int Id { get; set; }

        public RelatedEntity1102 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1104> ChildEntities { get; set; }
    }

    public class RelatedEntity1104
    {
        public int Id { get; set; }

        public RelatedEntity1103 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1105> ChildEntities { get; set; }
    }

    public class RelatedEntity1105
    {
        public int Id { get; set; }

        public RelatedEntity1104 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1106> ChildEntities { get; set; }
    }

    public class RelatedEntity1106
    {
        public int Id { get; set; }

        public RelatedEntity1105 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1107> ChildEntities { get; set; }
    }

    public class RelatedEntity1107
    {
        public int Id { get; set; }

        public RelatedEntity1106 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1108> ChildEntities { get; set; }
    }

    public class RelatedEntity1108
    {
        public int Id { get; set; }

        public RelatedEntity1107 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1109> ChildEntities { get; set; }
    }

    public class RelatedEntity1109
    {
        public int Id { get; set; }

        public RelatedEntity1108 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1110> ChildEntities { get; set; }
    }

    public class RelatedEntity1110
    {
        public int Id { get; set; }

        public RelatedEntity1109 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1111> ChildEntities { get; set; }
    }

    public class RelatedEntity1111
    {
        public int Id { get; set; }

        public RelatedEntity1110 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1112> ChildEntities { get; set; }
    }

    public class RelatedEntity1112
    {
        public int Id { get; set; }

        public RelatedEntity1111 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1113> ChildEntities { get; set; }
    }

    public class RelatedEntity1113
    {
        public int Id { get; set; }

        public RelatedEntity1112 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1114> ChildEntities { get; set; }
    }

    public class RelatedEntity1114
    {
        public int Id { get; set; }

        public RelatedEntity1113 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1115> ChildEntities { get; set; }
    }

    public class RelatedEntity1115
    {
        public int Id { get; set; }

        public RelatedEntity1114 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1116> ChildEntities { get; set; }
    }

    public class RelatedEntity1116
    {
        public int Id { get; set; }

        public RelatedEntity1115 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1117> ChildEntities { get; set; }
    }

    public class RelatedEntity1117
    {
        public int Id { get; set; }

        public RelatedEntity1116 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1118> ChildEntities { get; set; }
    }

    public class RelatedEntity1118
    {
        public int Id { get; set; }

        public RelatedEntity1117 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1119> ChildEntities { get; set; }
    }

    public class RelatedEntity1119
    {
        public int Id { get; set; }

        public RelatedEntity1118 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1120> ChildEntities { get; set; }
    }

    public class RelatedEntity1120
    {
        public int Id { get; set; }

        public RelatedEntity1119 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1121> ChildEntities { get; set; }
    }

    public class RelatedEntity1121
    {
        public int Id { get; set; }

        public RelatedEntity1120 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1122> ChildEntities { get; set; }
    }

    public class RelatedEntity1122
    {
        public int Id { get; set; }

        public RelatedEntity1121 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1123> ChildEntities { get; set; }
    }

    public class RelatedEntity1123
    {
        public int Id { get; set; }

        public RelatedEntity1122 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1124> ChildEntities { get; set; }
    }

    public class RelatedEntity1124
    {
        public int Id { get; set; }

        public RelatedEntity1123 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1125> ChildEntities { get; set; }
    }

    public class RelatedEntity1125
    {
        public int Id { get; set; }

        public RelatedEntity1124 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1126> ChildEntities { get; set; }
    }

    public class RelatedEntity1126
    {
        public int Id { get; set; }

        public RelatedEntity1125 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1127> ChildEntities { get; set; }
    }

    public class RelatedEntity1127
    {
        public int Id { get; set; }

        public RelatedEntity1126 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1128> ChildEntities { get; set; }
    }

    public class RelatedEntity1128
    {
        public int Id { get; set; }

        public RelatedEntity1127 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1129> ChildEntities { get; set; }
    }

    public class RelatedEntity1129
    {
        public int Id { get; set; }

        public RelatedEntity1128 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1130> ChildEntities { get; set; }
    }

    public class RelatedEntity1130
    {
        public int Id { get; set; }

        public RelatedEntity1129 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1131> ChildEntities { get; set; }
    }

    public class RelatedEntity1131
    {
        public int Id { get; set; }

        public RelatedEntity1130 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1132> ChildEntities { get; set; }
    }

    public class RelatedEntity1132
    {
        public int Id { get; set; }

        public RelatedEntity1131 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1133> ChildEntities { get; set; }
    }

    public class RelatedEntity1133
    {
        public int Id { get; set; }

        public RelatedEntity1132 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1134> ChildEntities { get; set; }
    }

    public class RelatedEntity1134
    {
        public int Id { get; set; }

        public RelatedEntity1133 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1135> ChildEntities { get; set; }
    }

    public class RelatedEntity1135
    {
        public int Id { get; set; }

        public RelatedEntity1134 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1136> ChildEntities { get; set; }
    }

    public class RelatedEntity1136
    {
        public int Id { get; set; }

        public RelatedEntity1135 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1137> ChildEntities { get; set; }
    }

    public class RelatedEntity1137
    {
        public int Id { get; set; }

        public RelatedEntity1136 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1138> ChildEntities { get; set; }
    }

    public class RelatedEntity1138
    {
        public int Id { get; set; }

        public RelatedEntity1137 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1139> ChildEntities { get; set; }
    }

    public class RelatedEntity1139
    {
        public int Id { get; set; }

        public RelatedEntity1138 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1140> ChildEntities { get; set; }
    }

    public class RelatedEntity1140
    {
        public int Id { get; set; }

        public RelatedEntity1139 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1141> ChildEntities { get; set; }
    }

    public class RelatedEntity1141
    {
        public int Id { get; set; }

        public RelatedEntity1140 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1142> ChildEntities { get; set; }
    }

    public class RelatedEntity1142
    {
        public int Id { get; set; }

        public RelatedEntity1141 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1143> ChildEntities { get; set; }
    }

    public class RelatedEntity1143
    {
        public int Id { get; set; }

        public RelatedEntity1142 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1144> ChildEntities { get; set; }
    }

    public class RelatedEntity1144
    {
        public int Id { get; set; }

        public RelatedEntity1143 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1145> ChildEntities { get; set; }
    }

    public class RelatedEntity1145
    {
        public int Id { get; set; }

        public RelatedEntity1144 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1146> ChildEntities { get; set; }
    }

    public class RelatedEntity1146
    {
        public int Id { get; set; }

        public RelatedEntity1145 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1147> ChildEntities { get; set; }
    }

    public class RelatedEntity1147
    {
        public int Id { get; set; }

        public RelatedEntity1146 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1148> ChildEntities { get; set; }
    }

    public class RelatedEntity1148
    {
        public int Id { get; set; }

        public RelatedEntity1147 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1149> ChildEntities { get; set; }
    }

    public class RelatedEntity1149
    {
        public int Id { get; set; }

        public RelatedEntity1148 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1150> ChildEntities { get; set; }
    }

    public class RelatedEntity1150
    {
        public int Id { get; set; }

        public RelatedEntity1149 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1151> ChildEntities { get; set; }
    }

    public class RelatedEntity1151
    {
        public int Id { get; set; }

        public RelatedEntity1150 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1152> ChildEntities { get; set; }
    }

    public class RelatedEntity1152
    {
        public int Id { get; set; }

        public RelatedEntity1151 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1153> ChildEntities { get; set; }
    }

    public class RelatedEntity1153
    {
        public int Id { get; set; }

        public RelatedEntity1152 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1154> ChildEntities { get; set; }
    }

    public class RelatedEntity1154
    {
        public int Id { get; set; }

        public RelatedEntity1153 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1155> ChildEntities { get; set; }
    }

    public class RelatedEntity1155
    {
        public int Id { get; set; }

        public RelatedEntity1154 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1156> ChildEntities { get; set; }
    }

    public class RelatedEntity1156
    {
        public int Id { get; set; }

        public RelatedEntity1155 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1157> ChildEntities { get; set; }
    }

    public class RelatedEntity1157
    {
        public int Id { get; set; }

        public RelatedEntity1156 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1158> ChildEntities { get; set; }
    }

    public class RelatedEntity1158
    {
        public int Id { get; set; }

        public RelatedEntity1157 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1159> ChildEntities { get; set; }
    }

    public class RelatedEntity1159
    {
        public int Id { get; set; }

        public RelatedEntity1158 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1160> ChildEntities { get; set; }
    }

    public class RelatedEntity1160
    {
        public int Id { get; set; }

        public RelatedEntity1159 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1161> ChildEntities { get; set; }
    }

    public class RelatedEntity1161
    {
        public int Id { get; set; }

        public RelatedEntity1160 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1162> ChildEntities { get; set; }
    }

    public class RelatedEntity1162
    {
        public int Id { get; set; }

        public RelatedEntity1161 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1163> ChildEntities { get; set; }
    }

    public class RelatedEntity1163
    {
        public int Id { get; set; }

        public RelatedEntity1162 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1164> ChildEntities { get; set; }
    }

    public class RelatedEntity1164
    {
        public int Id { get; set; }

        public RelatedEntity1163 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1165> ChildEntities { get; set; }
    }

    public class RelatedEntity1165
    {
        public int Id { get; set; }

        public RelatedEntity1164 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1166> ChildEntities { get; set; }
    }

    public class RelatedEntity1166
    {
        public int Id { get; set; }

        public RelatedEntity1165 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1167> ChildEntities { get; set; }
    }

    public class RelatedEntity1167
    {
        public int Id { get; set; }

        public RelatedEntity1166 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1168> ChildEntities { get; set; }
    }

    public class RelatedEntity1168
    {
        public int Id { get; set; }

        public RelatedEntity1167 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1169> ChildEntities { get; set; }
    }

    public class RelatedEntity1169
    {
        public int Id { get; set; }

        public RelatedEntity1168 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1170> ChildEntities { get; set; }
    }

    public class RelatedEntity1170
    {
        public int Id { get; set; }

        public RelatedEntity1169 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1171> ChildEntities { get; set; }
    }

    public class RelatedEntity1171
    {
        public int Id { get; set; }

        public RelatedEntity1170 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1172> ChildEntities { get; set; }
    }

    public class RelatedEntity1172
    {
        public int Id { get; set; }

        public RelatedEntity1171 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1173> ChildEntities { get; set; }
    }

    public class RelatedEntity1173
    {
        public int Id { get; set; }

        public RelatedEntity1172 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1174> ChildEntities { get; set; }
    }

    public class RelatedEntity1174
    {
        public int Id { get; set; }

        public RelatedEntity1173 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1175> ChildEntities { get; set; }
    }

    public class RelatedEntity1175
    {
        public int Id { get; set; }

        public RelatedEntity1174 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1176> ChildEntities { get; set; }
    }

    public class RelatedEntity1176
    {
        public int Id { get; set; }

        public RelatedEntity1175 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1177> ChildEntities { get; set; }
    }

    public class RelatedEntity1177
    {
        public int Id { get; set; }

        public RelatedEntity1176 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1178> ChildEntities { get; set; }
    }

    public class RelatedEntity1178
    {
        public int Id { get; set; }

        public RelatedEntity1177 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1179> ChildEntities { get; set; }
    }

    public class RelatedEntity1179
    {
        public int Id { get; set; }

        public RelatedEntity1178 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1180> ChildEntities { get; set; }
    }

    public class RelatedEntity1180
    {
        public int Id { get; set; }

        public RelatedEntity1179 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1181> ChildEntities { get; set; }
    }

    public class RelatedEntity1181
    {
        public int Id { get; set; }

        public RelatedEntity1180 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1182> ChildEntities { get; set; }
    }

    public class RelatedEntity1182
    {
        public int Id { get; set; }

        public RelatedEntity1181 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1183> ChildEntities { get; set; }
    }

    public class RelatedEntity1183
    {
        public int Id { get; set; }

        public RelatedEntity1182 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1184> ChildEntities { get; set; }
    }

    public class RelatedEntity1184
    {
        public int Id { get; set; }

        public RelatedEntity1183 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1185> ChildEntities { get; set; }
    }

    public class RelatedEntity1185
    {
        public int Id { get; set; }

        public RelatedEntity1184 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1186> ChildEntities { get; set; }
    }

    public class RelatedEntity1186
    {
        public int Id { get; set; }

        public RelatedEntity1185 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1187> ChildEntities { get; set; }
    }

    public class RelatedEntity1187
    {
        public int Id { get; set; }

        public RelatedEntity1186 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1188> ChildEntities { get; set; }
    }

    public class RelatedEntity1188
    {
        public int Id { get; set; }

        public RelatedEntity1187 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1189> ChildEntities { get; set; }
    }

    public class RelatedEntity1189
    {
        public int Id { get; set; }

        public RelatedEntity1188 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1190> ChildEntities { get; set; }
    }

    public class RelatedEntity1190
    {
        public int Id { get; set; }

        public RelatedEntity1189 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1191> ChildEntities { get; set; }
    }

    public class RelatedEntity1191
    {
        public int Id { get; set; }

        public RelatedEntity1190 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1192> ChildEntities { get; set; }
    }

    public class RelatedEntity1192
    {
        public int Id { get; set; }

        public RelatedEntity1191 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1193> ChildEntities { get; set; }
    }

    public class RelatedEntity1193
    {
        public int Id { get; set; }

        public RelatedEntity1192 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1194> ChildEntities { get; set; }
    }

    public class RelatedEntity1194
    {
        public int Id { get; set; }

        public RelatedEntity1193 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1195> ChildEntities { get; set; }
    }

    public class RelatedEntity1195
    {
        public int Id { get; set; }

        public RelatedEntity1194 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1196> ChildEntities { get; set; }
    }

    public class RelatedEntity1196
    {
        public int Id { get; set; }

        public RelatedEntity1195 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1197> ChildEntities { get; set; }
    }

    public class RelatedEntity1197
    {
        public int Id { get; set; }

        public RelatedEntity1196 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1198> ChildEntities { get; set; }
    }

    public class RelatedEntity1198
    {
        public int Id { get; set; }

        public RelatedEntity1197 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1199> ChildEntities { get; set; }
    }

    public class RelatedEntity1199
    {
        public int Id { get; set; }

        public RelatedEntity1198 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1200> ChildEntities { get; set; }
    }

    public class RelatedEntity1200
    {
        public int Id { get; set; }

        public RelatedEntity1199 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1201> ChildEntities { get; set; }
    }

    public class RelatedEntity1201
    {
        public int Id { get; set; }

        public RelatedEntity1200 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1202> ChildEntities { get; set; }
    }

    public class RelatedEntity1202
    {
        public int Id { get; set; }

        public RelatedEntity1201 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1203> ChildEntities { get; set; }
    }

    public class RelatedEntity1203
    {
        public int Id { get; set; }

        public RelatedEntity1202 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1204> ChildEntities { get; set; }
    }

    public class RelatedEntity1204
    {
        public int Id { get; set; }

        public RelatedEntity1203 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1205> ChildEntities { get; set; }
    }

    public class RelatedEntity1205
    {
        public int Id { get; set; }

        public RelatedEntity1204 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1206> ChildEntities { get; set; }
    }

    public class RelatedEntity1206
    {
        public int Id { get; set; }

        public RelatedEntity1205 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1207> ChildEntities { get; set; }
    }

    public class RelatedEntity1207
    {
        public int Id { get; set; }

        public RelatedEntity1206 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1208> ChildEntities { get; set; }
    }

    public class RelatedEntity1208
    {
        public int Id { get; set; }

        public RelatedEntity1207 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1209> ChildEntities { get; set; }
    }

    public class RelatedEntity1209
    {
        public int Id { get; set; }

        public RelatedEntity1208 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1210> ChildEntities { get; set; }
    }

    public class RelatedEntity1210
    {
        public int Id { get; set; }

        public RelatedEntity1209 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1211> ChildEntities { get; set; }
    }

    public class RelatedEntity1211
    {
        public int Id { get; set; }

        public RelatedEntity1210 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1212> ChildEntities { get; set; }
    }

    public class RelatedEntity1212
    {
        public int Id { get; set; }

        public RelatedEntity1211 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1213> ChildEntities { get; set; }
    }

    public class RelatedEntity1213
    {
        public int Id { get; set; }

        public RelatedEntity1212 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1214> ChildEntities { get; set; }
    }

    public class RelatedEntity1214
    {
        public int Id { get; set; }

        public RelatedEntity1213 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1215> ChildEntities { get; set; }
    }

    public class RelatedEntity1215
    {
        public int Id { get; set; }

        public RelatedEntity1214 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1216> ChildEntities { get; set; }
    }

    public class RelatedEntity1216
    {
        public int Id { get; set; }

        public RelatedEntity1215 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1217> ChildEntities { get; set; }
    }

    public class RelatedEntity1217
    {
        public int Id { get; set; }

        public RelatedEntity1216 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1218> ChildEntities { get; set; }
    }

    public class RelatedEntity1218
    {
        public int Id { get; set; }

        public RelatedEntity1217 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1219> ChildEntities { get; set; }
    }

    public class RelatedEntity1219
    {
        public int Id { get; set; }

        public RelatedEntity1218 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1220> ChildEntities { get; set; }
    }

    public class RelatedEntity1220
    {
        public int Id { get; set; }

        public RelatedEntity1219 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1221> ChildEntities { get; set; }
    }

    public class RelatedEntity1221
    {
        public int Id { get; set; }

        public RelatedEntity1220 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1222> ChildEntities { get; set; }
    }

    public class RelatedEntity1222
    {
        public int Id { get; set; }

        public RelatedEntity1221 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1223> ChildEntities { get; set; }
    }

    public class RelatedEntity1223
    {
        public int Id { get; set; }

        public RelatedEntity1222 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1224> ChildEntities { get; set; }
    }

    public class RelatedEntity1224
    {
        public int Id { get; set; }

        public RelatedEntity1223 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1225> ChildEntities { get; set; }
    }

    public class RelatedEntity1225
    {
        public int Id { get; set; }

        public RelatedEntity1224 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1226> ChildEntities { get; set; }
    }

    public class RelatedEntity1226
    {
        public int Id { get; set; }

        public RelatedEntity1225 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1227> ChildEntities { get; set; }
    }

    public class RelatedEntity1227
    {
        public int Id { get; set; }

        public RelatedEntity1226 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1228> ChildEntities { get; set; }
    }

    public class RelatedEntity1228
    {
        public int Id { get; set; }

        public RelatedEntity1227 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1229> ChildEntities { get; set; }
    }

    public class RelatedEntity1229
    {
        public int Id { get; set; }

        public RelatedEntity1228 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1230> ChildEntities { get; set; }
    }

    public class RelatedEntity1230
    {
        public int Id { get; set; }

        public RelatedEntity1229 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1231> ChildEntities { get; set; }
    }

    public class RelatedEntity1231
    {
        public int Id { get; set; }

        public RelatedEntity1230 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1232> ChildEntities { get; set; }
    }

    public class RelatedEntity1232
    {
        public int Id { get; set; }

        public RelatedEntity1231 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1233> ChildEntities { get; set; }
    }

    public class RelatedEntity1233
    {
        public int Id { get; set; }

        public RelatedEntity1232 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1234> ChildEntities { get; set; }
    }

    public class RelatedEntity1234
    {
        public int Id { get; set; }

        public RelatedEntity1233 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1235> ChildEntities { get; set; }
    }

    public class RelatedEntity1235
    {
        public int Id { get; set; }

        public RelatedEntity1234 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1236> ChildEntities { get; set; }
    }

    public class RelatedEntity1236
    {
        public int Id { get; set; }

        public RelatedEntity1235 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1237> ChildEntities { get; set; }
    }

    public class RelatedEntity1237
    {
        public int Id { get; set; }

        public RelatedEntity1236 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1238> ChildEntities { get; set; }
    }

    public class RelatedEntity1238
    {
        public int Id { get; set; }

        public RelatedEntity1237 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1239> ChildEntities { get; set; }
    }

    public class RelatedEntity1239
    {
        public int Id { get; set; }

        public RelatedEntity1238 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1240> ChildEntities { get; set; }
    }

    public class RelatedEntity1240
    {
        public int Id { get; set; }

        public RelatedEntity1239 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1241> ChildEntities { get; set; }
    }

    public class RelatedEntity1241
    {
        public int Id { get; set; }

        public RelatedEntity1240 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1242> ChildEntities { get; set; }
    }

    public class RelatedEntity1242
    {
        public int Id { get; set; }

        public RelatedEntity1241 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1243> ChildEntities { get; set; }
    }

    public class RelatedEntity1243
    {
        public int Id { get; set; }

        public RelatedEntity1242 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1244> ChildEntities { get; set; }
    }

    public class RelatedEntity1244
    {
        public int Id { get; set; }

        public RelatedEntity1243 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1245> ChildEntities { get; set; }
    }

    public class RelatedEntity1245
    {
        public int Id { get; set; }

        public RelatedEntity1244 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1246> ChildEntities { get; set; }
    }

    public class RelatedEntity1246
    {
        public int Id { get; set; }

        public RelatedEntity1245 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1247> ChildEntities { get; set; }
    }

    public class RelatedEntity1247
    {
        public int Id { get; set; }

        public RelatedEntity1246 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1248> ChildEntities { get; set; }
    }

    public class RelatedEntity1248
    {
        public int Id { get; set; }

        public RelatedEntity1247 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1249> ChildEntities { get; set; }
    }

    public class RelatedEntity1249
    {
        public int Id { get; set; }

        public RelatedEntity1248 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1250> ChildEntities { get; set; }
    }

    public class RelatedEntity1250
    {
        public int Id { get; set; }

        public RelatedEntity1249 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1251> ChildEntities { get; set; }
    }

    public class RelatedEntity1251
    {
        public int Id { get; set; }

        public RelatedEntity1250 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1252> ChildEntities { get; set; }
    }

    public class RelatedEntity1252
    {
        public int Id { get; set; }

        public RelatedEntity1251 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1253> ChildEntities { get; set; }
    }

    public class RelatedEntity1253
    {
        public int Id { get; set; }

        public RelatedEntity1252 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1254> ChildEntities { get; set; }
    }

    public class RelatedEntity1254
    {
        public int Id { get; set; }

        public RelatedEntity1253 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1255> ChildEntities { get; set; }
    }

    public class RelatedEntity1255
    {
        public int Id { get; set; }

        public RelatedEntity1254 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1256> ChildEntities { get; set; }
    }

    public class RelatedEntity1256
    {
        public int Id { get; set; }

        public RelatedEntity1255 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1257> ChildEntities { get; set; }
    }

    public class RelatedEntity1257
    {
        public int Id { get; set; }

        public RelatedEntity1256 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1258> ChildEntities { get; set; }
    }

    public class RelatedEntity1258
    {
        public int Id { get; set; }

        public RelatedEntity1257 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1259> ChildEntities { get; set; }
    }

    public class RelatedEntity1259
    {
        public int Id { get; set; }

        public RelatedEntity1258 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1260> ChildEntities { get; set; }
    }

    public class RelatedEntity1260
    {
        public int Id { get; set; }

        public RelatedEntity1259 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1261> ChildEntities { get; set; }
    }

    public class RelatedEntity1261
    {
        public int Id { get; set; }

        public RelatedEntity1260 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1262> ChildEntities { get; set; }
    }

    public class RelatedEntity1262
    {
        public int Id { get; set; }

        public RelatedEntity1261 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1263> ChildEntities { get; set; }
    }

    public class RelatedEntity1263
    {
        public int Id { get; set; }

        public RelatedEntity1262 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1264> ChildEntities { get; set; }
    }

    public class RelatedEntity1264
    {
        public int Id { get; set; }

        public RelatedEntity1263 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1265> ChildEntities { get; set; }
    }

    public class RelatedEntity1265
    {
        public int Id { get; set; }

        public RelatedEntity1264 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1266> ChildEntities { get; set; }
    }

    public class RelatedEntity1266
    {
        public int Id { get; set; }

        public RelatedEntity1265 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1267> ChildEntities { get; set; }
    }

    public class RelatedEntity1267
    {
        public int Id { get; set; }

        public RelatedEntity1266 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1268> ChildEntities { get; set; }
    }

    public class RelatedEntity1268
    {
        public int Id { get; set; }

        public RelatedEntity1267 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1269> ChildEntities { get; set; }
    }

    public class RelatedEntity1269
    {
        public int Id { get; set; }

        public RelatedEntity1268 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1270> ChildEntities { get; set; }
    }

    public class RelatedEntity1270
    {
        public int Id { get; set; }

        public RelatedEntity1269 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1271> ChildEntities { get; set; }
    }

    public class RelatedEntity1271
    {
        public int Id { get; set; }

        public RelatedEntity1270 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1272> ChildEntities { get; set; }
    }

    public class RelatedEntity1272
    {
        public int Id { get; set; }

        public RelatedEntity1271 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1273> ChildEntities { get; set; }
    }

    public class RelatedEntity1273
    {
        public int Id { get; set; }

        public RelatedEntity1272 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1274> ChildEntities { get; set; }
    }

    public class RelatedEntity1274
    {
        public int Id { get; set; }

        public RelatedEntity1273 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1275> ChildEntities { get; set; }
    }

    public class RelatedEntity1275
    {
        public int Id { get; set; }

        public RelatedEntity1274 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1276> ChildEntities { get; set; }
    }

    public class RelatedEntity1276
    {
        public int Id { get; set; }

        public RelatedEntity1275 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1277> ChildEntities { get; set; }
    }

    public class RelatedEntity1277
    {
        public int Id { get; set; }

        public RelatedEntity1276 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1278> ChildEntities { get; set; }
    }

    public class RelatedEntity1278
    {
        public int Id { get; set; }

        public RelatedEntity1277 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1279> ChildEntities { get; set; }
    }

    public class RelatedEntity1279
    {
        public int Id { get; set; }

        public RelatedEntity1278 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1280> ChildEntities { get; set; }
    }

    public class RelatedEntity1280
    {
        public int Id { get; set; }

        public RelatedEntity1279 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1281> ChildEntities { get; set; }
    }

    public class RelatedEntity1281
    {
        public int Id { get; set; }

        public RelatedEntity1280 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1282> ChildEntities { get; set; }
    }

    public class RelatedEntity1282
    {
        public int Id { get; set; }

        public RelatedEntity1281 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1283> ChildEntities { get; set; }
    }

    public class RelatedEntity1283
    {
        public int Id { get; set; }

        public RelatedEntity1282 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1284> ChildEntities { get; set; }
    }

    public class RelatedEntity1284
    {
        public int Id { get; set; }

        public RelatedEntity1283 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1285> ChildEntities { get; set; }
    }

    public class RelatedEntity1285
    {
        public int Id { get; set; }

        public RelatedEntity1284 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1286> ChildEntities { get; set; }
    }

    public class RelatedEntity1286
    {
        public int Id { get; set; }

        public RelatedEntity1285 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1287> ChildEntities { get; set; }
    }

    public class RelatedEntity1287
    {
        public int Id { get; set; }

        public RelatedEntity1286 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1288> ChildEntities { get; set; }
    }

    public class RelatedEntity1288
    {
        public int Id { get; set; }

        public RelatedEntity1287 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1289> ChildEntities { get; set; }
    }

    public class RelatedEntity1289
    {
        public int Id { get; set; }

        public RelatedEntity1288 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1290> ChildEntities { get; set; }
    }

    public class RelatedEntity1290
    {
        public int Id { get; set; }

        public RelatedEntity1289 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1291> ChildEntities { get; set; }
    }

    public class RelatedEntity1291
    {
        public int Id { get; set; }

        public RelatedEntity1290 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1292> ChildEntities { get; set; }
    }

    public class RelatedEntity1292
    {
        public int Id { get; set; }

        public RelatedEntity1291 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1293> ChildEntities { get; set; }
    }

    public class RelatedEntity1293
    {
        public int Id { get; set; }

        public RelatedEntity1292 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1294> ChildEntities { get; set; }
    }

    public class RelatedEntity1294
    {
        public int Id { get; set; }

        public RelatedEntity1293 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1295> ChildEntities { get; set; }
    }

    public class RelatedEntity1295
    {
        public int Id { get; set; }

        public RelatedEntity1294 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1296> ChildEntities { get; set; }
    }

    public class RelatedEntity1296
    {
        public int Id { get; set; }

        public RelatedEntity1295 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1297> ChildEntities { get; set; }
    }

    public class RelatedEntity1297
    {
        public int Id { get; set; }

        public RelatedEntity1296 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1298> ChildEntities { get; set; }
    }

    public class RelatedEntity1298
    {
        public int Id { get; set; }

        public RelatedEntity1297 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1299> ChildEntities { get; set; }
    }

    public class RelatedEntity1299
    {
        public int Id { get; set; }

        public RelatedEntity1298 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1300> ChildEntities { get; set; }
    }

    public class RelatedEntity1300
    {
        public int Id { get; set; }

        public RelatedEntity1299 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1301> ChildEntities { get; set; }
    }

    public class RelatedEntity1301
    {
        public int Id { get; set; }

        public RelatedEntity1300 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1302> ChildEntities { get; set; }
    }

    public class RelatedEntity1302
    {
        public int Id { get; set; }

        public RelatedEntity1301 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1303> ChildEntities { get; set; }
    }

    public class RelatedEntity1303
    {
        public int Id { get; set; }

        public RelatedEntity1302 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1304> ChildEntities { get; set; }
    }

    public class RelatedEntity1304
    {
        public int Id { get; set; }

        public RelatedEntity1303 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1305> ChildEntities { get; set; }
    }

    public class RelatedEntity1305
    {
        public int Id { get; set; }

        public RelatedEntity1304 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1306> ChildEntities { get; set; }
    }

    public class RelatedEntity1306
    {
        public int Id { get; set; }

        public RelatedEntity1305 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1307> ChildEntities { get; set; }
    }

    public class RelatedEntity1307
    {
        public int Id { get; set; }

        public RelatedEntity1306 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1308> ChildEntities { get; set; }
    }

    public class RelatedEntity1308
    {
        public int Id { get; set; }

        public RelatedEntity1307 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1309> ChildEntities { get; set; }
    }

    public class RelatedEntity1309
    {
        public int Id { get; set; }

        public RelatedEntity1308 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1310> ChildEntities { get; set; }
    }

    public class RelatedEntity1310
    {
        public int Id { get; set; }

        public RelatedEntity1309 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1311> ChildEntities { get; set; }
    }

    public class RelatedEntity1311
    {
        public int Id { get; set; }

        public RelatedEntity1310 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1312> ChildEntities { get; set; }
    }

    public class RelatedEntity1312
    {
        public int Id { get; set; }

        public RelatedEntity1311 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1313> ChildEntities { get; set; }
    }

    public class RelatedEntity1313
    {
        public int Id { get; set; }

        public RelatedEntity1312 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1314> ChildEntities { get; set; }
    }

    public class RelatedEntity1314
    {
        public int Id { get; set; }

        public RelatedEntity1313 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1315> ChildEntities { get; set; }
    }

    public class RelatedEntity1315
    {
        public int Id { get; set; }

        public RelatedEntity1314 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1316> ChildEntities { get; set; }
    }

    public class RelatedEntity1316
    {
        public int Id { get; set; }

        public RelatedEntity1315 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1317> ChildEntities { get; set; }
    }

    public class RelatedEntity1317
    {
        public int Id { get; set; }

        public RelatedEntity1316 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1318> ChildEntities { get; set; }
    }

    public class RelatedEntity1318
    {
        public int Id { get; set; }

        public RelatedEntity1317 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1319> ChildEntities { get; set; }
    }

    public class RelatedEntity1319
    {
        public int Id { get; set; }

        public RelatedEntity1318 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1320> ChildEntities { get; set; }
    }

    public class RelatedEntity1320
    {
        public int Id { get; set; }

        public RelatedEntity1319 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1321> ChildEntities { get; set; }
    }

    public class RelatedEntity1321
    {
        public int Id { get; set; }

        public RelatedEntity1320 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1322> ChildEntities { get; set; }
    }

    public class RelatedEntity1322
    {
        public int Id { get; set; }

        public RelatedEntity1321 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1323> ChildEntities { get; set; }
    }

    public class RelatedEntity1323
    {
        public int Id { get; set; }

        public RelatedEntity1322 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1324> ChildEntities { get; set; }
    }

    public class RelatedEntity1324
    {
        public int Id { get; set; }

        public RelatedEntity1323 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1325> ChildEntities { get; set; }
    }

    public class RelatedEntity1325
    {
        public int Id { get; set; }

        public RelatedEntity1324 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1326> ChildEntities { get; set; }
    }

    public class RelatedEntity1326
    {
        public int Id { get; set; }

        public RelatedEntity1325 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1327> ChildEntities { get; set; }
    }

    public class RelatedEntity1327
    {
        public int Id { get; set; }

        public RelatedEntity1326 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1328> ChildEntities { get; set; }
    }

    public class RelatedEntity1328
    {
        public int Id { get; set; }

        public RelatedEntity1327 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1329> ChildEntities { get; set; }
    }

    public class RelatedEntity1329
    {
        public int Id { get; set; }

        public RelatedEntity1328 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1330> ChildEntities { get; set; }
    }

    public class RelatedEntity1330
    {
        public int Id { get; set; }

        public RelatedEntity1329 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1331> ChildEntities { get; set; }
    }

    public class RelatedEntity1331
    {
        public int Id { get; set; }

        public RelatedEntity1330 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1332> ChildEntities { get; set; }
    }

    public class RelatedEntity1332
    {
        public int Id { get; set; }

        public RelatedEntity1331 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1333> ChildEntities { get; set; }
    }

    public class RelatedEntity1333
    {
        public int Id { get; set; }

        public RelatedEntity1332 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1334> ChildEntities { get; set; }
    }

    public class RelatedEntity1334
    {
        public int Id { get; set; }

        public RelatedEntity1333 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1335> ChildEntities { get; set; }
    }

    public class RelatedEntity1335
    {
        public int Id { get; set; }

        public RelatedEntity1334 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1336> ChildEntities { get; set; }
    }

    public class RelatedEntity1336
    {
        public int Id { get; set; }

        public RelatedEntity1335 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1337> ChildEntities { get; set; }
    }

    public class RelatedEntity1337
    {
        public int Id { get; set; }

        public RelatedEntity1336 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1338> ChildEntities { get; set; }
    }

    public class RelatedEntity1338
    {
        public int Id { get; set; }

        public RelatedEntity1337 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1339> ChildEntities { get; set; }
    }

    public class RelatedEntity1339
    {
        public int Id { get; set; }

        public RelatedEntity1338 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1340> ChildEntities { get; set; }
    }

    public class RelatedEntity1340
    {
        public int Id { get; set; }

        public RelatedEntity1339 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1341> ChildEntities { get; set; }
    }

    public class RelatedEntity1341
    {
        public int Id { get; set; }

        public RelatedEntity1340 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1342> ChildEntities { get; set; }
    }

    public class RelatedEntity1342
    {
        public int Id { get; set; }

        public RelatedEntity1341 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1343> ChildEntities { get; set; }
    }

    public class RelatedEntity1343
    {
        public int Id { get; set; }

        public RelatedEntity1342 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1344> ChildEntities { get; set; }
    }

    public class RelatedEntity1344
    {
        public int Id { get; set; }

        public RelatedEntity1343 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1345> ChildEntities { get; set; }
    }

    public class RelatedEntity1345
    {
        public int Id { get; set; }

        public RelatedEntity1344 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1346> ChildEntities { get; set; }
    }

    public class RelatedEntity1346
    {
        public int Id { get; set; }

        public RelatedEntity1345 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1347> ChildEntities { get; set; }
    }

    public class RelatedEntity1347
    {
        public int Id { get; set; }

        public RelatedEntity1346 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1348> ChildEntities { get; set; }
    }

    public class RelatedEntity1348
    {
        public int Id { get; set; }

        public RelatedEntity1347 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1349> ChildEntities { get; set; }
    }

    public class RelatedEntity1349
    {
        public int Id { get; set; }

        public RelatedEntity1348 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1350> ChildEntities { get; set; }
    }

    public class RelatedEntity1350
    {
        public int Id { get; set; }

        public RelatedEntity1349 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1351> ChildEntities { get; set; }
    }

    public class RelatedEntity1351
    {
        public int Id { get; set; }

        public RelatedEntity1350 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1352> ChildEntities { get; set; }
    }

    public class RelatedEntity1352
    {
        public int Id { get; set; }

        public RelatedEntity1351 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1353> ChildEntities { get; set; }
    }

    public class RelatedEntity1353
    {
        public int Id { get; set; }

        public RelatedEntity1352 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1354> ChildEntities { get; set; }
    }

    public class RelatedEntity1354
    {
        public int Id { get; set; }

        public RelatedEntity1353 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1355> ChildEntities { get; set; }
    }

    public class RelatedEntity1355
    {
        public int Id { get; set; }

        public RelatedEntity1354 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1356> ChildEntities { get; set; }
    }

    public class RelatedEntity1356
    {
        public int Id { get; set; }

        public RelatedEntity1355 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1357> ChildEntities { get; set; }
    }

    public class RelatedEntity1357
    {
        public int Id { get; set; }

        public RelatedEntity1356 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1358> ChildEntities { get; set; }
    }

    public class RelatedEntity1358
    {
        public int Id { get; set; }

        public RelatedEntity1357 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1359> ChildEntities { get; set; }
    }

    public class RelatedEntity1359
    {
        public int Id { get; set; }

        public RelatedEntity1358 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1360> ChildEntities { get; set; }
    }

    public class RelatedEntity1360
    {
        public int Id { get; set; }

        public RelatedEntity1359 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1361> ChildEntities { get; set; }
    }

    public class RelatedEntity1361
    {
        public int Id { get; set; }

        public RelatedEntity1360 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1362> ChildEntities { get; set; }
    }

    public class RelatedEntity1362
    {
        public int Id { get; set; }

        public RelatedEntity1361 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1363> ChildEntities { get; set; }
    }

    public class RelatedEntity1363
    {
        public int Id { get; set; }

        public RelatedEntity1362 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1364> ChildEntities { get; set; }
    }

    public class RelatedEntity1364
    {
        public int Id { get; set; }

        public RelatedEntity1363 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1365> ChildEntities { get; set; }
    }

    public class RelatedEntity1365
    {
        public int Id { get; set; }

        public RelatedEntity1364 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1366> ChildEntities { get; set; }
    }

    public class RelatedEntity1366
    {
        public int Id { get; set; }

        public RelatedEntity1365 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1367> ChildEntities { get; set; }
    }

    public class RelatedEntity1367
    {
        public int Id { get; set; }

        public RelatedEntity1366 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1368> ChildEntities { get; set; }
    }

    public class RelatedEntity1368
    {
        public int Id { get; set; }

        public RelatedEntity1367 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1369> ChildEntities { get; set; }
    }

    public class RelatedEntity1369
    {
        public int Id { get; set; }

        public RelatedEntity1368 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1370> ChildEntities { get; set; }
    }

    public class RelatedEntity1370
    {
        public int Id { get; set; }

        public RelatedEntity1369 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1371> ChildEntities { get; set; }
    }

    public class RelatedEntity1371
    {
        public int Id { get; set; }

        public RelatedEntity1370 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1372> ChildEntities { get; set; }
    }

    public class RelatedEntity1372
    {
        public int Id { get; set; }

        public RelatedEntity1371 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1373> ChildEntities { get; set; }
    }

    public class RelatedEntity1373
    {
        public int Id { get; set; }

        public RelatedEntity1372 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1374> ChildEntities { get; set; }
    }

    public class RelatedEntity1374
    {
        public int Id { get; set; }

        public RelatedEntity1373 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1375> ChildEntities { get; set; }
    }

    public class RelatedEntity1375
    {
        public int Id { get; set; }

        public RelatedEntity1374 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1376> ChildEntities { get; set; }
    }

    public class RelatedEntity1376
    {
        public int Id { get; set; }

        public RelatedEntity1375 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1377> ChildEntities { get; set; }
    }

    public class RelatedEntity1377
    {
        public int Id { get; set; }

        public RelatedEntity1376 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1378> ChildEntities { get; set; }
    }

    public class RelatedEntity1378
    {
        public int Id { get; set; }

        public RelatedEntity1377 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1379> ChildEntities { get; set; }
    }

    public class RelatedEntity1379
    {
        public int Id { get; set; }

        public RelatedEntity1378 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1380> ChildEntities { get; set; }
    }

    public class RelatedEntity1380
    {
        public int Id { get; set; }

        public RelatedEntity1379 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1381> ChildEntities { get; set; }
    }

    public class RelatedEntity1381
    {
        public int Id { get; set; }

        public RelatedEntity1380 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1382> ChildEntities { get; set; }
    }

    public class RelatedEntity1382
    {
        public int Id { get; set; }

        public RelatedEntity1381 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1383> ChildEntities { get; set; }
    }

    public class RelatedEntity1383
    {
        public int Id { get; set; }

        public RelatedEntity1382 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1384> ChildEntities { get; set; }
    }

    public class RelatedEntity1384
    {
        public int Id { get; set; }

        public RelatedEntity1383 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1385> ChildEntities { get; set; }
    }

    public class RelatedEntity1385
    {
        public int Id { get; set; }

        public RelatedEntity1384 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1386> ChildEntities { get; set; }
    }

    public class RelatedEntity1386
    {
        public int Id { get; set; }

        public RelatedEntity1385 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1387> ChildEntities { get; set; }
    }

    public class RelatedEntity1387
    {
        public int Id { get; set; }

        public RelatedEntity1386 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1388> ChildEntities { get; set; }
    }

    public class RelatedEntity1388
    {
        public int Id { get; set; }

        public RelatedEntity1387 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1389> ChildEntities { get; set; }
    }

    public class RelatedEntity1389
    {
        public int Id { get; set; }

        public RelatedEntity1388 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1390> ChildEntities { get; set; }
    }

    public class RelatedEntity1390
    {
        public int Id { get; set; }

        public RelatedEntity1389 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1391> ChildEntities { get; set; }
    }

    public class RelatedEntity1391
    {
        public int Id { get; set; }

        public RelatedEntity1390 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1392> ChildEntities { get; set; }
    }

    public class RelatedEntity1392
    {
        public int Id { get; set; }

        public RelatedEntity1391 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1393> ChildEntities { get; set; }
    }

    public class RelatedEntity1393
    {
        public int Id { get; set; }

        public RelatedEntity1392 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1394> ChildEntities { get; set; }
    }

    public class RelatedEntity1394
    {
        public int Id { get; set; }

        public RelatedEntity1393 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1395> ChildEntities { get; set; }
    }

    public class RelatedEntity1395
    {
        public int Id { get; set; }

        public RelatedEntity1394 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1396> ChildEntities { get; set; }
    }

    public class RelatedEntity1396
    {
        public int Id { get; set; }

        public RelatedEntity1395 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1397> ChildEntities { get; set; }
    }

    public class RelatedEntity1397
    {
        public int Id { get; set; }

        public RelatedEntity1396 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1398> ChildEntities { get; set; }
    }

    public class RelatedEntity1398
    {
        public int Id { get; set; }

        public RelatedEntity1397 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1399> ChildEntities { get; set; }
    }

    public class RelatedEntity1399
    {
        public int Id { get; set; }

        public RelatedEntity1398 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1400> ChildEntities { get; set; }
    }

    public class RelatedEntity1400
    {
        public int Id { get; set; }

        public RelatedEntity1399 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1401> ChildEntities { get; set; }
    }

    public class RelatedEntity1401
    {
        public int Id { get; set; }

        public RelatedEntity1400 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1402> ChildEntities { get; set; }
    }

    public class RelatedEntity1402
    {
        public int Id { get; set; }

        public RelatedEntity1401 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1403> ChildEntities { get; set; }
    }

    public class RelatedEntity1403
    {
        public int Id { get; set; }

        public RelatedEntity1402 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1404> ChildEntities { get; set; }
    }

    public class RelatedEntity1404
    {
        public int Id { get; set; }

        public RelatedEntity1403 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1405> ChildEntities { get; set; }
    }

    public class RelatedEntity1405
    {
        public int Id { get; set; }

        public RelatedEntity1404 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1406> ChildEntities { get; set; }
    }

    public class RelatedEntity1406
    {
        public int Id { get; set; }

        public RelatedEntity1405 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1407> ChildEntities { get; set; }
    }

    public class RelatedEntity1407
    {
        public int Id { get; set; }

        public RelatedEntity1406 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1408> ChildEntities { get; set; }
    }

    public class RelatedEntity1408
    {
        public int Id { get; set; }

        public RelatedEntity1407 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1409> ChildEntities { get; set; }
    }

    public class RelatedEntity1409
    {
        public int Id { get; set; }

        public RelatedEntity1408 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1410> ChildEntities { get; set; }
    }

    public class RelatedEntity1410
    {
        public int Id { get; set; }

        public RelatedEntity1409 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1411> ChildEntities { get; set; }
    }

    public class RelatedEntity1411
    {
        public int Id { get; set; }

        public RelatedEntity1410 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1412> ChildEntities { get; set; }
    }

    public class RelatedEntity1412
    {
        public int Id { get; set; }

        public RelatedEntity1411 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1413> ChildEntities { get; set; }
    }

    public class RelatedEntity1413
    {
        public int Id { get; set; }

        public RelatedEntity1412 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1414> ChildEntities { get; set; }
    }

    public class RelatedEntity1414
    {
        public int Id { get; set; }

        public RelatedEntity1413 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1415> ChildEntities { get; set; }
    }

    public class RelatedEntity1415
    {
        public int Id { get; set; }

        public RelatedEntity1414 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1416> ChildEntities { get; set; }
    }

    public class RelatedEntity1416
    {
        public int Id { get; set; }

        public RelatedEntity1415 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1417> ChildEntities { get; set; }
    }

    public class RelatedEntity1417
    {
        public int Id { get; set; }

        public RelatedEntity1416 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1418> ChildEntities { get; set; }
    }

    public class RelatedEntity1418
    {
        public int Id { get; set; }

        public RelatedEntity1417 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1419> ChildEntities { get; set; }
    }

    public class RelatedEntity1419
    {
        public int Id { get; set; }

        public RelatedEntity1418 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1420> ChildEntities { get; set; }
    }

    public class RelatedEntity1420
    {
        public int Id { get; set; }

        public RelatedEntity1419 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1421> ChildEntities { get; set; }
    }

    public class RelatedEntity1421
    {
        public int Id { get; set; }

        public RelatedEntity1420 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1422> ChildEntities { get; set; }
    }

    public class RelatedEntity1422
    {
        public int Id { get; set; }

        public RelatedEntity1421 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1423> ChildEntities { get; set; }
    }

    public class RelatedEntity1423
    {
        public int Id { get; set; }

        public RelatedEntity1422 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1424> ChildEntities { get; set; }
    }

    public class RelatedEntity1424
    {
        public int Id { get; set; }

        public RelatedEntity1423 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1425> ChildEntities { get; set; }
    }

    public class RelatedEntity1425
    {
        public int Id { get; set; }

        public RelatedEntity1424 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1426> ChildEntities { get; set; }
    }

    public class RelatedEntity1426
    {
        public int Id { get; set; }

        public RelatedEntity1425 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1427> ChildEntities { get; set; }
    }

    public class RelatedEntity1427
    {
        public int Id { get; set; }

        public RelatedEntity1426 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1428> ChildEntities { get; set; }
    }

    public class RelatedEntity1428
    {
        public int Id { get; set; }

        public RelatedEntity1427 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1429> ChildEntities { get; set; }
    }

    public class RelatedEntity1429
    {
        public int Id { get; set; }

        public RelatedEntity1428 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1430> ChildEntities { get; set; }
    }

    public class RelatedEntity1430
    {
        public int Id { get; set; }

        public RelatedEntity1429 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1431> ChildEntities { get; set; }
    }

    public class RelatedEntity1431
    {
        public int Id { get; set; }

        public RelatedEntity1430 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1432> ChildEntities { get; set; }
    }

    public class RelatedEntity1432
    {
        public int Id { get; set; }

        public RelatedEntity1431 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1433> ChildEntities { get; set; }
    }

    public class RelatedEntity1433
    {
        public int Id { get; set; }

        public RelatedEntity1432 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1434> ChildEntities { get; set; }
    }

    public class RelatedEntity1434
    {
        public int Id { get; set; }

        public RelatedEntity1433 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1435> ChildEntities { get; set; }
    }

    public class RelatedEntity1435
    {
        public int Id { get; set; }

        public RelatedEntity1434 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1436> ChildEntities { get; set; }
    }

    public class RelatedEntity1436
    {
        public int Id { get; set; }

        public RelatedEntity1435 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1437> ChildEntities { get; set; }
    }

    public class RelatedEntity1437
    {
        public int Id { get; set; }

        public RelatedEntity1436 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1438> ChildEntities { get; set; }
    }

    public class RelatedEntity1438
    {
        public int Id { get; set; }

        public RelatedEntity1437 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1439> ChildEntities { get; set; }
    }

    public class RelatedEntity1439
    {
        public int Id { get; set; }

        public RelatedEntity1438 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1440> ChildEntities { get; set; }
    }

    public class RelatedEntity1440
    {
        public int Id { get; set; }

        public RelatedEntity1439 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1441> ChildEntities { get; set; }
    }

    public class RelatedEntity1441
    {
        public int Id { get; set; }

        public RelatedEntity1440 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1442> ChildEntities { get; set; }
    }

    public class RelatedEntity1442
    {
        public int Id { get; set; }

        public RelatedEntity1441 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1443> ChildEntities { get; set; }
    }

    public class RelatedEntity1443
    {
        public int Id { get; set; }

        public RelatedEntity1442 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1444> ChildEntities { get; set; }
    }

    public class RelatedEntity1444
    {
        public int Id { get; set; }

        public RelatedEntity1443 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1445> ChildEntities { get; set; }
    }

    public class RelatedEntity1445
    {
        public int Id { get; set; }

        public RelatedEntity1444 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1446> ChildEntities { get; set; }
    }

    public class RelatedEntity1446
    {
        public int Id { get; set; }

        public RelatedEntity1445 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1447> ChildEntities { get; set; }
    }

    public class RelatedEntity1447
    {
        public int Id { get; set; }

        public RelatedEntity1446 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1448> ChildEntities { get; set; }
    }

    public class RelatedEntity1448
    {
        public int Id { get; set; }

        public RelatedEntity1447 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1449> ChildEntities { get; set; }
    }

    public class RelatedEntity1449
    {
        public int Id { get; set; }

        public RelatedEntity1448 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1450> ChildEntities { get; set; }
    }

    public class RelatedEntity1450
    {
        public int Id { get; set; }

        public RelatedEntity1449 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1451> ChildEntities { get; set; }
    }

    public class RelatedEntity1451
    {
        public int Id { get; set; }

        public RelatedEntity1450 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1452> ChildEntities { get; set; }
    }

    public class RelatedEntity1452
    {
        public int Id { get; set; }

        public RelatedEntity1451 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1453> ChildEntities { get; set; }
    }

    public class RelatedEntity1453
    {
        public int Id { get; set; }

        public RelatedEntity1452 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1454> ChildEntities { get; set; }
    }

    public class RelatedEntity1454
    {
        public int Id { get; set; }

        public RelatedEntity1453 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1455> ChildEntities { get; set; }
    }

    public class RelatedEntity1455
    {
        public int Id { get; set; }

        public RelatedEntity1454 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1456> ChildEntities { get; set; }
    }

    public class RelatedEntity1456
    {
        public int Id { get; set; }

        public RelatedEntity1455 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1457> ChildEntities { get; set; }
    }

    public class RelatedEntity1457
    {
        public int Id { get; set; }

        public RelatedEntity1456 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1458> ChildEntities { get; set; }
    }

    public class RelatedEntity1458
    {
        public int Id { get; set; }

        public RelatedEntity1457 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1459> ChildEntities { get; set; }
    }

    public class RelatedEntity1459
    {
        public int Id { get; set; }

        public RelatedEntity1458 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1460> ChildEntities { get; set; }
    }

    public class RelatedEntity1460
    {
        public int Id { get; set; }

        public RelatedEntity1459 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1461> ChildEntities { get; set; }
    }

    public class RelatedEntity1461
    {
        public int Id { get; set; }

        public RelatedEntity1460 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1462> ChildEntities { get; set; }
    }

    public class RelatedEntity1462
    {
        public int Id { get; set; }

        public RelatedEntity1461 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1463> ChildEntities { get; set; }
    }

    public class RelatedEntity1463
    {
        public int Id { get; set; }

        public RelatedEntity1462 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1464> ChildEntities { get; set; }
    }

    public class RelatedEntity1464
    {
        public int Id { get; set; }

        public RelatedEntity1463 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1465> ChildEntities { get; set; }
    }

    public class RelatedEntity1465
    {
        public int Id { get; set; }

        public RelatedEntity1464 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1466> ChildEntities { get; set; }
    }

    public class RelatedEntity1466
    {
        public int Id { get; set; }

        public RelatedEntity1465 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1467> ChildEntities { get; set; }
    }

    public class RelatedEntity1467
    {
        public int Id { get; set; }

        public RelatedEntity1466 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1468> ChildEntities { get; set; }
    }

    public class RelatedEntity1468
    {
        public int Id { get; set; }

        public RelatedEntity1467 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1469> ChildEntities { get; set; }
    }

    public class RelatedEntity1469
    {
        public int Id { get; set; }

        public RelatedEntity1468 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1470> ChildEntities { get; set; }
    }

    public class RelatedEntity1470
    {
        public int Id { get; set; }

        public RelatedEntity1469 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1471> ChildEntities { get; set; }
    }

    public class RelatedEntity1471
    {
        public int Id { get; set; }

        public RelatedEntity1470 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1472> ChildEntities { get; set; }
    }

    public class RelatedEntity1472
    {
        public int Id { get; set; }

        public RelatedEntity1471 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1473> ChildEntities { get; set; }
    }

    public class RelatedEntity1473
    {
        public int Id { get; set; }

        public RelatedEntity1472 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1474> ChildEntities { get; set; }
    }

    public class RelatedEntity1474
    {
        public int Id { get; set; }

        public RelatedEntity1473 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1475> ChildEntities { get; set; }
    }

    public class RelatedEntity1475
    {
        public int Id { get; set; }

        public RelatedEntity1474 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1476> ChildEntities { get; set; }
    }

    public class RelatedEntity1476
    {
        public int Id { get; set; }

        public RelatedEntity1475 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1477> ChildEntities { get; set; }
    }

    public class RelatedEntity1477
    {
        public int Id { get; set; }

        public RelatedEntity1476 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1478> ChildEntities { get; set; }
    }

    public class RelatedEntity1478
    {
        public int Id { get; set; }

        public RelatedEntity1477 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1479> ChildEntities { get; set; }
    }

    public class RelatedEntity1479
    {
        public int Id { get; set; }

        public RelatedEntity1478 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1480> ChildEntities { get; set; }
    }

    public class RelatedEntity1480
    {
        public int Id { get; set; }

        public RelatedEntity1479 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1481> ChildEntities { get; set; }
    }

    public class RelatedEntity1481
    {
        public int Id { get; set; }

        public RelatedEntity1480 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1482> ChildEntities { get; set; }
    }

    public class RelatedEntity1482
    {
        public int Id { get; set; }

        public RelatedEntity1481 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1483> ChildEntities { get; set; }
    }

    public class RelatedEntity1483
    {
        public int Id { get; set; }

        public RelatedEntity1482 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1484> ChildEntities { get; set; }
    }

    public class RelatedEntity1484
    {
        public int Id { get; set; }

        public RelatedEntity1483 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1485> ChildEntities { get; set; }
    }

    public class RelatedEntity1485
    {
        public int Id { get; set; }

        public RelatedEntity1484 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1486> ChildEntities { get; set; }
    }

    public class RelatedEntity1486
    {
        public int Id { get; set; }

        public RelatedEntity1485 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1487> ChildEntities { get; set; }
    }

    public class RelatedEntity1487
    {
        public int Id { get; set; }

        public RelatedEntity1486 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1488> ChildEntities { get; set; }
    }

    public class RelatedEntity1488
    {
        public int Id { get; set; }

        public RelatedEntity1487 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1489> ChildEntities { get; set; }
    }

    public class RelatedEntity1489
    {
        public int Id { get; set; }

        public RelatedEntity1488 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1490> ChildEntities { get; set; }
    }

    public class RelatedEntity1490
    {
        public int Id { get; set; }

        public RelatedEntity1489 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1491> ChildEntities { get; set; }
    }

    public class RelatedEntity1491
    {
        public int Id { get; set; }

        public RelatedEntity1490 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1492> ChildEntities { get; set; }
    }

    public class RelatedEntity1492
    {
        public int Id { get; set; }

        public RelatedEntity1491 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1493> ChildEntities { get; set; }
    }

    public class RelatedEntity1493
    {
        public int Id { get; set; }

        public RelatedEntity1492 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1494> ChildEntities { get; set; }
    }

    public class RelatedEntity1494
    {
        public int Id { get; set; }

        public RelatedEntity1493 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1495> ChildEntities { get; set; }
    }

    public class RelatedEntity1495
    {
        public int Id { get; set; }

        public RelatedEntity1494 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1496> ChildEntities { get; set; }
    }

    public class RelatedEntity1496
    {
        public int Id { get; set; }

        public RelatedEntity1495 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1497> ChildEntities { get; set; }
    }

    public class RelatedEntity1497
    {
        public int Id { get; set; }

        public RelatedEntity1496 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1498> ChildEntities { get; set; }
    }

    public class RelatedEntity1498
    {
        public int Id { get; set; }

        public RelatedEntity1497 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1499> ChildEntities { get; set; }
    }

    public class RelatedEntity1499
    {
        public int Id { get; set; }

        public RelatedEntity1498 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1500> ChildEntities { get; set; }
    }

    public class RelatedEntity1500
    {
        public int Id { get; set; }

        public RelatedEntity1499 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1501> ChildEntities { get; set; }
    }

    public class RelatedEntity1501
    {
        public int Id { get; set; }

        public RelatedEntity1500 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1502> ChildEntities { get; set; }
    }

    public class RelatedEntity1502
    {
        public int Id { get; set; }

        public RelatedEntity1501 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1503> ChildEntities { get; set; }
    }

    public class RelatedEntity1503
    {
        public int Id { get; set; }

        public RelatedEntity1502 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1504> ChildEntities { get; set; }
    }

    public class RelatedEntity1504
    {
        public int Id { get; set; }

        public RelatedEntity1503 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1505> ChildEntities { get; set; }
    }

    public class RelatedEntity1505
    {
        public int Id { get; set; }

        public RelatedEntity1504 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1506> ChildEntities { get; set; }
    }

    public class RelatedEntity1506
    {
        public int Id { get; set; }

        public RelatedEntity1505 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1507> ChildEntities { get; set; }
    }

    public class RelatedEntity1507
    {
        public int Id { get; set; }

        public RelatedEntity1506 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1508> ChildEntities { get; set; }
    }

    public class RelatedEntity1508
    {
        public int Id { get; set; }

        public RelatedEntity1507 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1509> ChildEntities { get; set; }
    }

    public class RelatedEntity1509
    {
        public int Id { get; set; }

        public RelatedEntity1508 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1510> ChildEntities { get; set; }
    }

    public class RelatedEntity1510
    {
        public int Id { get; set; }

        public RelatedEntity1509 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1511> ChildEntities { get; set; }
    }

    public class RelatedEntity1511
    {
        public int Id { get; set; }

        public RelatedEntity1510 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1512> ChildEntities { get; set; }
    }

    public class RelatedEntity1512
    {
        public int Id { get; set; }

        public RelatedEntity1511 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1513> ChildEntities { get; set; }
    }

    public class RelatedEntity1513
    {
        public int Id { get; set; }

        public RelatedEntity1512 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1514> ChildEntities { get; set; }
    }

    public class RelatedEntity1514
    {
        public int Id { get; set; }

        public RelatedEntity1513 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1515> ChildEntities { get; set; }
    }

    public class RelatedEntity1515
    {
        public int Id { get; set; }

        public RelatedEntity1514 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1516> ChildEntities { get; set; }
    }

    public class RelatedEntity1516
    {
        public int Id { get; set; }

        public RelatedEntity1515 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1517> ChildEntities { get; set; }
    }

    public class RelatedEntity1517
    {
        public int Id { get; set; }

        public RelatedEntity1516 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1518> ChildEntities { get; set; }
    }

    public class RelatedEntity1518
    {
        public int Id { get; set; }

        public RelatedEntity1517 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1519> ChildEntities { get; set; }
    }

    public class RelatedEntity1519
    {
        public int Id { get; set; }

        public RelatedEntity1518 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1520> ChildEntities { get; set; }
    }

    public class RelatedEntity1520
    {
        public int Id { get; set; }

        public RelatedEntity1519 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1521> ChildEntities { get; set; }
    }

    public class RelatedEntity1521
    {
        public int Id { get; set; }

        public RelatedEntity1520 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1522> ChildEntities { get; set; }
    }

    public class RelatedEntity1522
    {
        public int Id { get; set; }

        public RelatedEntity1521 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1523> ChildEntities { get; set; }
    }

    public class RelatedEntity1523
    {
        public int Id { get; set; }

        public RelatedEntity1522 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1524> ChildEntities { get; set; }
    }

    public class RelatedEntity1524
    {
        public int Id { get; set; }

        public RelatedEntity1523 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1525> ChildEntities { get; set; }
    }

    public class RelatedEntity1525
    {
        public int Id { get; set; }

        public RelatedEntity1524 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1526> ChildEntities { get; set; }
    }

    public class RelatedEntity1526
    {
        public int Id { get; set; }

        public RelatedEntity1525 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1527> ChildEntities { get; set; }
    }

    public class RelatedEntity1527
    {
        public int Id { get; set; }

        public RelatedEntity1526 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1528> ChildEntities { get; set; }
    }

    public class RelatedEntity1528
    {
        public int Id { get; set; }

        public RelatedEntity1527 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1529> ChildEntities { get; set; }
    }

    public class RelatedEntity1529
    {
        public int Id { get; set; }

        public RelatedEntity1528 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1530> ChildEntities { get; set; }
    }

    public class RelatedEntity1530
    {
        public int Id { get; set; }

        public RelatedEntity1529 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1531> ChildEntities { get; set; }
    }

    public class RelatedEntity1531
    {
        public int Id { get; set; }

        public RelatedEntity1530 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1532> ChildEntities { get; set; }
    }

    public class RelatedEntity1532
    {
        public int Id { get; set; }

        public RelatedEntity1531 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1533> ChildEntities { get; set; }
    }

    public class RelatedEntity1533
    {
        public int Id { get; set; }

        public RelatedEntity1532 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1534> ChildEntities { get; set; }
    }

    public class RelatedEntity1534
    {
        public int Id { get; set; }

        public RelatedEntity1533 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1535> ChildEntities { get; set; }
    }

    public class RelatedEntity1535
    {
        public int Id { get; set; }

        public RelatedEntity1534 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1536> ChildEntities { get; set; }
    }

    public class RelatedEntity1536
    {
        public int Id { get; set; }

        public RelatedEntity1535 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1537> ChildEntities { get; set; }
    }

    public class RelatedEntity1537
    {
        public int Id { get; set; }

        public RelatedEntity1536 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1538> ChildEntities { get; set; }
    }

    public class RelatedEntity1538
    {
        public int Id { get; set; }

        public RelatedEntity1537 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1539> ChildEntities { get; set; }
    }

    public class RelatedEntity1539
    {
        public int Id { get; set; }

        public RelatedEntity1538 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1540> ChildEntities { get; set; }
    }

    public class RelatedEntity1540
    {
        public int Id { get; set; }

        public RelatedEntity1539 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1541> ChildEntities { get; set; }
    }

    public class RelatedEntity1541
    {
        public int Id { get; set; }

        public RelatedEntity1540 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1542> ChildEntities { get; set; }
    }

    public class RelatedEntity1542
    {
        public int Id { get; set; }

        public RelatedEntity1541 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1543> ChildEntities { get; set; }
    }

    public class RelatedEntity1543
    {
        public int Id { get; set; }

        public RelatedEntity1542 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1544> ChildEntities { get; set; }
    }

    public class RelatedEntity1544
    {
        public int Id { get; set; }

        public RelatedEntity1543 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1545> ChildEntities { get; set; }
    }

    public class RelatedEntity1545
    {
        public int Id { get; set; }

        public RelatedEntity1544 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1546> ChildEntities { get; set; }
    }

    public class RelatedEntity1546
    {
        public int Id { get; set; }

        public RelatedEntity1545 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1547> ChildEntities { get; set; }
    }

    public class RelatedEntity1547
    {
        public int Id { get; set; }

        public RelatedEntity1546 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1548> ChildEntities { get; set; }
    }

    public class RelatedEntity1548
    {
        public int Id { get; set; }

        public RelatedEntity1547 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1549> ChildEntities { get; set; }
    }

    public class RelatedEntity1549
    {
        public int Id { get; set; }

        public RelatedEntity1548 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1550> ChildEntities { get; set; }
    }

    public class RelatedEntity1550
    {
        public int Id { get; set; }

        public RelatedEntity1549 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1551> ChildEntities { get; set; }
    }

    public class RelatedEntity1551
    {
        public int Id { get; set; }

        public RelatedEntity1550 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1552> ChildEntities { get; set; }
    }

    public class RelatedEntity1552
    {
        public int Id { get; set; }

        public RelatedEntity1551 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1553> ChildEntities { get; set; }
    }

    public class RelatedEntity1553
    {
        public int Id { get; set; }

        public RelatedEntity1552 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1554> ChildEntities { get; set; }
    }

    public class RelatedEntity1554
    {
        public int Id { get; set; }

        public RelatedEntity1553 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1555> ChildEntities { get; set; }
    }

    public class RelatedEntity1555
    {
        public int Id { get; set; }

        public RelatedEntity1554 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1556> ChildEntities { get; set; }
    }

    public class RelatedEntity1556
    {
        public int Id { get; set; }

        public RelatedEntity1555 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1557> ChildEntities { get; set; }
    }

    public class RelatedEntity1557
    {
        public int Id { get; set; }

        public RelatedEntity1556 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1558> ChildEntities { get; set; }
    }

    public class RelatedEntity1558
    {
        public int Id { get; set; }

        public RelatedEntity1557 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1559> ChildEntities { get; set; }
    }

    public class RelatedEntity1559
    {
        public int Id { get; set; }

        public RelatedEntity1558 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1560> ChildEntities { get; set; }
    }

    public class RelatedEntity1560
    {
        public int Id { get; set; }

        public RelatedEntity1559 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1561> ChildEntities { get; set; }
    }

    public class RelatedEntity1561
    {
        public int Id { get; set; }

        public RelatedEntity1560 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1562> ChildEntities { get; set; }
    }

    public class RelatedEntity1562
    {
        public int Id { get; set; }

        public RelatedEntity1561 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1563> ChildEntities { get; set; }
    }

    public class RelatedEntity1563
    {
        public int Id { get; set; }

        public RelatedEntity1562 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1564> ChildEntities { get; set; }
    }

    public class RelatedEntity1564
    {
        public int Id { get; set; }

        public RelatedEntity1563 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1565> ChildEntities { get; set; }
    }

    public class RelatedEntity1565
    {
        public int Id { get; set; }

        public RelatedEntity1564 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1566> ChildEntities { get; set; }
    }

    public class RelatedEntity1566
    {
        public int Id { get; set; }

        public RelatedEntity1565 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1567> ChildEntities { get; set; }
    }

    public class RelatedEntity1567
    {
        public int Id { get; set; }

        public RelatedEntity1566 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1568> ChildEntities { get; set; }
    }

    public class RelatedEntity1568
    {
        public int Id { get; set; }

        public RelatedEntity1567 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1569> ChildEntities { get; set; }
    }

    public class RelatedEntity1569
    {
        public int Id { get; set; }

        public RelatedEntity1568 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1570> ChildEntities { get; set; }
    }

    public class RelatedEntity1570
    {
        public int Id { get; set; }

        public RelatedEntity1569 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1571> ChildEntities { get; set; }
    }

    public class RelatedEntity1571
    {
        public int Id { get; set; }

        public RelatedEntity1570 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1572> ChildEntities { get; set; }
    }

    public class RelatedEntity1572
    {
        public int Id { get; set; }

        public RelatedEntity1571 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1573> ChildEntities { get; set; }
    }

    public class RelatedEntity1573
    {
        public int Id { get; set; }

        public RelatedEntity1572 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1574> ChildEntities { get; set; }
    }

    public class RelatedEntity1574
    {
        public int Id { get; set; }

        public RelatedEntity1573 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1575> ChildEntities { get; set; }
    }

    public class RelatedEntity1575
    {
        public int Id { get; set; }

        public RelatedEntity1574 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1576> ChildEntities { get; set; }
    }

    public class RelatedEntity1576
    {
        public int Id { get; set; }

        public RelatedEntity1575 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1577> ChildEntities { get; set; }
    }

    public class RelatedEntity1577
    {
        public int Id { get; set; }

        public RelatedEntity1576 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1578> ChildEntities { get; set; }
    }

    public class RelatedEntity1578
    {
        public int Id { get; set; }

        public RelatedEntity1577 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1579> ChildEntities { get; set; }
    }

    public class RelatedEntity1579
    {
        public int Id { get; set; }

        public RelatedEntity1578 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1580> ChildEntities { get; set; }
    }

    public class RelatedEntity1580
    {
        public int Id { get; set; }

        public RelatedEntity1579 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1581> ChildEntities { get; set; }
    }

    public class RelatedEntity1581
    {
        public int Id { get; set; }

        public RelatedEntity1580 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1582> ChildEntities { get; set; }
    }

    public class RelatedEntity1582
    {
        public int Id { get; set; }

        public RelatedEntity1581 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1583> ChildEntities { get; set; }
    }

    public class RelatedEntity1583
    {
        public int Id { get; set; }

        public RelatedEntity1582 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1584> ChildEntities { get; set; }
    }

    public class RelatedEntity1584
    {
        public int Id { get; set; }

        public RelatedEntity1583 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1585> ChildEntities { get; set; }
    }

    public class RelatedEntity1585
    {
        public int Id { get; set; }

        public RelatedEntity1584 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1586> ChildEntities { get; set; }
    }

    public class RelatedEntity1586
    {
        public int Id { get; set; }

        public RelatedEntity1585 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1587> ChildEntities { get; set; }
    }

    public class RelatedEntity1587
    {
        public int Id { get; set; }

        public RelatedEntity1586 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1588> ChildEntities { get; set; }
    }

    public class RelatedEntity1588
    {
        public int Id { get; set; }

        public RelatedEntity1587 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1589> ChildEntities { get; set; }
    }

    public class RelatedEntity1589
    {
        public int Id { get; set; }

        public RelatedEntity1588 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1590> ChildEntities { get; set; }
    }

    public class RelatedEntity1590
    {
        public int Id { get; set; }

        public RelatedEntity1589 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1591> ChildEntities { get; set; }
    }

    public class RelatedEntity1591
    {
        public int Id { get; set; }

        public RelatedEntity1590 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1592> ChildEntities { get; set; }
    }

    public class RelatedEntity1592
    {
        public int Id { get; set; }

        public RelatedEntity1591 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1593> ChildEntities { get; set; }
    }

    public class RelatedEntity1593
    {
        public int Id { get; set; }

        public RelatedEntity1592 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1594> ChildEntities { get; set; }
    }

    public class RelatedEntity1594
    {
        public int Id { get; set; }

        public RelatedEntity1593 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1595> ChildEntities { get; set; }
    }

    public class RelatedEntity1595
    {
        public int Id { get; set; }

        public RelatedEntity1594 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1596> ChildEntities { get; set; }
    }

    public class RelatedEntity1596
    {
        public int Id { get; set; }

        public RelatedEntity1595 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1597> ChildEntities { get; set; }
    }

    public class RelatedEntity1597
    {
        public int Id { get; set; }

        public RelatedEntity1596 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1598> ChildEntities { get; set; }
    }

    public class RelatedEntity1598
    {
        public int Id { get; set; }

        public RelatedEntity1597 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1599> ChildEntities { get; set; }
    }

    public class RelatedEntity1599
    {
        public int Id { get; set; }

        public RelatedEntity1598 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1600> ChildEntities { get; set; }
    }

    public class RelatedEntity1600
    {
        public int Id { get; set; }

        public RelatedEntity1599 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1601> ChildEntities { get; set; }
    }

    public class RelatedEntity1601
    {
        public int Id { get; set; }

        public RelatedEntity1600 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1602> ChildEntities { get; set; }
    }

    public class RelatedEntity1602
    {
        public int Id { get; set; }

        public RelatedEntity1601 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1603> ChildEntities { get; set; }
    }

    public class RelatedEntity1603
    {
        public int Id { get; set; }

        public RelatedEntity1602 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1604> ChildEntities { get; set; }
    }

    public class RelatedEntity1604
    {
        public int Id { get; set; }

        public RelatedEntity1603 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1605> ChildEntities { get; set; }
    }

    public class RelatedEntity1605
    {
        public int Id { get; set; }

        public RelatedEntity1604 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1606> ChildEntities { get; set; }
    }

    public class RelatedEntity1606
    {
        public int Id { get; set; }

        public RelatedEntity1605 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1607> ChildEntities { get; set; }
    }

    public class RelatedEntity1607
    {
        public int Id { get; set; }

        public RelatedEntity1606 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1608> ChildEntities { get; set; }
    }

    public class RelatedEntity1608
    {
        public int Id { get; set; }

        public RelatedEntity1607 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1609> ChildEntities { get; set; }
    }

    public class RelatedEntity1609
    {
        public int Id { get; set; }

        public RelatedEntity1608 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1610> ChildEntities { get; set; }
    }

    public class RelatedEntity1610
    {
        public int Id { get; set; }

        public RelatedEntity1609 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1611> ChildEntities { get; set; }
    }

    public class RelatedEntity1611
    {
        public int Id { get; set; }

        public RelatedEntity1610 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1612> ChildEntities { get; set; }
    }

    public class RelatedEntity1612
    {
        public int Id { get; set; }

        public RelatedEntity1611 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1613> ChildEntities { get; set; }
    }

    public class RelatedEntity1613
    {
        public int Id { get; set; }

        public RelatedEntity1612 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1614> ChildEntities { get; set; }
    }

    public class RelatedEntity1614
    {
        public int Id { get; set; }

        public RelatedEntity1613 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1615> ChildEntities { get; set; }
    }

    public class RelatedEntity1615
    {
        public int Id { get; set; }

        public RelatedEntity1614 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1616> ChildEntities { get; set; }
    }

    public class RelatedEntity1616
    {
        public int Id { get; set; }

        public RelatedEntity1615 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1617> ChildEntities { get; set; }
    }

    public class RelatedEntity1617
    {
        public int Id { get; set; }

        public RelatedEntity1616 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1618> ChildEntities { get; set; }
    }

    public class RelatedEntity1618
    {
        public int Id { get; set; }

        public RelatedEntity1617 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1619> ChildEntities { get; set; }
    }

    public class RelatedEntity1619
    {
        public int Id { get; set; }

        public RelatedEntity1618 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1620> ChildEntities { get; set; }
    }

    public class RelatedEntity1620
    {
        public int Id { get; set; }

        public RelatedEntity1619 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1621> ChildEntities { get; set; }
    }

    public class RelatedEntity1621
    {
        public int Id { get; set; }

        public RelatedEntity1620 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1622> ChildEntities { get; set; }
    }

    public class RelatedEntity1622
    {
        public int Id { get; set; }

        public RelatedEntity1621 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1623> ChildEntities { get; set; }
    }

    public class RelatedEntity1623
    {
        public int Id { get; set; }

        public RelatedEntity1622 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1624> ChildEntities { get; set; }
    }

    public class RelatedEntity1624
    {
        public int Id { get; set; }

        public RelatedEntity1623 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1625> ChildEntities { get; set; }
    }

    public class RelatedEntity1625
    {
        public int Id { get; set; }

        public RelatedEntity1624 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1626> ChildEntities { get; set; }
    }

    public class RelatedEntity1626
    {
        public int Id { get; set; }

        public RelatedEntity1625 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1627> ChildEntities { get; set; }
    }

    public class RelatedEntity1627
    {
        public int Id { get; set; }

        public RelatedEntity1626 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1628> ChildEntities { get; set; }
    }

    public class RelatedEntity1628
    {
        public int Id { get; set; }

        public RelatedEntity1627 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1629> ChildEntities { get; set; }
    }

    public class RelatedEntity1629
    {
        public int Id { get; set; }

        public RelatedEntity1628 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1630> ChildEntities { get; set; }
    }

    public class RelatedEntity1630
    {
        public int Id { get; set; }

        public RelatedEntity1629 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1631> ChildEntities { get; set; }
    }

    public class RelatedEntity1631
    {
        public int Id { get; set; }

        public RelatedEntity1630 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1632> ChildEntities { get; set; }
    }

    public class RelatedEntity1632
    {
        public int Id { get; set; }

        public RelatedEntity1631 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1633> ChildEntities { get; set; }
    }

    public class RelatedEntity1633
    {
        public int Id { get; set; }

        public RelatedEntity1632 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1634> ChildEntities { get; set; }
    }

    public class RelatedEntity1634
    {
        public int Id { get; set; }

        public RelatedEntity1633 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1635> ChildEntities { get; set; }
    }

    public class RelatedEntity1635
    {
        public int Id { get; set; }

        public RelatedEntity1634 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1636> ChildEntities { get; set; }
    }

    public class RelatedEntity1636
    {
        public int Id { get; set; }

        public RelatedEntity1635 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1637> ChildEntities { get; set; }
    }

    public class RelatedEntity1637
    {
        public int Id { get; set; }

        public RelatedEntity1636 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1638> ChildEntities { get; set; }
    }

    public class RelatedEntity1638
    {
        public int Id { get; set; }

        public RelatedEntity1637 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1639> ChildEntities { get; set; }
    }

    public class RelatedEntity1639
    {
        public int Id { get; set; }

        public RelatedEntity1638 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1640> ChildEntities { get; set; }
    }

    public class RelatedEntity1640
    {
        public int Id { get; set; }

        public RelatedEntity1639 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1641> ChildEntities { get; set; }
    }

    public class RelatedEntity1641
    {
        public int Id { get; set; }

        public RelatedEntity1640 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1642> ChildEntities { get; set; }
    }

    public class RelatedEntity1642
    {
        public int Id { get; set; }

        public RelatedEntity1641 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1643> ChildEntities { get; set; }
    }

    public class RelatedEntity1643
    {
        public int Id { get; set; }

        public RelatedEntity1642 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1644> ChildEntities { get; set; }
    }

    public class RelatedEntity1644
    {
        public int Id { get; set; }

        public RelatedEntity1643 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1645> ChildEntities { get; set; }
    }

    public class RelatedEntity1645
    {
        public int Id { get; set; }

        public RelatedEntity1644 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1646> ChildEntities { get; set; }
    }

    public class RelatedEntity1646
    {
        public int Id { get; set; }

        public RelatedEntity1645 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1647> ChildEntities { get; set; }
    }

    public class RelatedEntity1647
    {
        public int Id { get; set; }

        public RelatedEntity1646 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1648> ChildEntities { get; set; }
    }

    public class RelatedEntity1648
    {
        public int Id { get; set; }

        public RelatedEntity1647 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1649> ChildEntities { get; set; }
    }

    public class RelatedEntity1649
    {
        public int Id { get; set; }

        public RelatedEntity1648 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1650> ChildEntities { get; set; }
    }

    public class RelatedEntity1650
    {
        public int Id { get; set; }

        public RelatedEntity1649 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1651> ChildEntities { get; set; }
    }

    public class RelatedEntity1651
    {
        public int Id { get; set; }

        public RelatedEntity1650 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1652> ChildEntities { get; set; }
    }

    public class RelatedEntity1652
    {
        public int Id { get; set; }

        public RelatedEntity1651 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1653> ChildEntities { get; set; }
    }

    public class RelatedEntity1653
    {
        public int Id { get; set; }

        public RelatedEntity1652 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1654> ChildEntities { get; set; }
    }

    public class RelatedEntity1654
    {
        public int Id { get; set; }

        public RelatedEntity1653 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1655> ChildEntities { get; set; }
    }

    public class RelatedEntity1655
    {
        public int Id { get; set; }

        public RelatedEntity1654 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1656> ChildEntities { get; set; }
    }

    public class RelatedEntity1656
    {
        public int Id { get; set; }

        public RelatedEntity1655 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1657> ChildEntities { get; set; }
    }

    public class RelatedEntity1657
    {
        public int Id { get; set; }

        public RelatedEntity1656 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1658> ChildEntities { get; set; }
    }

    public class RelatedEntity1658
    {
        public int Id { get; set; }

        public RelatedEntity1657 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1659> ChildEntities { get; set; }
    }

    public class RelatedEntity1659
    {
        public int Id { get; set; }

        public RelatedEntity1658 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1660> ChildEntities { get; set; }
    }

    public class RelatedEntity1660
    {
        public int Id { get; set; }

        public RelatedEntity1659 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1661> ChildEntities { get; set; }
    }

    public class RelatedEntity1661
    {
        public int Id { get; set; }

        public RelatedEntity1660 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1662> ChildEntities { get; set; }
    }

    public class RelatedEntity1662
    {
        public int Id { get; set; }

        public RelatedEntity1661 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1663> ChildEntities { get; set; }
    }

    public class RelatedEntity1663
    {
        public int Id { get; set; }

        public RelatedEntity1662 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1664> ChildEntities { get; set; }
    }

    public class RelatedEntity1664
    {
        public int Id { get; set; }

        public RelatedEntity1663 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1665> ChildEntities { get; set; }
    }

    public class RelatedEntity1665
    {
        public int Id { get; set; }

        public RelatedEntity1664 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1666> ChildEntities { get; set; }
    }

    public class RelatedEntity1666
    {
        public int Id { get; set; }

        public RelatedEntity1665 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1667> ChildEntities { get; set; }
    }

    public class RelatedEntity1667
    {
        public int Id { get; set; }

        public RelatedEntity1666 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1668> ChildEntities { get; set; }
    }

    public class RelatedEntity1668
    {
        public int Id { get; set; }

        public RelatedEntity1667 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1669> ChildEntities { get; set; }
    }

    public class RelatedEntity1669
    {
        public int Id { get; set; }

        public RelatedEntity1668 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1670> ChildEntities { get; set; }
    }

    public class RelatedEntity1670
    {
        public int Id { get; set; }

        public RelatedEntity1669 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1671> ChildEntities { get; set; }
    }

    public class RelatedEntity1671
    {
        public int Id { get; set; }

        public RelatedEntity1670 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1672> ChildEntities { get; set; }
    }

    public class RelatedEntity1672
    {
        public int Id { get; set; }

        public RelatedEntity1671 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1673> ChildEntities { get; set; }
    }

    public class RelatedEntity1673
    {
        public int Id { get; set; }

        public RelatedEntity1672 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1674> ChildEntities { get; set; }
    }

    public class RelatedEntity1674
    {
        public int Id { get; set; }

        public RelatedEntity1673 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1675> ChildEntities { get; set; }
    }

    public class RelatedEntity1675
    {
        public int Id { get; set; }

        public RelatedEntity1674 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1676> ChildEntities { get; set; }
    }

    public class RelatedEntity1676
    {
        public int Id { get; set; }

        public RelatedEntity1675 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1677> ChildEntities { get; set; }
    }

    public class RelatedEntity1677
    {
        public int Id { get; set; }

        public RelatedEntity1676 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1678> ChildEntities { get; set; }
    }

    public class RelatedEntity1678
    {
        public int Id { get; set; }

        public RelatedEntity1677 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1679> ChildEntities { get; set; }
    }

    public class RelatedEntity1679
    {
        public int Id { get; set; }

        public RelatedEntity1678 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1680> ChildEntities { get; set; }
    }

    public class RelatedEntity1680
    {
        public int Id { get; set; }

        public RelatedEntity1679 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1681> ChildEntities { get; set; }
    }

    public class RelatedEntity1681
    {
        public int Id { get; set; }

        public RelatedEntity1680 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1682> ChildEntities { get; set; }
    }

    public class RelatedEntity1682
    {
        public int Id { get; set; }

        public RelatedEntity1681 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1683> ChildEntities { get; set; }
    }

    public class RelatedEntity1683
    {
        public int Id { get; set; }

        public RelatedEntity1682 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1684> ChildEntities { get; set; }
    }

    public class RelatedEntity1684
    {
        public int Id { get; set; }

        public RelatedEntity1683 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1685> ChildEntities { get; set; }
    }

    public class RelatedEntity1685
    {
        public int Id { get; set; }

        public RelatedEntity1684 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1686> ChildEntities { get; set; }
    }

    public class RelatedEntity1686
    {
        public int Id { get; set; }

        public RelatedEntity1685 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1687> ChildEntities { get; set; }
    }

    public class RelatedEntity1687
    {
        public int Id { get; set; }

        public RelatedEntity1686 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1688> ChildEntities { get; set; }
    }

    public class RelatedEntity1688
    {
        public int Id { get; set; }

        public RelatedEntity1687 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1689> ChildEntities { get; set; }
    }

    public class RelatedEntity1689
    {
        public int Id { get; set; }

        public RelatedEntity1688 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1690> ChildEntities { get; set; }
    }

    public class RelatedEntity1690
    {
        public int Id { get; set; }

        public RelatedEntity1689 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1691> ChildEntities { get; set; }
    }

    public class RelatedEntity1691
    {
        public int Id { get; set; }

        public RelatedEntity1690 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1692> ChildEntities { get; set; }
    }

    public class RelatedEntity1692
    {
        public int Id { get; set; }

        public RelatedEntity1691 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1693> ChildEntities { get; set; }
    }

    public class RelatedEntity1693
    {
        public int Id { get; set; }

        public RelatedEntity1692 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1694> ChildEntities { get; set; }
    }

    public class RelatedEntity1694
    {
        public int Id { get; set; }

        public RelatedEntity1693 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1695> ChildEntities { get; set; }
    }

    public class RelatedEntity1695
    {
        public int Id { get; set; }

        public RelatedEntity1694 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1696> ChildEntities { get; set; }
    }

    public class RelatedEntity1696
    {
        public int Id { get; set; }

        public RelatedEntity1695 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1697> ChildEntities { get; set; }
    }

    public class RelatedEntity1697
    {
        public int Id { get; set; }

        public RelatedEntity1696 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1698> ChildEntities { get; set; }
    }

    public class RelatedEntity1698
    {
        public int Id { get; set; }

        public RelatedEntity1697 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1699> ChildEntities { get; set; }
    }

    public class RelatedEntity1699
    {
        public int Id { get; set; }

        public RelatedEntity1698 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1700> ChildEntities { get; set; }
    }

    public class RelatedEntity1700
    {
        public int Id { get; set; }

        public RelatedEntity1699 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1701> ChildEntities { get; set; }
    }

    public class RelatedEntity1701
    {
        public int Id { get; set; }

        public RelatedEntity1700 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1702> ChildEntities { get; set; }
    }

    public class RelatedEntity1702
    {
        public int Id { get; set; }

        public RelatedEntity1701 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1703> ChildEntities { get; set; }
    }

    public class RelatedEntity1703
    {
        public int Id { get; set; }

        public RelatedEntity1702 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1704> ChildEntities { get; set; }
    }

    public class RelatedEntity1704
    {
        public int Id { get; set; }

        public RelatedEntity1703 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1705> ChildEntities { get; set; }
    }

    public class RelatedEntity1705
    {
        public int Id { get; set; }

        public RelatedEntity1704 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1706> ChildEntities { get; set; }
    }

    public class RelatedEntity1706
    {
        public int Id { get; set; }

        public RelatedEntity1705 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1707> ChildEntities { get; set; }
    }

    public class RelatedEntity1707
    {
        public int Id { get; set; }

        public RelatedEntity1706 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1708> ChildEntities { get; set; }
    }

    public class RelatedEntity1708
    {
        public int Id { get; set; }

        public RelatedEntity1707 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1709> ChildEntities { get; set; }
    }

    public class RelatedEntity1709
    {
        public int Id { get; set; }

        public RelatedEntity1708 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1710> ChildEntities { get; set; }
    }

    public class RelatedEntity1710
    {
        public int Id { get; set; }

        public RelatedEntity1709 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1711> ChildEntities { get; set; }
    }

    public class RelatedEntity1711
    {
        public int Id { get; set; }

        public RelatedEntity1710 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1712> ChildEntities { get; set; }
    }

    public class RelatedEntity1712
    {
        public int Id { get; set; }

        public RelatedEntity1711 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1713> ChildEntities { get; set; }
    }

    public class RelatedEntity1713
    {
        public int Id { get; set; }

        public RelatedEntity1712 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1714> ChildEntities { get; set; }
    }

    public class RelatedEntity1714
    {
        public int Id { get; set; }

        public RelatedEntity1713 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1715> ChildEntities { get; set; }
    }

    public class RelatedEntity1715
    {
        public int Id { get; set; }

        public RelatedEntity1714 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1716> ChildEntities { get; set; }
    }

    public class RelatedEntity1716
    {
        public int Id { get; set; }

        public RelatedEntity1715 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1717> ChildEntities { get; set; }
    }

    public class RelatedEntity1717
    {
        public int Id { get; set; }

        public RelatedEntity1716 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1718> ChildEntities { get; set; }
    }

    public class RelatedEntity1718
    {
        public int Id { get; set; }

        public RelatedEntity1717 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1719> ChildEntities { get; set; }
    }

    public class RelatedEntity1719
    {
        public int Id { get; set; }

        public RelatedEntity1718 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1720> ChildEntities { get; set; }
    }

    public class RelatedEntity1720
    {
        public int Id { get; set; }

        public RelatedEntity1719 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1721> ChildEntities { get; set; }
    }

    public class RelatedEntity1721
    {
        public int Id { get; set; }

        public RelatedEntity1720 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1722> ChildEntities { get; set; }
    }

    public class RelatedEntity1722
    {
        public int Id { get; set; }

        public RelatedEntity1721 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1723> ChildEntities { get; set; }
    }

    public class RelatedEntity1723
    {
        public int Id { get; set; }

        public RelatedEntity1722 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1724> ChildEntities { get; set; }
    }

    public class RelatedEntity1724
    {
        public int Id { get; set; }

        public RelatedEntity1723 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1725> ChildEntities { get; set; }
    }

    public class RelatedEntity1725
    {
        public int Id { get; set; }

        public RelatedEntity1724 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1726> ChildEntities { get; set; }
    }

    public class RelatedEntity1726
    {
        public int Id { get; set; }

        public RelatedEntity1725 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1727> ChildEntities { get; set; }
    }

    public class RelatedEntity1727
    {
        public int Id { get; set; }

        public RelatedEntity1726 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1728> ChildEntities { get; set; }
    }

    public class RelatedEntity1728
    {
        public int Id { get; set; }

        public RelatedEntity1727 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1729> ChildEntities { get; set; }
    }

    public class RelatedEntity1729
    {
        public int Id { get; set; }

        public RelatedEntity1728 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1730> ChildEntities { get; set; }
    }

    public class RelatedEntity1730
    {
        public int Id { get; set; }

        public RelatedEntity1729 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1731> ChildEntities { get; set; }
    }

    public class RelatedEntity1731
    {
        public int Id { get; set; }

        public RelatedEntity1730 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1732> ChildEntities { get; set; }
    }

    public class RelatedEntity1732
    {
        public int Id { get; set; }

        public RelatedEntity1731 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1733> ChildEntities { get; set; }
    }

    public class RelatedEntity1733
    {
        public int Id { get; set; }

        public RelatedEntity1732 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1734> ChildEntities { get; set; }
    }

    public class RelatedEntity1734
    {
        public int Id { get; set; }

        public RelatedEntity1733 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1735> ChildEntities { get; set; }
    }

    public class RelatedEntity1735
    {
        public int Id { get; set; }

        public RelatedEntity1734 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1736> ChildEntities { get; set; }
    }

    public class RelatedEntity1736
    {
        public int Id { get; set; }

        public RelatedEntity1735 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1737> ChildEntities { get; set; }
    }

    public class RelatedEntity1737
    {
        public int Id { get; set; }

        public RelatedEntity1736 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1738> ChildEntities { get; set; }
    }

    public class RelatedEntity1738
    {
        public int Id { get; set; }

        public RelatedEntity1737 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1739> ChildEntities { get; set; }
    }

    public class RelatedEntity1739
    {
        public int Id { get; set; }

        public RelatedEntity1738 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1740> ChildEntities { get; set; }
    }

    public class RelatedEntity1740
    {
        public int Id { get; set; }

        public RelatedEntity1739 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1741> ChildEntities { get; set; }
    }

    public class RelatedEntity1741
    {
        public int Id { get; set; }

        public RelatedEntity1740 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1742> ChildEntities { get; set; }
    }

    public class RelatedEntity1742
    {
        public int Id { get; set; }

        public RelatedEntity1741 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1743> ChildEntities { get; set; }
    }

    public class RelatedEntity1743
    {
        public int Id { get; set; }

        public RelatedEntity1742 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1744> ChildEntities { get; set; }
    }

    public class RelatedEntity1744
    {
        public int Id { get; set; }

        public RelatedEntity1743 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1745> ChildEntities { get; set; }
    }

    public class RelatedEntity1745
    {
        public int Id { get; set; }

        public RelatedEntity1744 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1746> ChildEntities { get; set; }
    }

    public class RelatedEntity1746
    {
        public int Id { get; set; }

        public RelatedEntity1745 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1747> ChildEntities { get; set; }
    }

    public class RelatedEntity1747
    {
        public int Id { get; set; }

        public RelatedEntity1746 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1748> ChildEntities { get; set; }
    }

    public class RelatedEntity1748
    {
        public int Id { get; set; }

        public RelatedEntity1747 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1749> ChildEntities { get; set; }
    }

    public class RelatedEntity1749
    {
        public int Id { get; set; }

        public RelatedEntity1748 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1750> ChildEntities { get; set; }
    }

    public class RelatedEntity1750
    {
        public int Id { get; set; }

        public RelatedEntity1749 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1751> ChildEntities { get; set; }
    }

    public class RelatedEntity1751
    {
        public int Id { get; set; }

        public RelatedEntity1750 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1752> ChildEntities { get; set; }
    }

    public class RelatedEntity1752
    {
        public int Id { get; set; }

        public RelatedEntity1751 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1753> ChildEntities { get; set; }
    }

    public class RelatedEntity1753
    {
        public int Id { get; set; }

        public RelatedEntity1752 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1754> ChildEntities { get; set; }
    }

    public class RelatedEntity1754
    {
        public int Id { get; set; }

        public RelatedEntity1753 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1755> ChildEntities { get; set; }
    }

    public class RelatedEntity1755
    {
        public int Id { get; set; }

        public RelatedEntity1754 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1756> ChildEntities { get; set; }
    }

    public class RelatedEntity1756
    {
        public int Id { get; set; }

        public RelatedEntity1755 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1757> ChildEntities { get; set; }
    }

    public class RelatedEntity1757
    {
        public int Id { get; set; }

        public RelatedEntity1756 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1758> ChildEntities { get; set; }
    }

    public class RelatedEntity1758
    {
        public int Id { get; set; }

        public RelatedEntity1757 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1759> ChildEntities { get; set; }
    }

    public class RelatedEntity1759
    {
        public int Id { get; set; }

        public RelatedEntity1758 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1760> ChildEntities { get; set; }
    }

    public class RelatedEntity1760
    {
        public int Id { get; set; }

        public RelatedEntity1759 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1761> ChildEntities { get; set; }
    }

    public class RelatedEntity1761
    {
        public int Id { get; set; }

        public RelatedEntity1760 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1762> ChildEntities { get; set; }
    }

    public class RelatedEntity1762
    {
        public int Id { get; set; }

        public RelatedEntity1761 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1763> ChildEntities { get; set; }
    }

    public class RelatedEntity1763
    {
        public int Id { get; set; }

        public RelatedEntity1762 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1764> ChildEntities { get; set; }
    }

    public class RelatedEntity1764
    {
        public int Id { get; set; }

        public RelatedEntity1763 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1765> ChildEntities { get; set; }
    }

    public class RelatedEntity1765
    {
        public int Id { get; set; }

        public RelatedEntity1764 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1766> ChildEntities { get; set; }
    }

    public class RelatedEntity1766
    {
        public int Id { get; set; }

        public RelatedEntity1765 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1767> ChildEntities { get; set; }
    }

    public class RelatedEntity1767
    {
        public int Id { get; set; }

        public RelatedEntity1766 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1768> ChildEntities { get; set; }
    }

    public class RelatedEntity1768
    {
        public int Id { get; set; }

        public RelatedEntity1767 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1769> ChildEntities { get; set; }
    }

    public class RelatedEntity1769
    {
        public int Id { get; set; }

        public RelatedEntity1768 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1770> ChildEntities { get; set; }
    }

    public class RelatedEntity1770
    {
        public int Id { get; set; }

        public RelatedEntity1769 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1771> ChildEntities { get; set; }
    }

    public class RelatedEntity1771
    {
        public int Id { get; set; }

        public RelatedEntity1770 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1772> ChildEntities { get; set; }
    }

    public class RelatedEntity1772
    {
        public int Id { get; set; }

        public RelatedEntity1771 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1773> ChildEntities { get; set; }
    }

    public class RelatedEntity1773
    {
        public int Id { get; set; }

        public RelatedEntity1772 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1774> ChildEntities { get; set; }
    }

    public class RelatedEntity1774
    {
        public int Id { get; set; }

        public RelatedEntity1773 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1775> ChildEntities { get; set; }
    }

    public class RelatedEntity1775
    {
        public int Id { get; set; }

        public RelatedEntity1774 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1776> ChildEntities { get; set; }
    }

    public class RelatedEntity1776
    {
        public int Id { get; set; }

        public RelatedEntity1775 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1777> ChildEntities { get; set; }
    }

    public class RelatedEntity1777
    {
        public int Id { get; set; }

        public RelatedEntity1776 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1778> ChildEntities { get; set; }
    }

    public class RelatedEntity1778
    {
        public int Id { get; set; }

        public RelatedEntity1777 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1779> ChildEntities { get; set; }
    }

    public class RelatedEntity1779
    {
        public int Id { get; set; }

        public RelatedEntity1778 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1780> ChildEntities { get; set; }
    }

    public class RelatedEntity1780
    {
        public int Id { get; set; }

        public RelatedEntity1779 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1781> ChildEntities { get; set; }
    }

    public class RelatedEntity1781
    {
        public int Id { get; set; }

        public RelatedEntity1780 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1782> ChildEntities { get; set; }
    }

    public class RelatedEntity1782
    {
        public int Id { get; set; }

        public RelatedEntity1781 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1783> ChildEntities { get; set; }
    }

    public class RelatedEntity1783
    {
        public int Id { get; set; }

        public RelatedEntity1782 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1784> ChildEntities { get; set; }
    }

    public class RelatedEntity1784
    {
        public int Id { get; set; }

        public RelatedEntity1783 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1785> ChildEntities { get; set; }
    }

    public class RelatedEntity1785
    {
        public int Id { get; set; }

        public RelatedEntity1784 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1786> ChildEntities { get; set; }
    }

    public class RelatedEntity1786
    {
        public int Id { get; set; }

        public RelatedEntity1785 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1787> ChildEntities { get; set; }
    }

    public class RelatedEntity1787
    {
        public int Id { get; set; }

        public RelatedEntity1786 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1788> ChildEntities { get; set; }
    }

    public class RelatedEntity1788
    {
        public int Id { get; set; }

        public RelatedEntity1787 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1789> ChildEntities { get; set; }
    }

    public class RelatedEntity1789
    {
        public int Id { get; set; }

        public RelatedEntity1788 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1790> ChildEntities { get; set; }
    }

    public class RelatedEntity1790
    {
        public int Id { get; set; }

        public RelatedEntity1789 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1791> ChildEntities { get; set; }
    }

    public class RelatedEntity1791
    {
        public int Id { get; set; }

        public RelatedEntity1790 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1792> ChildEntities { get; set; }
    }

    public class RelatedEntity1792
    {
        public int Id { get; set; }

        public RelatedEntity1791 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1793> ChildEntities { get; set; }
    }

    public class RelatedEntity1793
    {
        public int Id { get; set; }

        public RelatedEntity1792 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1794> ChildEntities { get; set; }
    }

    public class RelatedEntity1794
    {
        public int Id { get; set; }

        public RelatedEntity1793 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1795> ChildEntities { get; set; }
    }

    public class RelatedEntity1795
    {
        public int Id { get; set; }

        public RelatedEntity1794 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1796> ChildEntities { get; set; }
    }

    public class RelatedEntity1796
    {
        public int Id { get; set; }

        public RelatedEntity1795 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1797> ChildEntities { get; set; }
    }

    public class RelatedEntity1797
    {
        public int Id { get; set; }

        public RelatedEntity1796 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1798> ChildEntities { get; set; }
    }

    public class RelatedEntity1798
    {
        public int Id { get; set; }

        public RelatedEntity1797 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1799> ChildEntities { get; set; }
    }

    public class RelatedEntity1799
    {
        public int Id { get; set; }

        public RelatedEntity1798 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1800> ChildEntities { get; set; }
    }

    public class RelatedEntity1800
    {
        public int Id { get; set; }

        public RelatedEntity1799 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1801> ChildEntities { get; set; }
    }

    public class RelatedEntity1801
    {
        public int Id { get; set; }

        public RelatedEntity1800 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1802> ChildEntities { get; set; }
    }

    public class RelatedEntity1802
    {
        public int Id { get; set; }

        public RelatedEntity1801 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1803> ChildEntities { get; set; }
    }

    public class RelatedEntity1803
    {
        public int Id { get; set; }

        public RelatedEntity1802 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1804> ChildEntities { get; set; }
    }

    public class RelatedEntity1804
    {
        public int Id { get; set; }

        public RelatedEntity1803 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1805> ChildEntities { get; set; }
    }

    public class RelatedEntity1805
    {
        public int Id { get; set; }

        public RelatedEntity1804 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1806> ChildEntities { get; set; }
    }

    public class RelatedEntity1806
    {
        public int Id { get; set; }

        public RelatedEntity1805 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1807> ChildEntities { get; set; }
    }

    public class RelatedEntity1807
    {
        public int Id { get; set; }

        public RelatedEntity1806 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1808> ChildEntities { get; set; }
    }

    public class RelatedEntity1808
    {
        public int Id { get; set; }

        public RelatedEntity1807 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1809> ChildEntities { get; set; }
    }

    public class RelatedEntity1809
    {
        public int Id { get; set; }

        public RelatedEntity1808 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1810> ChildEntities { get; set; }
    }

    public class RelatedEntity1810
    {
        public int Id { get; set; }

        public RelatedEntity1809 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1811> ChildEntities { get; set; }
    }

    public class RelatedEntity1811
    {
        public int Id { get; set; }

        public RelatedEntity1810 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1812> ChildEntities { get; set; }
    }

    public class RelatedEntity1812
    {
        public int Id { get; set; }

        public RelatedEntity1811 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1813> ChildEntities { get; set; }
    }

    public class RelatedEntity1813
    {
        public int Id { get; set; }

        public RelatedEntity1812 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1814> ChildEntities { get; set; }
    }

    public class RelatedEntity1814
    {
        public int Id { get; set; }

        public RelatedEntity1813 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1815> ChildEntities { get; set; }
    }

    public class RelatedEntity1815
    {
        public int Id { get; set; }

        public RelatedEntity1814 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1816> ChildEntities { get; set; }
    }

    public class RelatedEntity1816
    {
        public int Id { get; set; }

        public RelatedEntity1815 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1817> ChildEntities { get; set; }
    }

    public class RelatedEntity1817
    {
        public int Id { get; set; }

        public RelatedEntity1816 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1818> ChildEntities { get; set; }
    }

    public class RelatedEntity1818
    {
        public int Id { get; set; }

        public RelatedEntity1817 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1819> ChildEntities { get; set; }
    }

    public class RelatedEntity1819
    {
        public int Id { get; set; }

        public RelatedEntity1818 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1820> ChildEntities { get; set; }
    }

    public class RelatedEntity1820
    {
        public int Id { get; set; }

        public RelatedEntity1819 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1821> ChildEntities { get; set; }
    }

    public class RelatedEntity1821
    {
        public int Id { get; set; }

        public RelatedEntity1820 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1822> ChildEntities { get; set; }
    }

    public class RelatedEntity1822
    {
        public int Id { get; set; }

        public RelatedEntity1821 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1823> ChildEntities { get; set; }
    }

    public class RelatedEntity1823
    {
        public int Id { get; set; }

        public RelatedEntity1822 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1824> ChildEntities { get; set; }
    }

    public class RelatedEntity1824
    {
        public int Id { get; set; }

        public RelatedEntity1823 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1825> ChildEntities { get; set; }
    }

    public class RelatedEntity1825
    {
        public int Id { get; set; }

        public RelatedEntity1824 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1826> ChildEntities { get; set; }
    }

    public class RelatedEntity1826
    {
        public int Id { get; set; }

        public RelatedEntity1825 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1827> ChildEntities { get; set; }
    }

    public class RelatedEntity1827
    {
        public int Id { get; set; }

        public RelatedEntity1826 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1828> ChildEntities { get; set; }
    }

    public class RelatedEntity1828
    {
        public int Id { get; set; }

        public RelatedEntity1827 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1829> ChildEntities { get; set; }
    }

    public class RelatedEntity1829
    {
        public int Id { get; set; }

        public RelatedEntity1828 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1830> ChildEntities { get; set; }
    }

    public class RelatedEntity1830
    {
        public int Id { get; set; }

        public RelatedEntity1829 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1831> ChildEntities { get; set; }
    }

    public class RelatedEntity1831
    {
        public int Id { get; set; }

        public RelatedEntity1830 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1832> ChildEntities { get; set; }
    }

    public class RelatedEntity1832
    {
        public int Id { get; set; }

        public RelatedEntity1831 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1833> ChildEntities { get; set; }
    }

    public class RelatedEntity1833
    {
        public int Id { get; set; }

        public RelatedEntity1832 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1834> ChildEntities { get; set; }
    }

    public class RelatedEntity1834
    {
        public int Id { get; set; }

        public RelatedEntity1833 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1835> ChildEntities { get; set; }
    }

    public class RelatedEntity1835
    {
        public int Id { get; set; }

        public RelatedEntity1834 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1836> ChildEntities { get; set; }
    }

    public class RelatedEntity1836
    {
        public int Id { get; set; }

        public RelatedEntity1835 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1837> ChildEntities { get; set; }
    }

    public class RelatedEntity1837
    {
        public int Id { get; set; }

        public RelatedEntity1836 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1838> ChildEntities { get; set; }
    }

    public class RelatedEntity1838
    {
        public int Id { get; set; }

        public RelatedEntity1837 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1839> ChildEntities { get; set; }
    }

    public class RelatedEntity1839
    {
        public int Id { get; set; }

        public RelatedEntity1838 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1840> ChildEntities { get; set; }
    }

    public class RelatedEntity1840
    {
        public int Id { get; set; }

        public RelatedEntity1839 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1841> ChildEntities { get; set; }
    }

    public class RelatedEntity1841
    {
        public int Id { get; set; }

        public RelatedEntity1840 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1842> ChildEntities { get; set; }
    }

    public class RelatedEntity1842
    {
        public int Id { get; set; }

        public RelatedEntity1841 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1843> ChildEntities { get; set; }
    }

    public class RelatedEntity1843
    {
        public int Id { get; set; }

        public RelatedEntity1842 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1844> ChildEntities { get; set; }
    }

    public class RelatedEntity1844
    {
        public int Id { get; set; }

        public RelatedEntity1843 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1845> ChildEntities { get; set; }
    }

    public class RelatedEntity1845
    {
        public int Id { get; set; }

        public RelatedEntity1844 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1846> ChildEntities { get; set; }
    }

    public class RelatedEntity1846
    {
        public int Id { get; set; }

        public RelatedEntity1845 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1847> ChildEntities { get; set; }
    }

    public class RelatedEntity1847
    {
        public int Id { get; set; }

        public RelatedEntity1846 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1848> ChildEntities { get; set; }
    }

    public class RelatedEntity1848
    {
        public int Id { get; set; }

        public RelatedEntity1847 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1849> ChildEntities { get; set; }
    }

    public class RelatedEntity1849
    {
        public int Id { get; set; }

        public RelatedEntity1848 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1850> ChildEntities { get; set; }
    }

    public class RelatedEntity1850
    {
        public int Id { get; set; }

        public RelatedEntity1849 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1851> ChildEntities { get; set; }
    }

    public class RelatedEntity1851
    {
        public int Id { get; set; }

        public RelatedEntity1850 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1852> ChildEntities { get; set; }
    }

    public class RelatedEntity1852
    {
        public int Id { get; set; }

        public RelatedEntity1851 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1853> ChildEntities { get; set; }
    }

    public class RelatedEntity1853
    {
        public int Id { get; set; }

        public RelatedEntity1852 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1854> ChildEntities { get; set; }
    }

    public class RelatedEntity1854
    {
        public int Id { get; set; }

        public RelatedEntity1853 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1855> ChildEntities { get; set; }
    }

    public class RelatedEntity1855
    {
        public int Id { get; set; }

        public RelatedEntity1854 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1856> ChildEntities { get; set; }
    }

    public class RelatedEntity1856
    {
        public int Id { get; set; }

        public RelatedEntity1855 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1857> ChildEntities { get; set; }
    }

    public class RelatedEntity1857
    {
        public int Id { get; set; }

        public RelatedEntity1856 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1858> ChildEntities { get; set; }
    }

    public class RelatedEntity1858
    {
        public int Id { get; set; }

        public RelatedEntity1857 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1859> ChildEntities { get; set; }
    }

    public class RelatedEntity1859
    {
        public int Id { get; set; }

        public RelatedEntity1858 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1860> ChildEntities { get; set; }
    }

    public class RelatedEntity1860
    {
        public int Id { get; set; }

        public RelatedEntity1859 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1861> ChildEntities { get; set; }
    }

    public class RelatedEntity1861
    {
        public int Id { get; set; }

        public RelatedEntity1860 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1862> ChildEntities { get; set; }
    }

    public class RelatedEntity1862
    {
        public int Id { get; set; }

        public RelatedEntity1861 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1863> ChildEntities { get; set; }
    }

    public class RelatedEntity1863
    {
        public int Id { get; set; }

        public RelatedEntity1862 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1864> ChildEntities { get; set; }
    }

    public class RelatedEntity1864
    {
        public int Id { get; set; }

        public RelatedEntity1863 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1865> ChildEntities { get; set; }
    }

    public class RelatedEntity1865
    {
        public int Id { get; set; }

        public RelatedEntity1864 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1866> ChildEntities { get; set; }
    }

    public class RelatedEntity1866
    {
        public int Id { get; set; }

        public RelatedEntity1865 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1867> ChildEntities { get; set; }
    }

    public class RelatedEntity1867
    {
        public int Id { get; set; }

        public RelatedEntity1866 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1868> ChildEntities { get; set; }
    }

    public class RelatedEntity1868
    {
        public int Id { get; set; }

        public RelatedEntity1867 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1869> ChildEntities { get; set; }
    }

    public class RelatedEntity1869
    {
        public int Id { get; set; }

        public RelatedEntity1868 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1870> ChildEntities { get; set; }
    }

    public class RelatedEntity1870
    {
        public int Id { get; set; }

        public RelatedEntity1869 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1871> ChildEntities { get; set; }
    }

    public class RelatedEntity1871
    {
        public int Id { get; set; }

        public RelatedEntity1870 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1872> ChildEntities { get; set; }
    }

    public class RelatedEntity1872
    {
        public int Id { get; set; }

        public RelatedEntity1871 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1873> ChildEntities { get; set; }
    }

    public class RelatedEntity1873
    {
        public int Id { get; set; }

        public RelatedEntity1872 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1874> ChildEntities { get; set; }
    }

    public class RelatedEntity1874
    {
        public int Id { get; set; }

        public RelatedEntity1873 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1875> ChildEntities { get; set; }
    }

    public class RelatedEntity1875
    {
        public int Id { get; set; }

        public RelatedEntity1874 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1876> ChildEntities { get; set; }
    }

    public class RelatedEntity1876
    {
        public int Id { get; set; }

        public RelatedEntity1875 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1877> ChildEntities { get; set; }
    }

    public class RelatedEntity1877
    {
        public int Id { get; set; }

        public RelatedEntity1876 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1878> ChildEntities { get; set; }
    }

    public class RelatedEntity1878
    {
        public int Id { get; set; }

        public RelatedEntity1877 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1879> ChildEntities { get; set; }
    }

    public class RelatedEntity1879
    {
        public int Id { get; set; }

        public RelatedEntity1878 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1880> ChildEntities { get; set; }
    }

    public class RelatedEntity1880
    {
        public int Id { get; set; }

        public RelatedEntity1879 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1881> ChildEntities { get; set; }
    }

    public class RelatedEntity1881
    {
        public int Id { get; set; }

        public RelatedEntity1880 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1882> ChildEntities { get; set; }
    }

    public class RelatedEntity1882
    {
        public int Id { get; set; }

        public RelatedEntity1881 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1883> ChildEntities { get; set; }
    }

    public class RelatedEntity1883
    {
        public int Id { get; set; }

        public RelatedEntity1882 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1884> ChildEntities { get; set; }
    }

    public class RelatedEntity1884
    {
        public int Id { get; set; }

        public RelatedEntity1883 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1885> ChildEntities { get; set; }
    }

    public class RelatedEntity1885
    {
        public int Id { get; set; }

        public RelatedEntity1884 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1886> ChildEntities { get; set; }
    }

    public class RelatedEntity1886
    {
        public int Id { get; set; }

        public RelatedEntity1885 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1887> ChildEntities { get; set; }
    }

    public class RelatedEntity1887
    {
        public int Id { get; set; }

        public RelatedEntity1886 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1888> ChildEntities { get; set; }
    }

    public class RelatedEntity1888
    {
        public int Id { get; set; }

        public RelatedEntity1887 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1889> ChildEntities { get; set; }
    }

    public class RelatedEntity1889
    {
        public int Id { get; set; }

        public RelatedEntity1888 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1890> ChildEntities { get; set; }
    }

    public class RelatedEntity1890
    {
        public int Id { get; set; }

        public RelatedEntity1889 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1891> ChildEntities { get; set; }
    }

    public class RelatedEntity1891
    {
        public int Id { get; set; }

        public RelatedEntity1890 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1892> ChildEntities { get; set; }
    }

    public class RelatedEntity1892
    {
        public int Id { get; set; }

        public RelatedEntity1891 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1893> ChildEntities { get; set; }
    }

    public class RelatedEntity1893
    {
        public int Id { get; set; }

        public RelatedEntity1892 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1894> ChildEntities { get; set; }
    }

    public class RelatedEntity1894
    {
        public int Id { get; set; }

        public RelatedEntity1893 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1895> ChildEntities { get; set; }
    }

    public class RelatedEntity1895
    {
        public int Id { get; set; }

        public RelatedEntity1894 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1896> ChildEntities { get; set; }
    }

    public class RelatedEntity1896
    {
        public int Id { get; set; }

        public RelatedEntity1895 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1897> ChildEntities { get; set; }
    }

    public class RelatedEntity1897
    {
        public int Id { get; set; }

        public RelatedEntity1896 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1898> ChildEntities { get; set; }
    }

    public class RelatedEntity1898
    {
        public int Id { get; set; }

        public RelatedEntity1897 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1899> ChildEntities { get; set; }
    }

    public class RelatedEntity1899
    {
        public int Id { get; set; }

        public RelatedEntity1898 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1900> ChildEntities { get; set; }
    }

    public class RelatedEntity1900
    {
        public int Id { get; set; }

        public RelatedEntity1899 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1901> ChildEntities { get; set; }
    }

    public class RelatedEntity1901
    {
        public int Id { get; set; }

        public RelatedEntity1900 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1902> ChildEntities { get; set; }
    }

    public class RelatedEntity1902
    {
        public int Id { get; set; }

        public RelatedEntity1901 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1903> ChildEntities { get; set; }
    }

    public class RelatedEntity1903
    {
        public int Id { get; set; }

        public RelatedEntity1902 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1904> ChildEntities { get; set; }
    }

    public class RelatedEntity1904
    {
        public int Id { get; set; }

        public RelatedEntity1903 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1905> ChildEntities { get; set; }
    }

    public class RelatedEntity1905
    {
        public int Id { get; set; }

        public RelatedEntity1904 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1906> ChildEntities { get; set; }
    }

    public class RelatedEntity1906
    {
        public int Id { get; set; }

        public RelatedEntity1905 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1907> ChildEntities { get; set; }
    }

    public class RelatedEntity1907
    {
        public int Id { get; set; }

        public RelatedEntity1906 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1908> ChildEntities { get; set; }
    }

    public class RelatedEntity1908
    {
        public int Id { get; set; }

        public RelatedEntity1907 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1909> ChildEntities { get; set; }
    }

    public class RelatedEntity1909
    {
        public int Id { get; set; }

        public RelatedEntity1908 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1910> ChildEntities { get; set; }
    }

    public class RelatedEntity1910
    {
        public int Id { get; set; }

        public RelatedEntity1909 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1911> ChildEntities { get; set; }
    }

    public class RelatedEntity1911
    {
        public int Id { get; set; }

        public RelatedEntity1910 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1912> ChildEntities { get; set; }
    }

    public class RelatedEntity1912
    {
        public int Id { get; set; }

        public RelatedEntity1911 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1913> ChildEntities { get; set; }
    }

    public class RelatedEntity1913
    {
        public int Id { get; set; }

        public RelatedEntity1912 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1914> ChildEntities { get; set; }
    }

    public class RelatedEntity1914
    {
        public int Id { get; set; }

        public RelatedEntity1913 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1915> ChildEntities { get; set; }
    }

    public class RelatedEntity1915
    {
        public int Id { get; set; }

        public RelatedEntity1914 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1916> ChildEntities { get; set; }
    }

    public class RelatedEntity1916
    {
        public int Id { get; set; }

        public RelatedEntity1915 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1917> ChildEntities { get; set; }
    }

    public class RelatedEntity1917
    {
        public int Id { get; set; }

        public RelatedEntity1916 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1918> ChildEntities { get; set; }
    }

    public class RelatedEntity1918
    {
        public int Id { get; set; }

        public RelatedEntity1917 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1919> ChildEntities { get; set; }
    }

    public class RelatedEntity1919
    {
        public int Id { get; set; }

        public RelatedEntity1918 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1920> ChildEntities { get; set; }
    }

    public class RelatedEntity1920
    {
        public int Id { get; set; }

        public RelatedEntity1919 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1921> ChildEntities { get; set; }
    }

    public class RelatedEntity1921
    {
        public int Id { get; set; }

        public RelatedEntity1920 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1922> ChildEntities { get; set; }
    }

    public class RelatedEntity1922
    {
        public int Id { get; set; }

        public RelatedEntity1921 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1923> ChildEntities { get; set; }
    }

    public class RelatedEntity1923
    {
        public int Id { get; set; }

        public RelatedEntity1922 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1924> ChildEntities { get; set; }
    }

    public class RelatedEntity1924
    {
        public int Id { get; set; }

        public RelatedEntity1923 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1925> ChildEntities { get; set; }
    }

    public class RelatedEntity1925
    {
        public int Id { get; set; }

        public RelatedEntity1924 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1926> ChildEntities { get; set; }
    }

    public class RelatedEntity1926
    {
        public int Id { get; set; }

        public RelatedEntity1925 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1927> ChildEntities { get; set; }
    }

    public class RelatedEntity1927
    {
        public int Id { get; set; }

        public RelatedEntity1926 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1928> ChildEntities { get; set; }
    }

    public class RelatedEntity1928
    {
        public int Id { get; set; }

        public RelatedEntity1927 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1929> ChildEntities { get; set; }
    }

    public class RelatedEntity1929
    {
        public int Id { get; set; }

        public RelatedEntity1928 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1930> ChildEntities { get; set; }
    }

    public class RelatedEntity1930
    {
        public int Id { get; set; }

        public RelatedEntity1929 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1931> ChildEntities { get; set; }
    }

    public class RelatedEntity1931
    {
        public int Id { get; set; }

        public RelatedEntity1930 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1932> ChildEntities { get; set; }
    }

    public class RelatedEntity1932
    {
        public int Id { get; set; }

        public RelatedEntity1931 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1933> ChildEntities { get; set; }
    }

    public class RelatedEntity1933
    {
        public int Id { get; set; }

        public RelatedEntity1932 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1934> ChildEntities { get; set; }
    }

    public class RelatedEntity1934
    {
        public int Id { get; set; }

        public RelatedEntity1933 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1935> ChildEntities { get; set; }
    }

    public class RelatedEntity1935
    {
        public int Id { get; set; }

        public RelatedEntity1934 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1936> ChildEntities { get; set; }
    }

    public class RelatedEntity1936
    {
        public int Id { get; set; }

        public RelatedEntity1935 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1937> ChildEntities { get; set; }
    }

    public class RelatedEntity1937
    {
        public int Id { get; set; }

        public RelatedEntity1936 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1938> ChildEntities { get; set; }
    }

    public class RelatedEntity1938
    {
        public int Id { get; set; }

        public RelatedEntity1937 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1939> ChildEntities { get; set; }
    }

    public class RelatedEntity1939
    {
        public int Id { get; set; }

        public RelatedEntity1938 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1940> ChildEntities { get; set; }
    }

    public class RelatedEntity1940
    {
        public int Id { get; set; }

        public RelatedEntity1939 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1941> ChildEntities { get; set; }
    }

    public class RelatedEntity1941
    {
        public int Id { get; set; }

        public RelatedEntity1940 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1942> ChildEntities { get; set; }
    }

    public class RelatedEntity1942
    {
        public int Id { get; set; }

        public RelatedEntity1941 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1943> ChildEntities { get; set; }
    }

    public class RelatedEntity1943
    {
        public int Id { get; set; }

        public RelatedEntity1942 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1944> ChildEntities { get; set; }
    }

    public class RelatedEntity1944
    {
        public int Id { get; set; }

        public RelatedEntity1943 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1945> ChildEntities { get; set; }
    }

    public class RelatedEntity1945
    {
        public int Id { get; set; }

        public RelatedEntity1944 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1946> ChildEntities { get; set; }
    }

    public class RelatedEntity1946
    {
        public int Id { get; set; }

        public RelatedEntity1945 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1947> ChildEntities { get; set; }
    }

    public class RelatedEntity1947
    {
        public int Id { get; set; }

        public RelatedEntity1946 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1948> ChildEntities { get; set; }
    }

    public class RelatedEntity1948
    {
        public int Id { get; set; }

        public RelatedEntity1947 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1949> ChildEntities { get; set; }
    }

    public class RelatedEntity1949
    {
        public int Id { get; set; }

        public RelatedEntity1948 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1950> ChildEntities { get; set; }
    }

    public class RelatedEntity1950
    {
        public int Id { get; set; }

        public RelatedEntity1949 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1951> ChildEntities { get; set; }
    }

    public class RelatedEntity1951
    {
        public int Id { get; set; }

        public RelatedEntity1950 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1952> ChildEntities { get; set; }
    }

    public class RelatedEntity1952
    {
        public int Id { get; set; }

        public RelatedEntity1951 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1953> ChildEntities { get; set; }
    }

    public class RelatedEntity1953
    {
        public int Id { get; set; }

        public RelatedEntity1952 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1954> ChildEntities { get; set; }
    }

    public class RelatedEntity1954
    {
        public int Id { get; set; }

        public RelatedEntity1953 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1955> ChildEntities { get; set; }
    }

    public class RelatedEntity1955
    {
        public int Id { get; set; }

        public RelatedEntity1954 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1956> ChildEntities { get; set; }
    }

    public class RelatedEntity1956
    {
        public int Id { get; set; }

        public RelatedEntity1955 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1957> ChildEntities { get; set; }
    }

    public class RelatedEntity1957
    {
        public int Id { get; set; }

        public RelatedEntity1956 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1958> ChildEntities { get; set; }
    }

    public class RelatedEntity1958
    {
        public int Id { get; set; }

        public RelatedEntity1957 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1959> ChildEntities { get; set; }
    }

    public class RelatedEntity1959
    {
        public int Id { get; set; }

        public RelatedEntity1958 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1960> ChildEntities { get; set; }
    }

    public class RelatedEntity1960
    {
        public int Id { get; set; }

        public RelatedEntity1959 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1961> ChildEntities { get; set; }
    }

    public class RelatedEntity1961
    {
        public int Id { get; set; }

        public RelatedEntity1960 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1962> ChildEntities { get; set; }
    }

    public class RelatedEntity1962
    {
        public int Id { get; set; }

        public RelatedEntity1961 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1963> ChildEntities { get; set; }
    }

    public class RelatedEntity1963
    {
        public int Id { get; set; }

        public RelatedEntity1962 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1964> ChildEntities { get; set; }
    }

    public class RelatedEntity1964
    {
        public int Id { get; set; }

        public RelatedEntity1963 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1965> ChildEntities { get; set; }
    }

    public class RelatedEntity1965
    {
        public int Id { get; set; }

        public RelatedEntity1964 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1966> ChildEntities { get; set; }
    }

    public class RelatedEntity1966
    {
        public int Id { get; set; }

        public RelatedEntity1965 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1967> ChildEntities { get; set; }
    }

    public class RelatedEntity1967
    {
        public int Id { get; set; }

        public RelatedEntity1966 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1968> ChildEntities { get; set; }
    }

    public class RelatedEntity1968
    {
        public int Id { get; set; }

        public RelatedEntity1967 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1969> ChildEntities { get; set; }
    }

    public class RelatedEntity1969
    {
        public int Id { get; set; }

        public RelatedEntity1968 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1970> ChildEntities { get; set; }
    }

    public class RelatedEntity1970
    {
        public int Id { get; set; }

        public RelatedEntity1969 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1971> ChildEntities { get; set; }
    }

    public class RelatedEntity1971
    {
        public int Id { get; set; }

        public RelatedEntity1970 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1972> ChildEntities { get; set; }
    }

    public class RelatedEntity1972
    {
        public int Id { get; set; }

        public RelatedEntity1971 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1973> ChildEntities { get; set; }
    }

    public class RelatedEntity1973
    {
        public int Id { get; set; }

        public RelatedEntity1972 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1974> ChildEntities { get; set; }
    }

    public class RelatedEntity1974
    {
        public int Id { get; set; }

        public RelatedEntity1973 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1975> ChildEntities { get; set; }
    }

    public class RelatedEntity1975
    {
        public int Id { get; set; }

        public RelatedEntity1974 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1976> ChildEntities { get; set; }
    }

    public class RelatedEntity1976
    {
        public int Id { get; set; }

        public RelatedEntity1975 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1977> ChildEntities { get; set; }
    }

    public class RelatedEntity1977
    {
        public int Id { get; set; }

        public RelatedEntity1976 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1978> ChildEntities { get; set; }
    }

    public class RelatedEntity1978
    {
        public int Id { get; set; }

        public RelatedEntity1977 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1979> ChildEntities { get; set; }
    }

    public class RelatedEntity1979
    {
        public int Id { get; set; }

        public RelatedEntity1978 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1980> ChildEntities { get; set; }
    }

    public class RelatedEntity1980
    {
        public int Id { get; set; }

        public RelatedEntity1979 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1981> ChildEntities { get; set; }
    }

    public class RelatedEntity1981
    {
        public int Id { get; set; }

        public RelatedEntity1980 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1982> ChildEntities { get; set; }
    }

    public class RelatedEntity1982
    {
        public int Id { get; set; }

        public RelatedEntity1981 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1983> ChildEntities { get; set; }
    }

    public class RelatedEntity1983
    {
        public int Id { get; set; }

        public RelatedEntity1982 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1984> ChildEntities { get; set; }
    }

    public class RelatedEntity1984
    {
        public int Id { get; set; }

        public RelatedEntity1983 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1985> ChildEntities { get; set; }
    }

    public class RelatedEntity1985
    {
        public int Id { get; set; }

        public RelatedEntity1984 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1986> ChildEntities { get; set; }
    }

    public class RelatedEntity1986
    {
        public int Id { get; set; }

        public RelatedEntity1985 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1987> ChildEntities { get; set; }
    }

    public class RelatedEntity1987
    {
        public int Id { get; set; }

        public RelatedEntity1986 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1988> ChildEntities { get; set; }
    }

    public class RelatedEntity1988
    {
        public int Id { get; set; }

        public RelatedEntity1987 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1989> ChildEntities { get; set; }
    }

    public class RelatedEntity1989
    {
        public int Id { get; set; }

        public RelatedEntity1988 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1990> ChildEntities { get; set; }
    }

    public class RelatedEntity1990
    {
        public int Id { get; set; }

        public RelatedEntity1989 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1991> ChildEntities { get; set; }
    }

    public class RelatedEntity1991
    {
        public int Id { get; set; }

        public RelatedEntity1990 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1992> ChildEntities { get; set; }
    }

    public class RelatedEntity1992
    {
        public int Id { get; set; }

        public RelatedEntity1991 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1993> ChildEntities { get; set; }
    }

    public class RelatedEntity1993
    {
        public int Id { get; set; }

        public RelatedEntity1992 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1994> ChildEntities { get; set; }
    }

    public class RelatedEntity1994
    {
        public int Id { get; set; }

        public RelatedEntity1993 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1995> ChildEntities { get; set; }
    }

    public class RelatedEntity1995
    {
        public int Id { get; set; }

        public RelatedEntity1994 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1996> ChildEntities { get; set; }
    }

    public class RelatedEntity1996
    {
        public int Id { get; set; }

        public RelatedEntity1995 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1997> ChildEntities { get; set; }
    }

    public class RelatedEntity1997
    {
        public int Id { get; set; }

        public RelatedEntity1996 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1998> ChildEntities { get; set; }
    }

    public class RelatedEntity1998
    {
        public int Id { get; set; }

        public RelatedEntity1997 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1999> ChildEntities { get; set; }
    }

    public class RelatedEntity1999
    {
        public int Id { get; set; }

        public RelatedEntity1998 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity2000> ChildEntities { get; set; }
    }

    public class RelatedEntity2000
    {
        public int Id { get; set; }

        public RelatedEntity1999 ParentEntity { get; set; }

        public IEnumerable<RelatedEntity1> ChildEntities { get; set; }
    }
}
